#ifdef FCGI
#include <fcgi_stdio.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "charsets.h"

static unsigned char* tables[N_OF_SETS][N_OF_SETS];

static char* charset_string[N_OF_SETS] = {
	"utf-8", "koi8-r", "windows-1251", "dos-866", "cp-866",
	"iso-8859-5", "x-mac-cyrillic"
};


static int charsets[][N_OF_SETS] = {
	//Unicode  koi8r cp1251	  dos  cp866  8859-5   mac
	{0x00a0,   0x9a,  0xa0,	 0xff,	0xff,  0xa0,  0xca}, // NO-BREAK SPACE
	{0x00a2,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xa2}, // CENT SIGN
	{0x00a3,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xa3}, // POUND SIGN
	{0x00a4,   0x00,  0xa4,	 0x00,	0xfd,  0x00,  0xff}, // CURRENCY SIGN
	{0x00a6,   0x00,  0xa6,	 0x00,	0x00,  0x00,  0x00}, // BROKEN BAR
	{0x00a7,   0x00,  0xa7,	 0x00,	0x00,  0xfd,  0xa4}, // SECTION SIGN
	{0x00a9,   0xbf,  0xa9,	 0xfc,	0x00,  0x00,  0xa9}, // COPYRIGHT SIGN
	{0x00ab,   0x00,  0xab,	 0x00,	0x00,  0x00,  0xc7}, // LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	{0x00ac,   0x00,  0xac,	 0x00,	0x00,  0x00,  0xc2}, // NOT SIGN
	{0x00ad,   0x00,  0xad,	 0x00,	0x00,  0xad,  0x00}, // SOFT HYPHEN
	{0x00ae,   0x00,  0xae,	 0x00,	0x00,  0x00,  0xa8}, // REGISTERED SIGN
	{0x00b0,   0x9c,  0xb0,	 0xf8,	0xf8,  0x00,  0xa1}, // DEGREE SIGN
	{0x00b1,   0x00,  0xb1,	 0x00,	0x00,  0x00,  0xb1}, // PLUS-MINUS SIGN
	{0x00b2,   0x9d,  0x00,	 0xfd,	0x00,  0x00,  0x00}, // SUPERSCRIPT TWO
	{0x00b5,   0x00,  0xb5,	 0x00,	0x00,  0x00,  0xb5}, // MICRO SIGN
	{0x00b6,   0x00,  0xb6,	 0x00,	0x00,  0x00,  0xa6}, // PILCROW SIGN
	{0x00b7,   0x9e,  0xb7,	 0xfa,	0xfa,  0x00,  0x00}, // MIDDLE DOT
	{0x00bb,   0x00,  0xbb,	 0x00,	0x00,  0x00,  0xc8}, // RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	{0x00f7,   0x9f,  0x00,	 0xf6,	0x00,  0x00,  0xd6}, // DIVISION SIGN
	{0x0192,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xc4}, // LATIN SMALL LETTER F WITH HOOK
	{0x0401,   0xb3,  0xa8,	 0xf0,	0xf0,  0xa1,  0xdd}, // CYRILLIC CAPITAL LETTER IO
	{0x0402,   0x00,  0x80,	 0x00,	0x00,  0xa2,  0xab}, // CYRILLIC CAPITAL LETTER DJE
	{0x0403,   0x00,  0x81,	 0x00,	0x00,  0xa3,  0xae}, // CYRILLIC CAPITAL LETTER GJE
	{0x0404,   0x00,  0xaa,	 0x00,	0xf2,  0xa4,  0xb8}, // CYRILLIC CAPITAL LETTER UKRAINIAN IE
	{0x0405,   0x00,  0xbd,	 0x00,	0x00,  0xa5,  0xc1}, // CYRILLIC CAPITAL LETTER DZE
	{0x0406,   0x00,  0xb2,	 0x00,	0x00,  0xa6,  0xa7}, // CYRILLIC CAPITAL LETTER BYELORUSSIAN-UKRAINIAN I
	{0x0407,   0x00,  0xaf,	 0x00,	0xf4,  0xa7,  0xba}, // CYRILLIC CAPITAL LETTER YI
	{0x0408,   0x00,  0xa3,	 0x00,	0x00,  0xa8,  0xb7}, // CYRILLIC CAPITAL LETTER JE
	{0x0409,   0x00,  0x8a,	 0x00,	0x00,  0xa9,  0xbc}, // CYRILLIC CAPITAL LETTER LJE
	{0x040a,   0x00,  0x8c,	 0x00,	0x00,  0xaa,  0xbe}, // CYRILLIC CAPITAL LETTER NJE
	{0x040b,   0x00,  0x8e,	 0x00,	0x00,  0xab,  0xcb}, // CYRILLIC CAPITAL LETTER TSHE
	{0x040c,   0x00,  0x8d,	 0x00,	0x00,  0xac,  0xcd}, // CYRILLIC CAPITAL LETTER KJE
	{0x040e,   0x00,  0xa1,	 0x00,	0xf6,  0xae,  0xd8}, // CYRILLIC CAPITAL LETTER SHORT U
	{0x040f,   0x00,  0x8f,	 0x00,	0x00,  0xaf,  0xda}, // CYRILLIC CAPITAL LETTER DZHE
	{0x0410,   0xe1,  0xc0,	 0x80,	0x80,  0xb0,  0x80}, // CYRILLIC CAPITAL LETTER A
	{0x0411,   0xe2,  0xc1,	 0x81,	0x81,  0xb1,  0x81}, // CYRILLIC CAPITAL LETTER BE
	{0x0412,   0xf7,  0xc2,	 0x82,	0x82,  0xb2,  0x82}, // CYRILLIC CAPITAL LETTER VE
	{0x0413,   0xe7,  0xc3,	 0x83,	0x83,  0xb3,  0x83}, // CYRILLIC CAPITAL LETTER GHE
	{0x0414,   0xe4,  0xc4,	 0x84,	0x84,  0xb4,  0x84}, // CYRILLIC CAPITAL LETTER DE
	{0x0415,   0xe5,  0xc5,	 0x85,	0x85,  0xb5,  0x85}, // CYRILLIC CAPITAL LETTER IE
	{0x0416,   0xf6,  0xc6,	 0x86,	0x86,  0xb6,  0x86}, // CYRILLIC CAPITAL LETTER ZHE
	{0x0417,   0xfa,  0xc7,	 0x87,	0x87,  0xb7,  0x87}, // CYRILLIC CAPITAL LETTER ZE
	{0x0418,   0xe9,  0xc8,	 0x88,	0x88,  0xb8,  0x88}, // CYRILLIC CAPITAL LETTER I
	{0x0419,   0xea,  0xc9,	 0x89,	0x89,  0xb9,  0x89}, // CYRILLIC CAPITAL LETTER SHORT I
	{0x041a,   0xeb,  0xca,	 0x8a,	0x8a,  0xba,  0x8a}, // CYRILLIC CAPITAL LETTER KA
	{0x041b,   0xec,  0xcb,	 0x8b,	0x8b,  0xbb,  0x8b}, // CYRILLIC CAPITAL LETTER EL
	{0x041c,   0xed,  0xcc,	 0x8c,	0x8c,  0xbc,  0x8c}, // CYRILLIC CAPITAL LETTER EM
	{0x041d,   0xee,  0xcd,	 0x8d,	0x8d,  0xbd,  0x8d}, // CYRILLIC CAPITAL LETTER EN
	{0x041e,   0xef,  0xce,	 0x8e,	0x8e,  0xbe,  0x8e}, // CYRILLIC CAPITAL LETTER O
	{0x041f,   0xf0,  0xcf,	 0x8f,	0x8f,  0xbf,  0x8f}, // CYRILLIC CAPITAL LETTER PE
	{0x0420,   0xf2,  0xd0,	 0x90,	0x90,  0xc0,  0x90}, // CYRILLIC CAPITAL LETTER ER
	{0x0421,   0xf3,  0xd1,	 0x91,	0x91,  0xc1,  0x91}, // CYRILLIC CAPITAL LETTER ES
	{0x0422,   0xf4,  0xd2,	 0x92,	0x92,  0xc2,  0x92}, // CYRILLIC CAPITAL LETTER TE
	{0x0423,   0xf5,  0xd3,	 0x93,	0x93,  0xc3,  0x93}, // CYRILLIC CAPITAL LETTER U
	{0x0424,   0xe6,  0xd4,	 0x94,	0x94,  0xc4,  0x94}, // CYRILLIC CAPITAL LETTER EF
	{0x0425,   0xe8,  0xd5,	 0x95,	0x95,  0xc5,  0x95}, // CYRILLIC CAPITAL LETTER HA
	{0x0426,   0xe3,  0xd6,	 0x96,	0x96,  0xc6,  0x96}, // CYRILLIC CAPITAL LETTER TSE
	{0x0427,   0xfe,  0xd7,	 0x97,	0x97,  0xc7,  0x97}, // CYRILLIC CAPITAL LETTER CHE
	{0x0428,   0xfb,  0xd8,	 0x98,	0x98,  0xc8,  0x98}, // CYRILLIC CAPITAL LETTER SHA
	{0x0429,   0xfd,  0xd9,	 0x99,	0x99,  0xc9,  0x99}, // CYRILLIC CAPITAL LETTER SHCHA
	{0x042a,   0xff,  0xda,	 0x9a,	0x9a,  0xca,  0x9a}, // CYRILLIC CAPITAL LETTER HARD SIGN
	{0x042b,   0xf9,  0xdb,	 0x9b,	0x9b,  0xcb,  0x9b}, // CYRILLIC CAPITAL LETTER YERU
	{0x042c,   0xf8,  0xdc,	 0x9c,	0x9c,  0xcc,  0x9c}, // CYRILLIC CAPITAL LETTER SOFT SIGN
	{0x042d,   0xfc,  0xdd,	 0x9d,	0x9d,  0xcd,  0x9d}, // CYRILLIC CAPITAL LETTER E
	{0x042e,   0xe0,  0xde,	 0x9e,	0x9e,  0xce,  0x9e}, // CYRILLIC CAPITAL LETTER YU
	{0x042f,   0xf1,  0xdf,	 0x9f,	0x9f,  0xcf,  0x9f}, // CYRILLIC CAPITAL LETTER YA
	{0x0430,   0xc1,  0xe0,	 0xa0,	0xa0,  0xd0,  0xe0}, // CYRILLIC SMALL LETTER A
	{0x0431,   0xc2,  0xe1,	 0xa1,	0xa1,  0xd1,  0xe1}, // CYRILLIC SMALL LETTER BE
	{0x0432,   0xd7,  0xe2,	 0xa2,	0xa2,  0xd2,  0xe2}, // CYRILLIC SMALL LETTER VE
	{0x0433,   0xc7,  0xe3,	 0xa3,	0xa3,  0xd3,  0xe3}, // CYRILLIC SMALL LETTER GHE
	{0x0434,   0xc4,  0xe4,	 0xa4,	0xa4,  0xd4,  0xe4}, // CYRILLIC SMALL LETTER DE
	{0x0435,   0xc5,  0xe5,	 0xa5,	0xa5,  0xd5,  0xe5}, // CYRILLIC SMALL LETTER IE
	{0x0436,   0xd6,  0xe6,	 0xa6,	0xa6,  0xd6,  0xe6}, // CYRILLIC SMALL LETTER ZHE
	{0x0437,   0xda,  0xe7,	 0xa7,	0xa7,  0xd7,  0xe7}, // CYRILLIC SMALL LETTER ZE
	{0x0438,   0xc9,  0xe8,	 0xa8,	0xa8,  0xd8,  0xe8}, // CYRILLIC SMALL LETTER I
	{0x0439,   0xca,  0xe9,	 0xa9,	0xa9,  0xd9,  0xe9}, // CYRILLIC SMALL LETTER SHORT I
	{0x043a,   0xcb,  0xea,	 0xaa,	0xaa,  0xda,  0xea}, // CYRILLIC SMALL LETTER KA
	{0x043b,   0xcc,  0xeb,	 0xab,	0xab,  0xdb,  0xeb}, // CYRILLIC SMALL LETTER EL
	{0x043c,   0xcd,  0xec,	 0xac,	0xac,  0xdc,  0xec}, // CYRILLIC SMALL LETTER EM
	{0x043d,   0xce,  0xed,	 0xad,	0xad,  0xdd,  0xed}, // CYRILLIC SMALL LETTER EN
	{0x043e,   0xcf,  0xee,	 0xae,	0xae,  0xde,  0xee}, // CYRILLIC SMALL LETTER O
	{0x043f,   0xd0,  0xef,	 0xaf,	0xaf,  0xdf,  0xef}, // CYRILLIC SMALL LETTER PE
	{0x0440,   0xd2,  0xf0,	 0xe0,	0xe0,  0xe0,  0xf0}, // CYRILLIC SMALL LETTER ER
	{0x0441,   0xd3,  0xf1,	 0xe1,	0xe1,  0xe1,  0xf1}, // CYRILLIC SMALL LETTER ES
	{0x0442,   0xd4,  0xf2,	 0xe2,	0xe2,  0xe2,  0xf2}, // CYRILLIC SMALL LETTER TE
	{0x0443,   0xd5,  0xf3,	 0xe3,	0xe3,  0xe3,  0xf3}, // CYRILLIC SMALL LETTER U
	{0x0444,   0xc6,  0xf4,	 0xe4,	0xe4,  0xe4,  0xf4}, // CYRILLIC SMALL LETTER EF
	{0x0445,   0xc8,  0xf5,	 0xe5,	0xe5,  0xe5,  0xf5}, // CYRILLIC SMALL LETTER HA
	{0x0446,   0xc3,  0xf6,	 0xe6,	0xe6,  0xe6,  0xf6}, // CYRILLIC SMALL LETTER TSE
	{0x0447,   0xde,  0xf7,	 0xe7,	0xe7,  0xe7,  0xf7}, // CYRILLIC SMALL LETTER CHE
	{0x0448,   0xdb,  0xf8,	 0xe8,	0xe8,  0xe8,  0xf8}, // CYRILLIC SMALL LETTER SHA
	{0x0449,   0xdd,  0xf9,	 0xe9,	0xe9,  0xe9,  0xf9}, // CYRILLIC SMALL LETTER SHCHA
	{0x044a,   0xdf,  0xfa,	 0xea,	0xea,  0xea,  0xfa}, // CYRILLIC SMALL LETTER HARD SIGN
	{0x044b,   0xd9,  0xfb,	 0xeb,	0xeb,  0xeb,  0xfb}, // CYRILLIC SMALL LETTER YERU
	{0x044c,   0xd8,  0xfc,	 0xec,	0xec,  0xec,  0xfc}, // CYRILLIC SMALL LETTER SOFT SIGN
	{0x044d,   0xdc,  0xfd,	 0xed,	0xed,  0xed,  0xfd}, // CYRILLIC SMALL LETTER E
	{0x044e,   0xc0,  0xfe,	 0xee,	0xee,  0xee,  0xfe}, // CYRILLIC SMALL LETTER YU
	{0x044f,   0xd1,  0xff,	 0xef,	0xef,  0xef,  0xdf}, // CYRILLIC SMALL LETTER YA
	{0x0451,   0xa3,  0xb8,	 0xf1,	0xf1,  0xf1,  0xde}, // CYRILLIC SMALL LETTER IO
	{0x0452,   0x00,  0x90,	 0x00,	0x00,  0xf2,  0xac}, // CYRILLIC SMALL LETTER DJE
	{0x0453,   0x00,  0x83,	 0x00,	0x00,  0xf3,  0xaf}, // CYRILLIC SMALL LETTER GJE
	{0x0454,   0x00,  0xba,	 0x00,	0xf3,  0xf4,  0xb9}, // CYRILLIC SMALL LETTER UKRAINIAN IE
	{0x0455,   0x00,  0xbe,	 0x00,	0x00,  0xf5,  0xcf}, // CYRILLIC SMALL LETTER DZE
	{0x0456,   0x00,  0xb3,	 0x00,	0x00,  0xf6,  0xb4}, // CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
	{0x0457,   0x00,  0xbf,	 0x00,	0xf5,  0xf7,  0xbb}, // CYRILLIC SMALL LETTER YI
	{0x0458,   0x00,  0xbc,	 0x00,	0x00,  0xf8,  0xc0}, // CYRILLIC SMALL LETTER JE
	{0x0459,   0x00,  0x9a,	 0x00,	0x00,  0xf9,  0xbd}, // CYRILLIC SMALL LETTER LJE
	{0x045a,   0x00,  0x9c,	 0x00,	0x00,  0xfa,  0xbf}, // CYRILLIC SMALL LETTER NJE
	{0x045b,   0x00,  0x9e,	 0x00,	0x00,  0xfb,  0xcc}, // CYRILLIC SMALL LETTER TSHE
	{0x045c,   0x00,  0x9d,	 0x00,	0x00,  0xfc,  0xce}, // CYRILLIC SMALL LETTER KJE
	{0x045e,   0x00,  0xa2,	 0x00,	0xf7,  0xfe,  0xd9}, // CYRILLIC SMALL LETTER SHORT U
	{0x045f,   0x00,  0x9f,	 0x00,	0x00,  0xff,  0xdb}, // CYRILLIC SMALL LETTER DZHE
	{0x0490,   0x00,  0xa5,	 0x00,	0x00,  0x00,  0x00}, // CYRILLIC CAPITAL LETTER GHE WITH UPTURN
	{0x0491,   0x00,  0xb4,	 0x00,	0x00,  0x00,  0x00}, // CYRILLIC SMALL LETTER GHE WITH UPTURN
	{0x2013,   0x00,  0x96,	 0x00,	0x00,  0x00,  0xd0}, // EN DASH
	{0x2014,   0x00,  0x97,	 0x00,	0x00,  0x00,  0xd1}, // EM DASH
	{0x2018,   0x00,  0x91,	 0x00,	0x00,  0x00,  0xd4}, // LEFT SINGLE QUOTATION MARK
	{0x2019,   0x00,  0x92,	 0x00,	0x00,  0x00,  0xd5}, // RIGHT SINGLE QUOTATION MARK
	{0x201a,   0x00,  0x82,	 0x00,	0x00,  0x00,  0x00}, // SINGLE LOW-9,QUOTATION MARK
	{0x201c,   0x00,  0x93,	 0x00,	0x00,  0x00,  0xd2}, // LEFT DOUBLE QUOTATION MARK
	{0x201d,   0x00,  0x94,	 0x00,	0x00,  0x00,  0xd3}, // RIGHT DOUBLE QUOTATION MARK
	{0x201e,   0x00,  0x84,	 0x00,	0x00,  0x00,  0xd7}, // DOUBLE LOW-9,QUOTATION MARK
	{0x2020,   0x00,  0x86,	 0x00,	0x00,  0x00,  0xa0}, // DAGGER
	{0x2021,   0x00,  0x87,	 0x00,	0x00,  0x00,  0x00}, // DOUBLE DAGGER
	{0x2022,   0x00,  0x95,	 0x00,	0x00,  0x00,  0xa5}, // BULLET
	{0x2026,   0x00,  0x85,	 0x00,	0x00,  0x00,  0xc9}, // HORIZONTAL ELLIPSIS
	{0x2030,   0x00,  0x89,	 0x00,	0x00,  0x00,  0x00}, // PER MILLE SIGN
	{0x2039,   0x00,  0x8b,	 0x00,	0x00,  0x00,  0x00}, // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	{0x203a,   0x00,  0x9b,	 0x00,	0x00,  0x00,  0x00}, // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	{0x2116,   0x00,  0xb9,	 0x00,	0xfc,  0xf0,  0xdc}, // NUMERO SIGN
	{0x2122,   0x00,  0x99,	 0x00,	0x00,  0x00,  0xaa}, // TRADE MARK SIGN
	{0x2202,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xb6}, // PARTIAL DIFFERENTIAL
	{0x2206,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xc6}, // INCREMENT
	{0x2219,   0x95,  0x00,	 0xf9,	0xf9,  0x00,  0x00}, // BULLET OPERATOR
	{0x221a,   0x96,  0x00,	 0xfb,	0xfb,  0x00,  0xc3}, // SQUARE ROOT
	{0x221e,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xb0}, // INFINITY
	{0x2248,   0x97,  0x00,	 0xf7,	0x00,  0x00,  0xc5}, // ALMOST EQUAL TO
	{0x2260,   0x00,  0x00,	 0x00,	0x00,  0x00,  0xad}, // NOT EQUAL TO
	{0x2264,   0x98,  0x00,	 0xf3,	0x00,  0x00,  0xb2}, // LESS-THAN OR EQUAL TO
	{0x2265,   0x99,  0x00,	 0xf2,	0x00,  0x00,  0xb3}, // GREATER-THAN OR EQUAL TO
	{0x2320,   0x93,  0x00,	 0xf4,	0x00,  0x00,  0x00}, // TOP HALF INTEGRAL
	{0x2321,   0x9b,  0x00,	 0xf5,	0x00,  0x00,  0x00}, // BOTTOM HALF INTEGRAL
	{0x2500,   0x80,  0x00,	 0xc4,	0xc4,  0x00,  0x00}, // BOX DRAWINGS LIGHT HORIZONTAL
	{0x2502,   0x81,  0x00,	 0xb3,	0xb3,  0x00,  0x00}, // BOX DRAWINGS LIGHT VERTICAL
	{0x250c,   0x82,  0x00,	 0xda,	0xda,  0x00,  0x00}, // BOX DRAWINGS LIGHT DOWN AND RIGHT
	{0x2510,   0x83,  0x00,	 0xbf,	0xbf,  0x00,  0x00}, // BOX DRAWINGS LIGHT DOWN AND LEFT
	{0x2514,   0x84,  0x00,	 0xc0,	0xc0,  0x00,  0x00}, // BOX DRAWINGS LIGHT UP AND RIGHT
	{0x2518,   0x85,  0x00,	 0xd9,	0xd9,  0x00,  0x00}, // BOX DRAWINGS LIGHT UP AND LEFT
	{0x251c,   0x86,  0x00,	 0xc3,	0xc3,  0x00,  0x00}, // BOX DRAWINGS LIGHT VERTICAL AND RIGHT
	{0x2524,   0x87,  0x00,	 0xb4,	0xb4,  0x00,  0x00}, // BOX DRAWINGS LIGHT VERTICAL AND LEFT
	{0x252c,   0x88,  0x00,	 0xc2,	0xc2,  0x00,  0x00}, // BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
	{0x2534,   0x89,  0x00,	 0xc1,	0xc1,  0x00,  0x00}, // BOX DRAWINGS LIGHT UP AND HORIZONTAL
	{0x253c,   0x8a,  0x00,	 0xc5,	0xc5,  0x00,  0x00}, // BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
	{0x2550,   0xa0,  0x00,	 0xcd,	0xcd,  0x00,  0x00}, // BOX DRAWINGS DOUBLE HORIZONTAL
	{0x2551,   0xa1,  0x00,	 0xba,	0xba,  0x00,  0x00}, // BOX DRAWINGS DOUBLE VERTICAL
	{0x2552,   0xa2,  0x00,	 0xd5,	0xd5,  0x00,  0x00}, // BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
	{0x2553,   0xa4,  0x00,	 0xd6,	0xd6,  0x00,  0x00}, // BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
	{0x2554,   0xa5,  0x00,	 0xc9,	0xc9,  0x00,  0x00}, // BOX DRAWINGS DOUBLE DOWN AND RIGHT
	{0x2555,   0xa6,  0x00,	 0xb8,	0xb8,  0x00,  0x00}, // BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
	{0x2556,   0xa7,  0x00,	 0xb7,	0xb7,  0x00,  0x00}, // BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
	{0x2557,   0xa8,  0x00,	 0xbb,	0xbb,  0x00,  0x00}, // BOX DRAWINGS DOUBLE DOWN AND LEFT
	{0x2558,   0xa9,  0x00,	 0xd4,	0xd4,  0x00,  0x00}, // BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
	{0x2559,   0xaa,  0x00,	 0xd3,	0xd3,  0x00,  0x00}, // BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
	{0x255a,   0xab,  0x00,	 0xc8,	0xc8,  0x00,  0x00}, // BOX DRAWINGS DOUBLE UP AND RIGHT
	{0x255b,   0xac,  0x00,	 0xbe,	0xbe,  0x00,  0x00}, // BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
	{0x255c,   0xad,  0x00,	 0xbd,	0xbd,  0x00,  0x00}, // BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
	{0x255d,   0xae,  0x00,	 0xbc,	0xbc,  0x00,  0x00}, // BOX DRAWINGS DOUBLE UP AND LEFT
	{0x255e,   0xaf,  0x00,	 0xc6,	0xc6,  0x00,  0x00}, // BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
	{0x255f,   0xb0,  0x00,	 0xc7,	0xc7,  0x00,  0x00}, // BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
	{0x2560,   0xb1,  0x00,	 0xcc,	0xcc,  0x00,  0x00}, // BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
	{0x2561,   0xb2,  0x00,	 0xb5,	0xb5,  0x00,  0x00}, // BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
	{0x2562,   0xb4,  0x00,	 0xb6,	0xb6,  0x00,  0x00}, // BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
	{0x2563,   0xb5,  0x00,	 0xb9,	0xb9,  0x00,  0x00}, // BOX DRAWINGS DOUBLE VERTICAL AND LEFT
	{0x2564,   0xb6,  0x00,	 0xd1,	0xd1,  0x00,  0x00}, // BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
	{0x2565,   0xb7,  0x00,	 0xd2,	0xd2,  0x00,  0x00}, // BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
	{0x2566,   0xb8,  0x00,	 0xcb,	0xcb,  0x00,  0x00}, // BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
	{0x2567,   0xb9,  0x00,	 0xcf,	0xcf,  0x00,  0x00}, // BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
	{0x2568,   0xba,  0x00,	 0xd0,	0xd0,  0x00,  0x00}, // BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
	{0x2569,   0xbb,  0x00,	 0xca,	0xca,  0x00,  0x00}, // BOX DRAWINGS DOUBLE UP AND HORIZONTAL
	{0x256a,   0xbc,  0x00,	 0xd8,	0xd8,  0x00,  0x00}, // BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
	{0x256b,   0xbd,  0x00,	 0xd7,	0xd7,  0x00,  0x00}, // BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
	{0x256c,   0xbe,  0x00,	 0xce,	0xce,  0x00,  0x00}, // BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
	{0x2580,   0x8b,  0x00,	 0xdf,	0xdf,  0x00,  0x00}, // UPPER HALF BLOCK
	{0x2584,   0x8c,  0x00,	 0xdc,	0xdc,  0x00,  0x00}, // LOWER HALF BLOCK
	{0x2588,   0x8d,  0x00,	 0xdb,	0xdb,  0x00,  0x00}, // FULL BLOCK
	{0x258c,   0x8e,  0x00,	 0xdd,	0xdd,  0x00,  0x00}, // LEFT HALF BLOCK
	{0x2590,   0x8f,  0x00,	 0xde,	0xde,  0x00,  0x00}, // RIGHT HALF BLOCK
	{0x2591,   0x90,  0x00,	 0xb0,	0xb0,  0x00,  0x00}, // LIGHT SHADE
	{0x2592,   0x91,  0x00,	 0xb1,	0xb1,  0x00,  0x00}, // MEDIUM SHADE
	{0x2593,   0x92,  0x00,	 0xb2,	0xb2,  0x00,  0x00}, // DARK SHADE
	{0x25a0,   0x94,  0x00,	 0xfe,	0xfe,  0x00,  0x00}	 // BLACK SQUARE
};

unsigned char*
makeArray(int fromCode, int toCode)
{
	unsigned int i;
	unsigned char* array = (unsigned char*) malloc(256);
	if (!array) return NULL;

	for (i=0; i<256; i++) array[i] = i;
	if (fromCode==toCode || fromCode == UTF || toCode == UTF) return array;

	for (i = 0; i < sizeof(charsets) / sizeof(charsets[0]); i++){
		if (charsets[i][fromCode] && charsets[i][toCode])
			array[charsets[i][fromCode]] = charsets[i][toCode];
	}
	return array;
}

int
unicode2char(int unicode, int code)
{
	int start = 0;
	int midd = 0;
	int end = sizeof(charsets) / sizeof(charsets[0]);

	if (unicode < 0x80) return unicode;

	while (start < end) {
		int* u;
		midd = (start + end) >> 1;
		u = charsets[midd];
		if (unicode == *u) return u[code];
		if (unicode < *u) end = midd;
		else start = midd + 1;
	}
	return 0;
}

int
fromUTF8(unsigned char* s, int n, int code)
{
	int c = 0;
	char* r = s;
	for (; n; n--, s++) {
		int utf;
		if (!*s) {
			*r++ = 0;
			c++;
			continue;
		} else if (*s < 0x80) { // one byte
			utf = *s;
		} else if (*s < 0xC2) { // incorrect byte
			continue;
		} else if (*s < 0xE0) { // two bytes
			if (n < 2) continue;// shortage

			utf = (int)(*s++ & 0x1F) << 6; n--;

			if (*s < 0x80 || *s > 0xBF) continue; // incorrect 2nd byte
			utf |= *s & 0x3F;

		} else if (*s < 0xF0) { // three bytes
			if (n < 3) continue;// shortage

			utf = (int)(*s++ & 0xF) << 12; n--;

			if (*s < 0xA0 || *s > 0xBF) continue; // incorrect 2nd byte
			utf |= (int)(*s++ & 0x3F) << 6; n--;

			if (*s < 0x80 || *s > 0xBF) continue; // incorrect 3rd byte
			utf |= *s & 0x3F;
		} else {				// four bytes,
			continue; // unsupported
		}
		*r++ = unicode2char(utf, code) ? : '.';
		c++;
	}
	return c;
}


int
recode(char* s, int n, int fromCode, int toCode)
{
	unsigned char* table;
	int c = n;

	if (fromCode < 0 || fromCode >= N_OF_SETS) fromCode = DEFAULT_CS;
	if (toCode < 0 || toCode >= N_OF_SETS) toCode = DEFAULT_CS;
	if (fromCode == toCode) return c;

	if (toCode == UTF) return 0; // convert to utf-8 unsupported
	if (fromCode == UTF) return fromUTF8(s, n, toCode);

	table = tables[fromCode][toCode];
	if (!table) {
		table = tables[fromCode][toCode] = makeArray(fromCode, toCode);
		if (!table) return 0;
	}

	for (; n; n--, s++) *s = table[(unsigned char)(*s)];
	return c;
}

int
cs_by_string(char* cs)
{
	int i;

	if (!cs) return DEFAULT_CS;

	for (i = 0; i < N_OF_SETS; i++) {
		if (!strcmp(charset_string[i], cs)) return i;
	}
	return DEFAULT_CS;
}

char*
cs_by_id(int code)
{
	if (code < 0 || code >= N_OF_SETS) code = DEFAULT_CS;
	return charset_string[code];
}

