#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "settings.h"
#include "strings.h"
#include "misc.h"

#include "country.h"


static unsigned* ranges = NULL;
static unsigned char* rangesCodes = NULL;
static int nOfRanges = 0;

static int nOfCountries = -1;
static char* buffer = NULL;
static char** domains = NULL;
static char** countries = NULL;

static time_t mtime = 0;

char
loadIPRanges()
{
	char* name = "conf/ip2country";
	char buf[SHORT_STR];
	FILE* file;
	int i = 0;

	mtime = fileTime(name);

	if (!(file = fopen(name, "r"))) {
		message(SYS|ERROR, "cannot open %s", name);
		return FAILED;
	}
	if (!fgets(buf, SHORT_STR, file) ||
		(nOfRanges = atol(buf)) < 1) {
		message(SYS|ERROR, "cannot read %s", name);
		fclose(file);
		return FAILED;
	}

	FREE(ranges);
	FREE(rangesCodes);

	ranges = MALLOC(nOfRanges * sizeof(unsigned));
	rangesCodes = MALLOC(nOfRanges);

	while (i < nOfRanges && fgets(buf, SHORT_STR, file)) {
		unsigned ip = string2ip(buf);
		int code = strchr(buf, ' ') ? atol(strchr(buf, ' ')) : -1;
		if (ip != 0xFFFFFFFF && code != -1) {
			ranges[i] = ip;
			rangesCodes[i++] = code;
		}
	}
	fclose(file);
	if (i < nOfRanges) message(MESSAGE, "ranges mismatch (%d < %d)", i, nOfRanges);
	nOfRanges = i;
	return OK;
}

void
refreshRanges()
{
	if (mtime != fileTime("conf/ip2country")) loadIPRanges();
}

short
ip2country(unsigned ip)
{
	int start = 0;
	int midd = 0;
	int end = nOfRanges;
	int res, next;

	if (!ranges) {
		static char tried = 0;
		if (tried) return 0;
		tried = 1;
		if (!loadIPRanges()) return 0;
		end = nOfRanges;
	}

	while (start < end) {
		unsigned current;
		midd = (start + end) >> 1;
		current = ranges[midd];
		res = ip < current ? -1 : ip > current ? 1 : 0;
		next = midd + 1 >= nOfRanges || ip < ranges[midd+1] ? -1 : 1;
		if (res >= 0 && next < 0) return rangesCodes[midd];
		if (res < 0) end = midd;
		else start = midd + 1;
	}
	return 0;
}

void
loadCountries()
{
	long size;
	char* name = "conf/countries";
	FILE* cc;
	int i = 0;

	nOfCountries = 0;
	cc = fopen(name, "r");
	if (!cc) {
		message(SYS|ERROR, "cannot open file %s", name);
		buffer = MALLOC(1);
		return;
	}

	size = fileSize(name);
	buffer = (char*) MALLOC(size);
	domains = (char**) MALLOC(MAX_COUNTRIES * sizeof(char*));
	countries = (char**) MALLOC(MAX_COUNTRIES * sizeof(char*));

	fread(buffer, size, 1, cc);
	fclose(cc);

	while(i < size) {
		char* pos;
		int id = atol(buffer + i);
		if (id < 0 || id >= MAX_COUNTRIES) break;

		pos = strchr(buffer + i, ' '); if (!pos) break;
		i = pos - buffer;
		buffer[i++] = 0;
		while(buffer[i] == ' ') i++;
		domains[id] = buffer + i;

		pos = strchr(buffer + i, ' '); if (!pos) break;
		i = pos - buffer;
		buffer[i++] = 0;
		while(buffer[i] == ' ') i++;
		countries[id] = buffer + i;

		pos = strchr(buffer + i, '\n'); if (!pos) break;
		i = pos - buffer;
		buffer[i++] = 0;
		nOfCountries++;
	}
}

char*
id2country(short id)
{
	if (nOfCountries < 0) loadCountries();
	if (0 <= id && id < nOfCountries) return countries[id];
	if (nOfCountries > 0) return countries[0];
	return "";
}

char*
id2domain(short id)
{
	if (nOfCountries < 0) loadCountries();
	if (0 <= id && id < nOfCountries) return domains[id];
	if (nOfCountries > 0) return domains[0];
	return "";
}

short
domain2id(char* domain)
{
	int id;
	if (nOfCountries < 0) loadCountries();
	for (id = 0; id < nOfCountries; id++)
		if (!strcmp(domains[id], domain)) return id;
	return 0;
}
