/*	Copyright (C) 1999 Maxim Zotov

	Description: some general functions

	$Id: misc.c,v 1.12 2002/10/24 22:41:04 maxim Exp $
*/

#ifdef FCGI
#include <fcgi_stdio.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dirent.h>

#include "misc.h"
#include "settings.h"
#include "strings.h"
#include "ptrinfo.h"


#if DEBUGLOG
int Debug = 1;
#else
int Debug = 0;
#endif

char* skin;
char* progName = "undef";
char* logName[TYPE_MASK+1] = {"debug.log", "message.log", "error.log", "action.log", "access.log"};

int myPID;
time_t now;
struct tm now_tm;
short now_day;
char now_str[SHORT_STR];


#define N_OF_YEARS 10
static short days[N_OF_YEARS];

short
getDay(struct tm* tm)
{
	if (!days[1]) {
		int i, y;
		days[0] = 0;
		for (i = 0, y = 2000; i < N_OF_YEARS - 1; i++, y++)
			days[i + 1] = days[i] + (isleap(y) ? 366 : 365);
	}
	return 100 <= tm->tm_year && tm->tm_year <= 100 + N_OF_YEARS - 1 ?
		days[tm->tm_year - 100] + tm->tm_yday : 0;
}

struct tm*
getDate(short day)
{
	static struct tm tm;
	static short mdays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	int y = 0, m = 0;

	if (!days[1]) {
		int i, y;
		days[0] = 0;
		for (i = 0, y = 2000; i < N_OF_YEARS - 1; i++, y++)
			days[i + 1] = days[i] + (isleap(y) ? 366 : 365);
	}

	while (y < N_OF_YEARS - 1 && day >= days[y + 1]) y++;
	day -= days[y];
	y += 2000;

	memset(&tm, 0, sizeof(tm));
	tm.tm_yday = day;

	mdays[1] = isleap(y) ? 29 : 28;
	while (m < 11 && day >= mdays[m]) day -= mdays[m++];

	tm.tm_mday = day + 1;
	tm.tm_mon = m;
	tm.tm_year = y - 1900;

	return &tm;
}

time_t
getTimeByWeek(short year, short week)
{
	static struct tm tm;
	static short mdays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	int m = 0;
	int day = week * 7 - 1;

	mdays[1] = isleap(year) ? 29 : 28;
	while (m < 11 && day >= mdays[m]) day -= mdays[m++];
	if (day > 31) day = 31;

	memset(&tm, 0, sizeof(tm));
	tm.tm_mday = day + 1;
	tm.tm_mon = m;
	tm.tm_year = year - 1900;

	return mktime(&tm);
}

void
getTime()
{
	static time_t prev = 0;
	time(&now);
	if (now != prev) {
		now_tm = *(localtime(&now));
		now_day = getDay(&now_tm);
		strftime(now_str, SHORT_STR, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&now));
		prev = now;
	}
}

char*
time2str(time_t t, char type)
{
	static int call = 0;
	static char string[4][64];
	struct tm* ts = type < HTTP_TIME ? localtime(&t) : gmtime(&t);
	char* format =
		type == LOCAL_TIME	 ? "%d/%b/%Y:%H:%M:%S" :
		type == LOCAL_DATE	 ? "%d/%b/%Y" :
		type == TODAY_TIME	 ? "%H:%M:%S" :
		type == SQL_DATE	 ? "%Y-%m-%d" :
		type == SQL_TIME	 ? "%Y-%m-%d %H:%M:%S" :
		type == HTTP_TIME	 ? "%a, %d %b %Y %H:%M:%S GMT" :
		type == COOKIE_TIME	 ? "%a, %d-%b-%Y %H:%M:%S GMT" :
		"";
	char *s = string[++call % 4];
	strftime(s, 64, format, ts);
	return s;
}

unsigned
biRound(unsigned value, unsigned n)
{
	if (value >= 0x80000000) return 0x80000000;
	while (value > n) n <<= 1;
	return n;
}


short
countBits(int v)
{
	short bits;
	for(bits = 0; v; v<<=1) if (v < 0) bits++;
	return bits;
}


int
fileSize(char* filename)
{
	struct stat st;
	if (stat(filename, &st) != -1) {
		return st.st_size;
	}
//	message(SYS|ERROR, "cannot stat '%s'", filename);
	return -2048;
}

int
fdSize(int fd)
{
	struct stat st;
	if (fstat(fd, &st) != -1) {
		return st.st_size;
	}
//	message(SYS|ERROR, "cannot stat '%s'", filename);
	return -2048;
}

int
fileSizeKb(char* filename)
{
	return (fileSize(filename) + 512) / 1024;
}


time_t
fileTime(char* filename)
{
	struct stat st;
	if (stat(filename, &st) != -1) return st.st_mtime;
	message(SYS|ERROR, "cannot stat '%s'", filename);
	return 0;
}

char
exists(char* filename)
{
	struct stat st;
	return stat(filename, &st) == 0;
}

char
makedir(char* dir)
{
	char* slash = dir + strlen(dir) - 1;
	if (*slash == '/') *slash = 0;
	if (!exists(dir)) {
		char dir_[1024];
		char temp[1024];
		char* ptr;
		char* slash;
		strncpy(dir_, dir, 1023); dir_[1023] = 0;
		*temp = 0;
		ptr = dir_;
		do {
			(slash = strchr(ptr, '/')) && (*slash = 0);
			strcat(temp, ptr);
			if (*temp && !exists(temp)) {
				message(MESSAGE, "creating directory %s", temp);
				if (mkdir(temp, 0777)) {
					message(SYS|ERROR, "cannot make directory %s", temp);
					return FAILED;
				}
			}
			strcat(temp, "/");

			if (slash) ptr = slash + 1;
		} while (slash);
	}
	if (*slash == 0) *slash = '/';
	return OK;
}

char
writePidFile(char* name, char bg)
{
	int fd;
	char file[PATH_LEN];
	char* p;
	if (chdir(ROOT)) message(SYS|ERROR, "cannot chdir to %s", ROOT);

	catstr(file, PATH_LEN, PIDS, "/", name, NULL);

	if ((fd = open(file, O_CREAT|O_RDWR, 0644)) < 0)
		return message(SYS|ERROR, "cannot open %s", file);

	if (bg) {
		int r;
		int null;
		if (flock(fd, LOCK_EX|LOCK_NB) < 0) {
			int n = read(fd, file, PATH_LEN);
			if (n > 0) file[n] = 0;
			else message(SYS|ERROR, "cannot read %s", file);
			return message(ERROR, "another daemon is running [%s]", file);
		}

		if ((r = fork()) < 0) return message(SYS|ERROR, "cannot fork");
		if (r) exit(0);
		setsid();

		if ((null = open("/dev/null", O_RDWR)) < 0) {
			message(SYS|ERROR, "cannot open %s", "/dev/null");
		} else {
			dup2(null, STDIN_FILENO);
			dup2(null, STDOUT_FILENO);
			dup2(null, STDERR_FILENO);
			if (null > 2) close(null);
		}
	}

	ftruncate(fd, 0);
	p = itoa(myPID = getpid());
	write(fd, p, strlen(p));
	message(MESSAGE, "%s is starting to work", p);

	if (!bg && close(fd)) return message(SYS|ERROR, "cannot close %s", file);
	return OK;
}

char
message__(char* file, int pos, int flags, char *format, ...)
{
	static int* log = NULL;
	static char* string = NULL;
	static int logs_day = -1;
	va_list ap;
	int l;
	int err = errno;
	int type = flags & TYPE_MASK;


#if DEBUGLOG
	if (type == DEBUG && !Debug) return FAILED;
#endif

	if (!string && !(string = (char*) calloc(LONG_STR, sizeof(char))))
		return FAILED;
	if (!log) {
		if (!(log = calloc(TYPE_MASK + 1, sizeof(int)))) return FAILED;
		memset(log, -1, (TYPE_MASK + 1) * sizeof(int));
	}

	getTime();

	if (type == CLOSE_LOGS || logs_day != now_day) {
		int t;
		for (t = 0; t <= TYPE_MASK; t++) {
			if (log[t] >= 0 && log[t] != STDERR_FILENO) {
				close(log[t]);
				log[t] = -1;
			}
		}
		if (type == CLOSE_LOGS) return FAILED;
		logs_day = now_day;
	}

	if (log[type] < 0) {
		if (logName[type] && *logName[type] && log[type] < 0) {
			char str[PATH_LEN];
			char today[SHORT_STR];
			strftime(today, SHORT_STR, "%Y-%m-%d", &now_tm);
			l = catstr(str, PATH_LEN, "logs/", today, NULL);
			if (!exists(str) && mkdir(str, 0777) && errno != EEXIST) {
				logName[type] = NULL;
				log[type] = STDERR_FILENO;
				message(SYS|ERROR, "cannot create directory %s", str);
			} else {
				catstr(str + l, PATH_LEN - l, "/", logName[type], NULL);
				log[type] = open(str, O_CREAT|O_APPEND|O_WRONLY, 0644);
				if (log[type] < 0) {
					logName[type] = NULL;
					message(SYS|ERROR, "cannot open log file %s for appending", str);
					log[type] = STDERR_FILENO;
				} else {
                    char tmp[PATH_LEN];
					catstr(tmp, PATH_LEN, "logs/current.", itoa(getpid()), NULL);
					if (!symlink(today, tmp)) {
						if (rename(tmp, "logs/current")) message(SYS|ERROR, "cannot rename %s - %s", tmp, "logs/current");
					} else {
						message(SYS|ERROR, "cannot link %s -> %s", tmp, today);
					}
				}
			}
		} else {
			log[type] = STDERR_FILENO;
		}
	}

#if DEBUGLOG
	if (type == DEBUG) {
		struct timeval tv;
		gettimeofday(&tv, NULL);
		//if (now != tv.tv_sec) getTime();
		l = snprintf(string, LONG_STR,
					 "%02d:%02d:%02d.%03d|%s|%s:%d| ",
					 now_tm.tm_hour, now_tm.tm_min, now_tm.tm_sec, (int)(tv.tv_usec/1000),
					 progName, file, pos);
	} else
#endif
		l = snprintf(string, LONG_STR, "%02d:%02d:%02d|%s| ",
					 now_tm.tm_hour, now_tm.tm_min, now_tm.tm_sec,
					 progName);

	va_start(ap, format);
	l += vsnprintf(string + l, LONG_STR - l, format, ap);
	va_end(ap);

	if (l < LONG_STR) {
		if ((flags & SYS)) {
			l += snprintf(string + l, LONG_STR - l,
					": %s\n", strerror(err));
			if (l >= LONG_STR) {
				string[LONG_STR - 1] = '\n';
				l = LONG_STR;
			}
		} else {
			string[l++] = '\n';
		}
	} else {
		string[LONG_STR - 1] = '\n';
		l = LONG_STR;
	}


	write(log[type], string, l);
	errno = err;

	return FAILED;
}

int
addCount(char* file, char doFlush)
{
	static FILE* log = NULL;
	static int count = 0;
	static int intercount = 0;
	static int error = 0;

	if (!doFlush) {
		count++;
		if ((++intercount) < 1000) return count;
	}

	if (!log && !error) {
		char name[PATH_LEN];
		snprintf(name, PATH_LEN, "%s/%s-count", LOGS, file);
		if (!(log = fopen(name, "r+")) &&
			!(log = fopen(name, "w+"))) {
			message(SYS|ERROR, "cannot open file %s", name);
			error = 1;
		}
	}
	if (!error) {
		int c = 0;
		char s[32];
		fseek(log, 0, SEEK_SET);
		fgets(s, 32, log);
		if (*s) sscanf(s, "%d", &c);
		fseek(log, 0, SEEK_SET);
		if (c>=1000000000) c -= 1000000000;
		c += intercount;
		fprintf(log, "%-10d\n", c);
		fflush(log);
	}
	if (doFlush) {
		fclose(log);
		log = NULL;
	}
	intercount = 0;
	return count;
}

void*
malloc_wrapper(size_t size, char* file, int line)
{
	void* ptr = calloc(size, 1);
	debug("allocating %d bytes on %s:%d", size, file, line);
	if (!ptr) message(ERROR, "cannot allocate %d bytes on %s:%d", size, file, line);
#if DEBUGLOG
	else addPtr(ptr, size, file, line);
#endif
	return ptr;
}

void*
realloc_wrapper(void* ptr, size_t size, char* file, int line)
{
#if DEBUGLOG
	if (ptr) delPtr(ptr, file, line);
#endif
	ptr = realloc(ptr, size);
	debug("reallocating %d bytes on %s:%d", size, file, line);
	if (!ptr) message(ERROR, "cannot reallocate %d bytes on %s:%d", size, file, line);
#if DEBUGLOG
	else addPtr(ptr, size, file, line);
#endif
	return ptr;
}

void*
free_wrapper(void* ptr, char* file, int line)
{
	if (ptr) {
#if DEBUGLOG
		delPtr(ptr, file, line);
#endif
		free(ptr);
	}
	debug("free on %s:%d", file, line);
	return NULL;
}

char* lastPositionFile = NULL;
int lastPositionLine = 0;

void
fix_position(char* file, int line)
{
	lastPositionFile = file;
	lastPositionLine = line;
}


char*
readFile(char* fileName, int* size)
{
	char* buf;
	FILE* f;
	int fs = fileSize(fileName);
	int readed;
	if (size) *size = fs;
	if (fs < 0) return NULL;
	if (!(f = fopen(fileName, "r"))) {
		message(SYS|ERROR, "cannot open %s", fileName);
		return NULL;
	}
	buf = MALLOC(fs + 1);
	buf[fs] = 0;
	if ((readed = fread(buf, 1, fs, f)) < fs)
		message(ERROR, "short read %s (%d < %d)", fileName, readed, fs);
	if (fclose(f))
		message(SYS|ERROR, "cannot close %s", fileName);
	return buf;
}

char
removeDir(char* name)
{
	DIR* dir;
	struct dirent* ent;
	char result = OK;

	if (!(dir = opendir(name))) return message(SYS|ERROR, "cannot opendir %s", name);
	readdir(dir);
	readdir(dir);
	while((ent = readdir(dir))) {
		char filename[PATH_LEN];
		struct stat s;

		snprintf(filename, PATH_LEN, "%s/%s", name, ent->d_name);
		if (lstat(filename, &s)) result = message(SYS|ERROR, "cannot stat %s", filename);
		else if (S_ISDIR(s.st_mode)) result &= removeDir(filename);
		else if (unlink(filename)) result = message(SYS|ERROR, "cannot unlink %s", filename);
	}
	closedir(dir);
	if (rmdir(name)) result = message(SYS|ERROR, "cannot rmdir %s", name);

	return result;
}


int
initServerSocket(int domain, char* address, int port)
{
	int fd = -1;

	do {
		struct sockaddr_in addrI;
		struct sockaddr_un addrU;
		struct sockaddr* addr;
		int len;
		int onoff = 1;


		if ((fd = socket(domain, SOCK_STREAM, 0)) <= 0) {
			message(SYS|ERROR, "cannot create socket");
			break;
		}

		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &onoff, sizeof(onoff));
		fcntl(fd, F_SETFL, O_NONBLOCK);

		if (domain == AF_INET) {
			memset(&addrI, 0, sizeof(addrI));
			addrI.sin_family = AF_INET;
			if (address && !inet_aton(address, &addrI.sin_addr)) {
				message(ERROR, "bad address: %s", address);
				break;
			} else {
				addrI.sin_addr.s_addr = INADDR_ANY;
			}
			addrI.sin_port = htons(port);
			addr = (struct sockaddr*)& addrI;
			len = sizeof(addrI);
		} else {
			memset(&addrU, 0, sizeof(addrU));
			addrU.sun_family = AF_UNIX;
			strcpy(addrU.sun_path, "data/sockets/");
			strcat(addrU.sun_path, address);
			unlink(addrU.sun_path);
			addr = (struct sockaddr*)&addrU;
			len = sizeof(addrU.sun_family) + strlen(addrU.sun_path);
		}

		if (bind(fd, addr, len) < 0) {
			message(SYS|ERROR, "cannot bind on %s:%d", address, port);
			break;
		}

		if (listen(fd, 10) < 0) {
			message(SYS|ERROR, "cannot listen for %s:%d", address, port);
			break;
		}

		return fd;
	} while (FALSE);

	if (fd >= 0) close(fd);
	return -1;
}


int
doConnect(int domain, char* address, int port, int timeout)
{
	int fd = -1;

	do {
		struct sockaddr_in addrI;
		struct sockaddr_un addrU;
		struct sockaddr* addr;
		int len;

		if ((fd = socket(domain, SOCK_STREAM, 0)) < 0) {
			message(SYS|ERROR, "cannot create socket");
			break;
		}

		fcntl(fd, F_SETFL, O_NONBLOCK);

		if (domain == AF_INET) {
			memset(&addrI, 0, sizeof(addrI));
			if (!inet_aton(address, &addrI.sin_addr)) {
				message(ERROR, "bad address: %s", address);
				break;
			}
			addrI.sin_port = htons(port);
			addrI.sin_family = AF_INET;
			addr = (struct sockaddr*)& addrI;
			len = sizeof(addrI);
		} else {
			addrU.sun_family = AF_UNIX;
			strcpy(addrU.sun_path, "data/sockets/");
			strcat(addrU.sun_path, address);

			addr = (struct sockaddr*)& addrU;
			len = sizeof(addrU.sun_family) + strlen(addrU.sun_path);
		}

		if (connect(fd, addr, len) < 0) {
			struct pollfd fds;
			int r;

			if (errno != EINPROGRESS) {
				message(SYS|ERROR, "cannot connect");
				break;
			}

			fds.events = POLLOUT;
			fds.fd = fd;
			r = poll(&fds, 1, timeout * 1000);
			if (r > 0) {
				int err = 0;
				int l = sizeof(err);
				if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &l) < 0 || err) {
					if (err) errno = err;
					message(SYS|ERROR, "cannot connect to %s:%d", address, port);
					break;
				}
			} else {
				if (r < 0) message(SYS|ERROR, "cannot poll");
				else message(ERROR, "connect timeout (%s:%d)", address, port);
				break;
			}
		}

		return fd;
	} while(FALSE);

	if (fd >= 0) close(fd);
	return -1;

}

int
readMessage(int fd, char* buffer, int maxSize, int timeout)
{
	int size = -1;
	int received = 0;
	time_t last = now + timeout;
	struct timeval start;
	struct timeval finish;

	gettimeofday(&start, NULL);

	do {
		int r, n;
		struct pollfd fds;
		fds.events = POLLIN;
		fds.fd = fd;
		r = poll(&fds, 1, 1000);
		getTime();
		switch(r) {
		case -1:
			if (errno != EINTR) {
				message(SYS|ERROR, "cannot poll");
				return -1;
			}
		case 0:
			if (now >= last) {
				gettimeofday(&finish, NULL);
				message(ERROR, "readMessage timeout (%ld ms)",
						(finish.tv_sec - start.tv_sec)*1000 + (finish.tv_usec/1000 - start.tv_usec/1000));
				return -1;
			}
			break;
		default:
			n = read(fd, buffer + received, maxSize - received);
			debug("read: %d bytes", n);
			if (n <= 0) {
				if (n < 0 && errno == EWOULDBLOCK) break;
				message((n<0?SYS:0)|ERROR, "cannot read");
				return -1;
			}
			received += n;

			if (size < 0 && received >= sizeof(int))
				size = *((int*)buffer);

			if (received == size) {
//				gettimeofday(&finish, NULL);
//				message(MESSAGE, "readMessage: %ld ms",
//						(finish.tv_sec - start.tv_sec)*1000 + (finish.tv_usec/1000 - start.tv_usec/1000));
				return received;
			}

			if (received == maxSize) {
				message(ERROR, "too big message (%d > %d)", size, maxSize);
				return -1;
			}
			if (now >= last) {
				gettimeofday(&finish, NULL);
				message(ERROR, "readMessage timeout (%d/%d bytes, %ld ms)",
						received, size,
						(finish.tv_sec - start.tv_sec)*1000 + (finish.tv_usec/1000 - start.tv_usec/1000));
				return -1;
			}
		}
	} while (TRUE);
	return -1;
}

int
writeMessage(int fd, char* buffer, int size, int timeout)
{
	int sent = 0;
	time_t last = now + timeout;

	*((int*)buffer) = size;

	do {
		int r, n;
		struct pollfd fds;

		n = write(fd, buffer + sent, size - sent);
		debug("%d> written %d", n);
		if (n > 0) {
			sent += n;
			if (sent == size) return sent;

			if (now >= last) {
				message(ERROR, "writeMessage timeout");
				return -1;
			}
		} else if (n == 0 || errno != EWOULDBLOCK) {
			message((n<0?SYS:0)|ERROR, "cannot write");
			return -1;
		}


		fds.events = POLLOUT;
		fds.fd = fd;
		r = poll(&fds, 1, 1000);
		getTime();
		switch(r) {
		case -1:
			if (errno != EINTR) {
				message(SYS|ERROR, "cannot poll");
				return -1;
			}
		case 0:
			if (now >= last) {
				message(ERROR, "writeMessage timeout");
				return -1;
			}
		}
	} while (TRUE);
	return -1;
}


int
snprintf__(char* file, int pos, char* string, int length, char* fmt, ...)
{
	va_list ap;
	int n;
	if (length <= 0) {
		message(ERROR, "no buffer for snprintf on %s:%d \"%s\"", file, pos, fmt);
		return 0;
	}
	va_start(ap, fmt);
	n = vsnprintf(string, length, fmt, ap);
	va_end(ap);

	if (n < 0) {
		message(ERROR, "snprintf negative on %s:%d \"%s\"", file, pos, fmt);
		return 0;
	} else if (n >= length) {
		message(ERROR, "snprintf overflow on %s:%d \"%s\": %d >= %d", file, pos, fmt, n, length);
		return length - 1;
	} else {
		return n;
	}
}

