/*	Copyright (C) 1999 Maxim Zotov

Description: some strings functions

$Id: strings.h,v 1.15 2002/11/01 22:56:21 maxim Exp $
*/

#ifndef STRINGS_H
#define STRINGS_H

#include "list.h"

typedef struct String_ {
	unsigned next;
	char length;
	char string[0];
} String;

typedef struct StringBlock_ {
	unsigned crc;
	unsigned length;
	char data[0];
} StringBlock;

typedef struct Strings_ {
	unsigned hashSize;
	unsigned nOfBlocks;
	unsigned count;
	unsigned readedBlock;
	unsigned changed;

	unsigned* hash;
	StringBlock** blocks;
	StringBlock* block;

	void* map;
	int mapFD;
	unsigned mapSize;
} Strings;

Strings* newStrings(unsigned hashSize);
unsigned string2id(Strings* s, char* str);
char* id2string(Strings* s, unsigned id);
char writeStrings(Strings* s, char* file, char force);
Strings* readStrings(char* file, unsigned size);
void showStrings(Strings* s);
Strings* freeStrings(Strings* s);
Strings* mapStrings(char* file);


typedef struct StringPool_struct {
	int size;
	char* store;

	int nOfParts;
	char** parts;

	char** strings;
	int count;
} StringPool;


char* addStringToPool(StringPool* pool, char* string);
StringPool* freeStringPool(StringPool* pool);

int hexDigit(char c);
char* unhash(char* str);
char* unescape(char* string, char plusIsSpace, char isUTF);
char* escape(char* string, char plusIsSpace);
char* escapeLink(char* string);

char* quote(char* string);
char* quotes(char* string);

void html_escape(List** list, ...);
void sql_escape(List** list, ...);
void sql_unescape(char* string, ...);
void remove_cr(char* string, ...);
char* chomp(char* str);
char* trimSpaces(char* string);
char* strtolower(unsigned char* s);
char tolower(int c);

char* cat_strings(char* string, ...);
char* add_strings(char** string, ...);
char* add_string(char** string, char* add);

int catstr(char* string, int max_size, ...);

unsigned hash(char *s);

char isalnum(char c);
char isspace(char c);
char isdigit(char c);
char ischar(unsigned char c);
char issafe(char c);

char* itoa(int n);
char* itos(int n);

int string2ip(char* string);
char* ip2string(int ip);

int reverse(char* s);

extern char toLower[256];
void initToLower();
char* copystring(char* string);

int split(char* string, char* del, int max_len, ...);

char* fmt(char* format, ...);


char* long2str(unsigned u);
char* longs2str(unsigned u1, unsigned u2);
long str2long(char* str);

char checkEmail(char* email);
char checkPassword(char* password);
char checkNick(char* nick);

int stripLast(char* url);

char* humanSize(long long size);
char* humanPeriod(int period);

char* strncpy_(char* dest, char* src, int maxLength);



#define PUT(object) {\
	memcpy(ptr, &(object), sizeof(object)); \
	ptr += sizeof(object);\
	}

#define PUT_STR(string) {\
	short length = string ? strlen(string) + 1 : 0;\
	if (length > LONG_STR || length < 0) {string[LONG_STR - 1] = 0; length = LONG_STR;}\
	memcpy(ptr, &length, sizeof(length));\
	ptr += sizeof(length);\
	if (length > 0) memcpy(ptr, string, length);\
	ptr += length;\
	}

#define PUT_VALUES(values) {\
	if (values && values[0]) {\
	short count = 1; while(values[count]) count++;\
	count *= sizeof(*values);\
	memcpy(ptr, &count, sizeof(count));\
	ptr += sizeof(count);\
	memcpy(ptr, values, count)\
	ptr += count;\
	} else {memset(ptr, (short)0, sizeof(short));\
	ptr += sizeof(short);}\
	}

#define PUT_ARRAY(array, cnt) {\
	if (array && cnt > 0) {\
	short n = cnt * sizeof(*(array));\
	*((short*)ptr) = n;\
	ptr += sizeof(n);\
	memcpy(ptr, array, n);\
	ptr += n;\
	} else {memset(ptr, (short)0, sizeof(short));\
	ptr += sizeof(short);}\
	}


#define GET(object) {\
	memcpy(&object, ptr, sizeof(object)); \
	ptr += sizeof(object);\
	}

#define GET_STR(string) {\
	short* length = (short*) ptr;\
	ptr += sizeof(*length);\
	if (*length){string = ptr; ptr += *length;}\
	else string = NULL;\
	}

#define GET_VALUES(values) {\
	short count; memcpy(&count, ptr, sizeof(count));\
	ptr += sizeof(count);\
	if (count) {values = ptr; ptr += count;}\
	else values = NULL;\
	}

#define GET_ARRAY(array, cnt) {\
	short count; memcpy(&count, ptr, sizeof(count));\
	ptr += sizeof(count);\
	if (count > 0) {memcpy(array, ptr,\
	count > cnt * sizeof(*array) ? cnt * sizeof(*array) : count);\
	ptr += count;}\
	}

#endif

