/* $Id: config.c,v 1.4 2002/03/26 20:52:51 maxim Exp $
 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "common/misc.h"
#include "common/settings.h"
#include "common/strings.h"
#include "common/country.h"
//#include "common/charsets.h"

#include "config.h"
#include "core.h"
#include "httpd.h"
#include "access.h"
#include "filters.h"


#define CONFIG_NAME_DEF "httpd.conf"
#define CONNECTIONS_DEF 1000
#define PROXY_CONNECTIONS_DEF 100
#define MAX_BUFFER (128*1024)
#define MAX_POST_SIZE (256*1024)
#define KEEP_ALIVE_TIMEOUT_DEF 15
#define TIMEOUT_DEF 60
#define REQUEST_TIMEOUT_DEF 10
#define EXPIRES_DEF 0
#define KEEP_ALIVE_DEF ON
#define CONNECTIONS_PER_IP 50
#define LOG_DEF ON
#define LISTING_DEF OFF
#define THROTTLE_DEF OFF
#define DETACH_ROOT_DEF ON
#define SERVER_TOKENS_DEF ON
#define SKIP_WWW_DEF OFF
#define SKIP_PORT_DEF ON
#define FILES_CHECK_DEF 30
#define FILES_KEEP_DEF 10
#define CHARSET_DEF -1
#define SRC_CHARSET_DEF -1


#define MAX_TOKENS 64

Config* config;

enum {
	CONFIG = 1,
	SERVER = 2,
	ALIAS = 4
};


typedef int (*Setter)(char** tokens, int count);

typedef struct Option_struct {
	char* name;
	Setter setter;
	int context;
} Option;

int setConnections(char** tokens, int count);
int setProxyConnections(char** tokens, int count);
int setMaxBuffer(char** tokens, int count);
int setSendBuffer(char** tokens, int count);
int setLowWatermark(char** tokens, int count);
int setMaxPostSize(char** tokens, int count);
int setTimeout(char** tokens, int count);
int setRequestTimeout(char** tokens, int count);
int setExpires(char** tokens, int count);
int setKeepAliveTimeout(char** tokens, int count);
int setConnectionsPerIP(char** tokens, int count);
int setListen(char** tokens, int count);
int setMime(char** tokens, int count);
int setAdmin(char** tokens, int count);
int setSkipWWW(char** tokens, int count);
int setSkipPort(char** tokens, int count);
int setFilesCheck(char** tokens, int count);
int setFilesKeep(char** tokens, int count);
int setIndex(char** tokens, int count);
int setReverse(char** tokens, int count);
int setListing(char** tokens, int count);
int setServerTokens(char** tokens, int count);
int setCharset(char** tokens, int count);
int setSrcCharset(char** tokens, int count);
int setServer(char** tokens, int count);
int setAlias(char** tokens, int count);
int setCache(char** tokens, int count);
int setAccess(char** tokens, int count);
int setKeepAlive(char** tokens, int count);
int setAccess(char** tokens, int count);
int setFilter(char** tokens, int count);
int setLog(char** tokens, int count);
int setThrottle(char** tokens, int count);
int setDetachRoot(char** tokens, int count);

Server* currentServer;
Alias* currentAlias;

int context = 0;

Option options[] = {
	{"ProxyConnections",	setProxyConnections,	CONFIG},
	{"KeepAliveTimeout",	setKeepAliveTimeout,	CONFIG},
	{"ConnectionsPerIP",	setConnectionsPerIP,	CONFIG},
	{"Connections", setConnections, CONFIG},
	{"MaxBuffer",	setMaxBuffer,	CONFIG},
	{"MaxPostSize", setMaxPostSize, CONFIG},
	{"SendBuffer",  setSendBuffer,  CONFIG},
	{"LowWatermark",setLowWatermark,CONFIG},
	{"Listen",		setListen,		CONFIG},
	{"MIME",		setMime,		CONFIG},
	{"SkipWWW",		setSkipWWW,		CONFIG},
	{"FilesCheck",	setFilesCheck,	CONFIG},
	{"FilesKeep",	setFilesKeep,	CONFIG},
	{"SkipPort",	setSkipPort,	CONFIG},
	{"Admin",		setAdmin,		CONFIG|SERVER},
	{"Log",			setLog,			CONFIG|SERVER|ALIAS},
	{"Timeout",		setTimeout,		CONFIG|SERVER|ALIAS},
	{"RequestTimeout",setRequestTimeout, CONFIG},
	{"Expires",		setExpires,		CONFIG|SERVER|ALIAS},
	{"Charset",		setCharset,		CONFIG|SERVER|ALIAS},
	{"SrcCharset",	setSrcCharset,	CONFIG|SERVER|ALIAS},
	{"Index",		setIndex,		CONFIG|SERVER|ALIAS},
	{"Reverse",		setReverse,		CONFIG|SERVER|ALIAS},
	{"Listing",		setListing,		CONFIG|SERVER|ALIAS},
	{"KeepAlive",	setKeepAlive,	CONFIG|SERVER|ALIAS},
	{"Server",		setServer,		CONFIG|SERVER|ALIAS},
	{"Alias",		setAlias,		SERVER|ALIAS},
	{"Cache",		setCache,		SERVER|ALIAS},
	{"Access",		setAccess,		CONFIG|SERVER|ALIAS},
	{"Filter",		setFilter,		CONFIG|SERVER|ALIAS},
	{"Throttle",	setThrottle,	CONFIG|SERVER|ALIAS},
	{"DetachRoot",	setDetachRoot,	CONFIG},
	{"ServerTokens",setServerTokens,CONFIG}
};


int
setDefaults()
{
	if (!(config = MALLOC(sizeof(Config)))) return FAILED;
	config->configName = strdup(CONFIG_NAME_DEF);
	config->connections = CONNECTIONS_DEF;
	config->proxyConnections = PROXY_CONNECTIONS_DEF;
	config->connectionsPerIP = CONNECTIONS_PER_IP;
	config->keepAliveTimeout = KEEP_ALIVE_TIMEOUT_DEF;
	config->maxBuffer = MAX_BUFFER;
	config->maxPostSize = MAX_POST_SIZE;
	config->timeout = TIMEOUT_DEF;
	config->requestTimeout = REQUEST_TIMEOUT_DEF;
	config->expires = EXPIRES_DEF;
	config->keepalive = KEEP_ALIVE_DEF;
	config->log = LOG_DEF;
	config->listing = LISTING_DEF;
	config->throttle = THROTTLE_DEF;
	config->detachRoot = DETACH_ROOT_DEF;
	config->serverTokens = SERVER_TOKENS_DEF;
	config->skipWWW = SKIP_WWW_DEF;
	config->skipPort = SKIP_PORT_DEF;
	config->filesCheck = FILES_CHECK_DEF;
	config->filesKeep = FILES_KEEP_DEF;
	//	config->charset = CHARSET_DEF;
	config->srcCharset = SRC_CHARSET_DEF;

	config->nOfListeners = 0;
	return OK;
}

void
freeSimpleList(SimpleList* list)
{
	while(list) {
		SimpleList* next = list->next;
		FREE(list->name);
		FREE(list);
		list = next;
	}
}

void
freeCache(Cache* cache)
{
    FREE(cache);
}

void
freeAccess(Access* access, Access* parent)
{
	while(access && access != parent) {
		Access* next = access->next;
		if (access->type == PASSWD) {
			freePairs(access->users);
			FREE(access->file);
			FREE(access->passwd);
			FREE(access->realm);
		} else if (access->type == REFERRER) {
			FREE(access->referrer);
		} else if (access->type == AGENT) {
			FREE(access->agent);
		}
		FREE(access->redirect);
		FREE(access);

		access = next;
	}
}

void
freeAlias(Alias* alias, Server* server)
{
    int i;
	if (alias->cache != server->cache) freeCache(alias->cache);
	if (alias->proxy) {
		Proxy* start = alias->proxy;
		Proxy* proxy = alias->proxy;
		do {
			Proxy* next = proxy->next;
			FREE(proxy->location);
			FREE(proxy);
			proxy = next;
		} while (proxy != start);
	}
    for (i = 0; i < 8; i++) FREE(alias->aux[i]);
	freeAccess(alias->access, server->access);
	FREE(alias->tail);
	FREE(alias->tailLength);
	FREE(alias->ctype);
	FREE(alias->ctypeLength);
	FREE(alias->alias);
	FREE(alias->location);
	FREE(alias);
}


void
freeServer(Server* server, Config* config)
{
	int i;
	Alias* alias = server->aliases;
	if (server->admin != config->admin) FREE(server->admin);
	if (server->indexes != config->indexes) freeSimpleList(server->indexes);
	if (server->reverse != config->reverse) freeSimpleList(server->reverse);
	while(alias) {
		Alias* next = alias->next;
		freeAlias(alias, server);
		alias = next;
	}
	freeCache(server->cache);
	freeAccess(server->access, config->access);
	for (i = 0; i < server->nOfAlters; i++)
		FREE(server->alter[i]);
	FREE(server->alterHash);
	FREE(server->alter);
	FREE(server->name);
	FREE(server);
}

Config*
freeConfig(Config* config)
{
	Listener* listener;
	Server* server;
    Backend* backend;

	if (!config) return NULL;
	debug("freeConfig(%p)", config);

	FREE(config->configName);
	FREE(config->admin);

	freeSimpleList(config->extensions);
	freeSimpleList(config->mimes);
	freeSimpleList(config->indexes);
	freeSimpleList(config->reverse);

	listener = config->listeners;
	while(listener) {
		Listener* next = listener->next;
		FREE(listener->error);
		FREE(listener);
		listener = next;
	}

	server = config->servers;
	while(server) {
		Server* next = server->next;
		freeServer(server, config);
		server = next;
	}
	freeAccess(config->access, NULL);

	backend = config->backends;
	while(backend) {
		Backend* next = backend->next;
		free(backend->name);
		FREE(backend);
		backend = next;
	}

	return FREE(config);
}

int
getTokens(char* str, char** tokens, int max_tokens)
{
	int cnt = 0, i;
	char* token;
	token = strtok(str, " \t");
	if (!token || *token == '#') return 0;
	while(cnt < max_tokens && (tokens[cnt] = token)) {
		token = strtok(NULL, " \t");
		cnt++;
	}
	for (i = 0; i < cnt - 1; i++) {
		if (tokens[i][strlen(tokens[i]) - 1] == '\\') {
			tokens[i][strlen(tokens[i]) - 1] = ' ';
			memmove(tokens[i] + strlen(tokens[i]),
					tokens[i + 1],
					strlen(tokens[i + 1]) + 1);

			memmove(tokens + i + 1, tokens + i + 2, (sizeof(char*)) * (cnt - i - 2));
			i--;
			cnt--;
		}
	}
	return cnt;
}


int
setOption(char** tokens, int count, char* name, int line)
{
	size_t i;
	for (i = 0; i < sizeof(options)/sizeof(options[0]); i++) {
		if (strcasecmp(tokens[0], options[i].name) == 0) {
			if (count < 2) return message(ERROR, "undefined value for option `%s' at %s:%d", tokens[0], name, line);
			if ((context & options[i].context))
				return options[i].setter(tokens + 1, count - 1);
			else
				return message(ERROR, "wrong context for `%s' at %s:%d", tokens[0], name, line);
		}
	}
	return message(ERROR, "unknown option `%s'", tokens[0]);
}


int
sizeValue(char* token)
{
    char* e;
	int n = strtol(token, &e, 10);
    int mult = 1;

	switch(*e) {
	case 0:
	case 'k':
	case 'K': mult = 1024; break;
	case 'M': mult = 1024*1024; break;
	default:  message(ERROR, "bad size multiplier: %s, treating as bytes", token);
	}

	return n * mult;
}

int
booleanValue(char* token)
{
	return
		!strcasecmp(token, "on") ? ON :
		!strcasecmp(token, "yes") ? ON :
		!strcasecmp(token, "true") ? ON :
		!strcasecmp(token, "off") ? OFF :
		!strcasecmp(token, "no") ? OFF :
		!strcasecmp(token, "false") ? OFF :
		message(ERROR, "bad boolean value: %s, treating as OFF", token);;
}

int
timeValue(char* token)
{
    char* e;
	int n = strtol(token, &e, 10);
	int mult = 1;

	switch(*e) {
	case 0:
	case 's': mult = 1; break;
	case 'm': mult = 60; break;
	case 'h': mult = 60*60; break;
	case 'd': mult = 24*60*60; break;
	case 'W': mult = 7*24*60*60; break;
	case 'M': mult = 30*24*60*60; break;
	case 'Y': mult = 365*24*60*60; break;
	default:  message(ERROR, "bad time multiplier: %s, treating as seconds", token);
	}

	return n * mult;
}


int
setConnections(char** tokens, int count)
{
	config->connections = atol(tokens[0]);
	if (config->connections <= 0 || config->connections > 8192)
		return message(ERROR, "bad connections value: %s", tokens[0]);
	return OK;
}

int
setProxyConnections(char** tokens, int count)
{
	config->proxyConnections = atol(tokens[0]);
	if (config->proxyConnections <= 0 || config->proxyConnections > 1024)
		return message(ERROR, "bad connections value: %s", tokens[0]);
	return OK;
}

int
setConnectionsPerIP(char** tokens, int count)
{
	config->connectionsPerIP = atol(tokens[0]);
	if (config->connectionsPerIP <= 0)
		return message(ERROR, "bad ConnectionsPerIP value: %s", tokens[0]);
	return OK;
}

int
setKeepAliveTimeout(char** tokens, int count)
{
	config->keepAliveTimeout = timeValue(tokens[0]);
	if (config->keepAliveTimeout <= 0 || config->keepAliveTimeout > 10*60)
		return message(ERROR, "bad keepAliveTimeout value: %s", tokens[0]);
	return OK;
}

int
setMaxBuffer(char** tokens, int count)
{
	config->maxBuffer = atol(tokens[0]);
	if (config->maxBuffer <= 0 || config->maxBuffer > 1024)
		return message(ERROR, "bad maxBuffer value: %s", tokens[0]);
	config->maxBuffer *= 1024;
	return OK;
}

int
setSendBuffer(char** tokens, int count)
{
	config->sendBuffer = atol(tokens[0]);
	if (config->sendBuffer <= 0 || config->sendBuffer > 1024)
		return message(ERROR, "bad sendBuffer value: %s", tokens[0]);
	config->sendBuffer *= 1024;
	return OK;
}

int
setLowWatermark(char** tokens, int count)
{
	config->lowWatermark = atol(tokens[0]);
	if (config->lowWatermark <= 0 || config->lowWatermark > 1024)
		return message(ERROR, "bad lowWatermark value: %s", tokens[0]);
	config->lowWatermark *= 1024;
	return OK;
}

int
setMaxPostSize(char** tokens, int count)
{
	config->maxPostSize = atol(tokens[0]);
	if (config->maxPostSize <= 0 || config->maxPostSize > 16384)
		return message(ERROR, "bad maxPostSize value: %s", tokens[0]);
	config->maxPostSize *= 1024;
	return OK;
}

int
setTimeout(char** tokens, int count)
{
	int v = timeValue(tokens[0]);
	if (v <= 0 || v > 10*60)
		return message(ERROR, "bad timeout value: %s", tokens[0]);
	if (context == CONFIG) config->timeout = v;
	else if (context == SERVER) currentServer->timeout = v;
	else currentAlias->timeout = v;
	return OK;
}

int
setRequestTimeout(char** tokens, int count)
{
	int v = timeValue(tokens[0]);
	if (v <= 0 || v > 10*60)
		return message(ERROR, "bad timeout value: %s", tokens[0]);
	config->requestTimeout = v;
	return OK;
}

int
setExpires(char** tokens, int count)
{
	int v;
	if (!strcasecmp(tokens[0], "past"))
		v = EXPIRES_PAST;
    else if (!strcasecmp(tokens[0], "future"))
		v = EXPIRES_FUTURE;
	else {
        v = timeValue(tokens[0]);
		if (!v || v < -10*365*24*60*60 || v > 10*365*24*60*60)
			return message(ERROR, "bad expires value: %s", tokens[0]);
	}
	if (context == CONFIG) config->expires = v;
	else if (context == SERVER) currentServer->expires = v;
	else currentAlias->expires = v;
	return OK;
}


int
setListen(char** tokens, int count)
{
	int p = 0;
	Listener* listener = MALLOC(sizeof(Listener));
	if (!listener) return FAILED;

	if (count > 1) {
		if (inet_aton(tokens[0], &(listener->addr)) == 0)
			return message(ERROR, "bad address value: %s", tokens[0]);
		p = 1;
	}

	listener->port = atol(tokens[p]);
	if (listener->port <= 0 || listener->port > 65535)
		return message(ERROR, "bad port value: %s", tokens[p]);

	listener->type = LISTENER;
	listener->fd = -1;
	listener->next = config->listeners;
	config->listeners = listener;
	config->nOfListeners++;
	return OK;
}

int
setMime(char** tokens, int count)
{
	SimpleList* mime;
	int i;
	if (!strchr(tokens[0], '/'))
		return message(ERROR, "bad mime type: %s", tokens[0]);


	mime = MALLOC(sizeof(SimpleList));
	if (!mime) return FAILED;
	mime->name = strdup(tokens[0]);
	for (i = 1; i < count; i++) {
		SimpleList* extension = config->extensions;
		char ext[SHORT_STR];
        strtolower(strncpy_(ext, tokens[i], SHORT_STR));
		while (extension) {
			if (extension->link == mime->name &&
				strcmp(ext, extension->name) == 0)
                break;
			extension = extension->next;
		}
		if (extension) continue;

		extension = MALLOC(sizeof(SimpleList));
		if (!extension) return FAILED;
		extension->name = strdup(ext);
		extension->link = mime->name;
		extension->next = config->extensions;
		config->extensions = extension;
	}
	mime->next = config->mimes;
	config->mimes = mime;
	return OK;
}


int
setAdmin(char** tokens, int count)
{
	char* admin = strdup(tokens[0]);
	if (context == CONFIG) {
		FREE(config->admin);
		config->admin = admin;
	} else {
		FREE(currentServer->admin);
		currentServer->admin = admin;
	}
	return OK;
}

int
setSkipWWW(char** tokens, int count)
{
	config->skipWWW = booleanValue(tokens[0]);
	return OK;
}

int
setSkipPort(char** tokens, int count)
{
	config->skipPort = booleanValue(tokens[0]);
	return OK;
}

int
setFilesCheck(char** tokens, int count)
{
	config->filesCheck = atol(tokens[0]);
	if (config->filesCheck < 1 || config->filesCheck > 60*60)
		return message(ERROR, "bad filesCheck value: %s (out of [1 - 60*60])", tokens[0]);
	return OK;
}

int
setFilesKeep(char** tokens, int count)
{
	config->filesKeep = atol(tokens[0]);
	if (config->filesKeep < 1 || config->filesKeep > 60*60)
		return message(ERROR, "bad filesKeep value: %s (out of [1 - 60*60])", tokens[0]);
	return OK;
}

int
setIndex(char** tokens, int count)
{
	int i;
	for (i = count - 1; i >= 0; i--) {
		SimpleList* index = MALLOC(sizeof(SimpleList));
		if (!index) return FAILED;
		index->name = strdup(tokens[i]);
		if (context == CONFIG) {
			index->next = config->indexes;
			config->indexes = index;
		} else if (context == SERVER) {
			index->next = currentServer->indexes;
			currentServer->indexes = index;
		} else {
			index->next = currentAlias->indexes;
			currentAlias->indexes = index;
		}
	}
	return OK;
}

int
setReverse(char** tokens, int count)
{
	int i;
	for (i = count - 1; i >= 0; i--) {
		SimpleList* rev = MALLOC(sizeof(SimpleList));
		if (!rev) return FAILED;
		rev->name = strtolower(strdup(tokens[i]));
        reverse(rev->name);
		if (context == CONFIG) {
			rev->next = config->reverse;
			config->reverse = rev;
		} else if (context == SERVER) {
			rev->next = currentServer->reverse;
			currentServer->reverse = rev;
		} else {
			rev->next = currentAlias->reverse;
			currentAlias->reverse = rev;
		}
	}
	return OK;
}

#ifdef CHARSETS_H
int
setCharset(char** tokens, int count)
{
	int cs = strcasecmp(tokens[0], "off") ? cs_by_string(tokens[0]) : -1;
	if (context == CONFIG) {
		config->charset = cs;
	} else if (context == SERVER) {
		currentServer->charset = cs;
	} else {
		currentAlias->charset = cs;
	}
	return OK;
}

int
setSrcCharset(char** tokens, int count)
{
	int cs = strcasecmp(tokens[0], "off") ? cs_by_string(tokens[0]) : -1;
	if (context == CONFIG) {
		config->srcCharset = cs;
	} else if (context == SERVER) {
		currentServer->srcCharset = cs;
	} else {
		currentAlias->srcCharset = cs;
	}
	return OK;
}
#else
int
setCharset(char** tokens, int count)
{
	return message(ERROR, "charsets unsupported");
}

int
setSrcCharset(char** tokens, int count)
{
	return message(ERROR, "charsets unsupported");
}

#endif


int
setKeepAlive(char** tokens, int count)
{
	int v = booleanValue(tokens[0]);

	if (context == CONFIG) {
		config->keepalive = v;
	} else if (context == SERVER) {
		currentServer->keepalive = v;
	} else {
		currentAlias->keepalive = v;
	}
	return OK;
}

int
setLog(char** tokens, int count)
{
	int v = !strcasecmp(tokens[0], "errors") ? ERRORS :
		booleanValue(tokens[0]);

	if (context == CONFIG) {
		config->log = v;
	} else if (context == SERVER) {
		currentServer->log = v;
	} else {
		currentAlias->log = v;
	}
	return OK;
}

int
setThrottle(char** tokens, int count)
{
	int* v =
		context == CONFIG ? &config->throttle :
		context == SERVER ? &currentServer->throttle :
		&currentAlias->throttle;

	*v = booleanValue(tokens[0]);

	return OK;
}

int
setDetachRoot(char** tokens, int count)
{
	config->detachRoot = booleanValue(tokens[0]);
	return OK;
}



int
setListing(char** tokens, int count)
{
	int v = !strcasecmp(tokens[0], "simple") ? 2 : booleanValue(tokens[0]);

	if (context == CONFIG) {
		config->listing = v;
	} else if (context == SERVER) {
		currentServer->listing = v;
	} else {
		currentAlias->listing = v;
	}
	return OK;
}

int
setServerTokens(char** tokens, int count)
{
	config->serverTokens = !strcasecmp(tokens[0], "name") ? 2 : booleanValue(tokens[0]);
    return OK;
}

int
setServer(char** tokens, int count)
{
	Server* server = MALLOC(sizeof(Server));
	if (!server) return FAILED;
	server->name = strdup(tokens[0]);
	server->hash = hash(server->name);
	if (--count) {
		int i;
		server->alter = MALLOC((server->nOfAlters = count) * sizeof(char*));
		server->alterHash = MALLOC(count * sizeof(unsigned));
		for (i = 0; i < count; i++) {
			server->alter[i] = strdup(tokens[i + 1]);
			server->alterHash[i] = hash(server->alter[i]);
		}
	}
	server->keepalive = -1;
	server->log = -1;
	server->listing = -1;
	server->throttle = -1;

	server->next = config->servers;
	config->servers = server;

	currentServer = server;
	context = SERVER;
	return OK;
}


int
setProxy(Alias* alias, char** tokens, int count)
{
	int i;
	alias->type = PROXY;
	for (i = 0; i < count; i++) {
		char* c = strchr(tokens[i], ':');
		Proxy* proxy;
		struct hostent* he;
		int n;
		unsigned short port;

		if (!c) return message(ERROR, "bad address: %s", tokens[i]);
		*c++ = 0;

		if (!(he = gethostbyname(tokens[i]))) {
			if (h_errno == NETDB_INTERNAL)
				return message(SYS|ERROR, "cannot resolve %s", tokens[i]);
			return message(ERROR, "cannot resolve %s: %s", tokens[i],
						   h_errno == HOST_NOT_FOUND ? "HOST_NOT_FOUND" :
						   h_errno == NO_RECOVERY ? "NO_RECOVERY" :
						   h_errno == NO_DATA ? "NO_DATA" :
						   "unknown error");
		}

		if (!(port = strtoul(c, &c, 10)))
			return message(ERROR, "bad port: %s", c);

		n = strlen(c);
		if (*c && c[n - 1] == '$') {
			alias->exactLocation = YES;
			c[--n] = 0;
		}
		if (*c && c[n - 1] == '/') c[--n] = 0;

		for (n = 0; he->h_addr_list[n]; n++) {
            Backend* backend = config->backends;
			proxy = MALLOC(sizeof(Proxy));
			proxy->location = strdup(c);

			while (backend) {
				if (backend->port == port &&
					!memcmp(&(backend->addr), he->h_addr_list[n], sizeof(backend->addr)))
					break;
				backend = backend->next;
			}
			if (!backend) {
				backend = MALLOC(sizeof(Backend));
				memcpy(&(backend->addr), he->h_addr_list[n], sizeof(backend->addr));
                backend->name = strdup(tokens[i]);
				backend->port = port;
				backend->state = OK;
				backend->next = config->backends;
				config->backends = backend;
			}
			proxy->backend = backend;

			if (alias->proxy) {
				proxy->next = alias->proxy->next;
				alias->proxy->next = proxy;
			} else {
				alias->proxy = proxy->next = proxy;
			}
		}
	}
	return OK;
}

int
setAlias(char** tokens, int count)
{
	Alias* alias = MALLOC(sizeof(Alias));
	char* c;
	if (!alias) return FAILED;
	if (count < 2) return message(ERROR, "Alias must have at least two parameters");

	alias->alias = strdup(tokens[0]);
	alias->length = strlen(alias->alias);
	alias->keepalive = -1;
	alias->log = -1;
	alias->listing = -1;
	alias->throttle = -1;

	if (tokens[1][0] == '/') {
		if (count == 2) alias->type = LOCAL;
		else if (strcasecmp(tokens[2], "CGI") == 0)
			alias->type = CGI;
		else return message(ERROR, "unknown special: %s", tokens[2]);
		alias->location = strdup(tokens[1]);
	} else if (strncasecmp(tokens[1], "http://", strlen("http://")) == 0) {
		alias->type = REDIRECT;
		alias->location = strdup(tokens[1]);
	} else if (strcasecmp(tokens[1], "STATUS") == 0) {
		alias->type = STATUS;
	} else if (strcasecmp(tokens[1], "EMPTY") == 0) {
		alias->type = EMPTY;
	} else if (strcasecmp(tokens[1], "PIXEL") == 0) {
		alias->type = PIXEL;
		if (count > 2) { // pixel color
            unsigned color = strtoul(tokens[2], NULL, 16);
			char* c = (char*)&color;
			alias->aux[0] = MALLOC(3);
            alias->aux[0][0] = c[2];
            alias->aux[0][1] = c[1];
            alias->aux[0][2] = c[0];
		}
	} else if ((c = strchr(tokens[1], ':'))) {
		setProxy(alias, tokens + 1, count - 1);
#ifdef COUNTER_ADDON
	} else if (strcasecmp(tokens[1], "COUNTER") == 0) {
		alias->type = COUNTER;
	} else if (strcasecmp(tokens[1], "COUNTER_LOGO") == 0) {
		alias->type = COUNTER_LOGO;
	} else if (strcasecmp(tokens[1], "COUNTER_CLICK") == 0) {
		alias->type = COUNTER_CLICK;
	} else if (strcasecmp(tokens[1], "COUNTER_VALUES") == 0) {
		alias->type = COUNTER_VALUES;
#endif
#ifdef _3DNEWS_ADDON
	} else if (strcasecmp(tokens[1], "3DNEWS_REDIRECT") == 0) {
		alias->type = _3DNEWS_REDIRECT;
		if (!(tokens[2] && tokens[3] && tokens[4]))
			return message(ERROR, "3DNEWS_REDIRECT must follows by 'country mirror-URL original-URL'");
		alias->aux[0] = strdup(tokens[2]); // country
		alias->aux[1] = strdup(tokens[3]); // mirror for this country
		alias->aux[2] = strdup(tokens[4]); // original
	} else if (strcasecmp(tokens[1], "3DNEWS_MIRROR") == 0) {
		alias->type = _3DNEWS_MIRROR;
		alias->location = strdup(tokens[2]);
		alias->aux[0] = strdup(tokens[3]);
#endif
	} else {
		return message(ERROR, "unknown type of alias: %s", tokens[1]);
	}


	if (alias->length > 0 && alias->alias[alias->length - 1] == '$') {
		alias->exact = YES;
		alias->alias[--(alias->length)] = 0;
	} else if ((c = strchr(alias->alias, '*'))) {
		char* s = strtolower(c);
		int i = 0;
		alias->nOfTails = 1;
		while((c = (strchr(c + 1, ',')))) alias->nOfTails++;
		alias->tail = MALLOC(alias->nOfTails * sizeof(char*));
		alias->tailLength = MALLOC(alias->nOfTails * sizeof(int));
		do {
			*s++ = 0;
			if (*s == '*') s++;
			alias->tail[i++] = s;
		} while((s = (strchr(s + 1, ','))));

		for (i = 0; i < alias->nOfTails; i++)
			alias->tailLength[i] = strlen(alias->tail[i]);

		alias->length = strlen(alias->alias);
	} else if ((c = strchr(alias->alias, '|'))) {
		char* s = strtolower(c);
		int i = 0;
		alias->nOfCtype = 1;
		while((c = (strchr(c + 1, ',')))) alias->nOfCtype++;
		alias->ctype = MALLOC(alias->nOfCtype * sizeof(char*));
		alias->ctypeLength = MALLOC(alias->nOfCtype * sizeof(int));
		do {
			*s++ = 0;
			alias->ctype[i++] = s;
		} while((s = (strchr(s + 1, ','))));

		for (i = 0; i < alias->nOfCtype; i++) {
			alias->ctypeLength[i] = strlen(alias->ctype[i]);
			if (alias->ctype[i][alias->ctypeLength[i] - 1] == '*')
				alias->ctype[i][--(alias->ctypeLength[i])] = 0;
		}

		alias->length = strlen(alias->alias);
	}

	if (alias->alias[alias->length - 1] == '/')
		alias->alias[--(alias->length)] = 0;


	if (alias->location && alias->location[strlen(alias->location) - 1] == '$') {
		alias->exactLocation = YES;
		alias->location[strlen(alias->location) - 1] = 0;
	}


	if (alias->location && alias->location[strlen(alias->location) - 1] == '/')
		alias->location[strlen(alias->location) - 1] = 0;


	alias->next = currentServer->aliases;

	currentServer->aliases = alias;

	currentAlias = alias;
	context = ALIAS;
	return OK;
}

int
setCache(char** tokens, int count)
{
	Cache* cache;
	int size = -1;
	int age = -1;
	int v;
	char* end;

	v = strtol(tokens[0], &end, 10);
	if (*end == 'k')
		size = v*1024;
	else if (*end == 'K')
		size = v*1024;
	else if (*end == 'M')
		size = v*1024*1024;
	else if (*end == 'm')
		age = v*60;
	else if (*end == 'h')
		age = v*60*60;
	else
		return message(ERROR, "bad value for cache: %s", tokens[0]);

	if (size == 0 || size > 1024*1024*1024)
		return message(ERROR, "bad value for size of cache: %s", tokens[0]);
	if (age == 0 || age > 100*60*60)
		return message(ERROR, "bad value for age of cache: %s", tokens[0]);


	if (count > 1) {
		v = strtol(tokens[1], &end, 10);
		if (*end == 'k')
			size = v*1024;
		else if (*end == 'K')
			size = v*1024;
		else if (*end == 'M')
			size = v*1024*1024;
		else if (*end == 'm')
			age = v*60;
		else if (*end == 'h')
			age = v*60*60;
		else
			return message(ERROR, "bad value for cache: %s", tokens[1]);
		if (size == 0 || size > 1024*1024*1024)
			return message(ERROR, "bad value for size of cache: %s", tokens[1]);
		if (age == 0 || age > 100*60*60)
			return message(ERROR, "bad value for age of cache: %s", tokens[1]);
	}


	cache = MALLOC(sizeof(Cache));
	if (!cache) return FAILED;
	cache->size = size;
	cache->age = age;
	if (context == SERVER) {
		if (currentServer->cache)
			return message(ERROR, "cache for server %s already defined", currentServer->name);
		currentServer->cache = cache;
	} else {
		if (currentAlias->cache)
			return message(ERROR, "cache for alias %s%s already defined",
						   currentServer->name, currentAlias->alias);
		currentAlias->cache = cache;
	}
	return OK;
}

char
refreshPasswd(Access* access)
{
	int fd;
	int n;
	char* s;
	time_t mtime = fileTime(access->passwd);
	if (mtime <= access->mtime) return OK;

	access->users = freePairs(access->users);
	access->file = FREE(access->file);

	if ((fd = open(access->passwd, O_RDONLY)) < 0)
		return message(SYS|ERROR, "cannot open %s", access->passwd);

	n = fileSize(access->passwd);
	access->file = MALLOC(n + 1);
	read(fd, access->file, n);
	close(fd);

	access->mtime = mtime;
	access->users = newPairs(n/30?:4, NULL);

	s = access->file;
	do {
		char* name;
		char* passwd;
		char* eol = strchr(s, '\n');
		if (eol) {
			if (eol > s && eol[-1] == '\r') eol[-1] = 0;
			*eol++ = 0;
		}

		name = s;
		if ((passwd = strchr(name, ':'))) {
			char* end;
			*passwd++ = 0;
			if ((end = strchr(name, ':'))) *end = 0;

			setPair(access->users, name, passwd);
		}
		s = eol;
	} while(s);

	return OK;
}

char
parseAccess(Access* access, char* token)
{

	// allowed types:
	//	 any: *
	//	 country: ru
	//	 ip: 192.168.50.44
	//	 network: 192.168.50.0/24, 192.168.*, 192.168.
	//	 passwd-file: User:conf/passwd Site\ admin:/etc/httpd/passwd
	//	 referrer: referrer=self referrer=foreign referrer=empty referrer=site.ru
	//	 user agent: agent=DTS\ Agent
	int l = strlen(token);
	char* c;

	if (token[l - 1] == ',') token[--l] = 0;

	if (l == 1 && *token == '*') {
		access->type = ANY;
		return OK;
	}

	if (l == 2) {
		access->type = COUNTRY;
		access->country = domain2id(token);
		// hm, id can be zero, so no error checking yet :(
		//if (!(access->country = domain2id(token)))
		//return message(ERROR, "unknown domain: %s", token);
		return OK;
	}

	if ((l = string2ip(token)) != -1) {
		char* slash;
		access->type = NETWORK;
		access->network = l;
		access->mask = 0xFFFFFFFF;

		if ((slash = strchr(token, '/'))) {
			int bits = atol(slash + 1);
			if (bits <= 0 || bits > 32) return message(ERROR, "wrong bits count");
			access->mask <<= 32 - bits;
			if ((access->network & access->mask) != access->network)
				message(MESSAGE, "warning: network has more bits when mask: %s", token);
		}
		return OK;
	}

	if ((l = atol(token)) > 0 && l <= 255) {
		int ip = 0;
		unsigned char* n = (unsigned char*)&ip;
		int count = 3;
		char* str = token;
		while (1) {
			char* e;
			n[count] = strtol(str, &e, 10);
			if (e == str) break;
			if (count && *e == '.') str = e + 1;
			else break;
			if (!*str || (*str == '*' && !str[1])) {
				access->type = NETWORK;
				access->network = ip;
				access->mask = 0xFFFFFFFF << (count*8);
				return OK;
			}
			--count;
		}
	}

	if (!memcmp(token, "referrer=", strlen("referrer="))) {
		c = token + strlen("referrer=");
		strtolower(c);

		access->type = REFERRER;

		if (!strcmp(c, "empty"))
			access->refType = REFERRER_EMPTY;
		else if (!strcmp(c, "self"))
			access->refType = REFERRER_SELF;
		else if (!strcmp(c, "foreign"))
			access->refType = REFERRER_FOREIGN;
		else {
			access->refType = REFERRER_SITE;
			access->referrer = strdup(c);
		}

		return OK;
	}

	if (!memcmp(token, "agent=", strlen("agent="))) {
		c = token + strlen("agent=");
        if (!*c) return FAILED;

		access->type = AGENT;
        access->agent = strdup(c);

		return OK;
	}

	if ((c = strchr(token, ':'))) {
		*c++ = 0;
		if (!exists(c)) return message(ERROR, "password file %s does not exists", c);

		base64initialize();

		access->type = PASSWD;
		access->realm = strdup(token);
		access->passwd = strdup(c);
		c[-1] = ':';
		return refreshPasswd(access);
	}

	return FAILED;
}

int
setAccess(char** tokens, int count)
{
	Access* next;
	int what = 0;
	int i;

	if (context == CONFIG) {
		next = config->access;
	} else if (context == SERVER) {
		next = currentServer->access ? : config->access;
	} else {
		next = currentAlias->access ? : currentServer->access ? : config->access;
	}

	//Access Allow conf/passwd, Allow 192.168.50.0/24
	//Access Deny *, Allow ru, Allow ua
	//Access Deny * > /only-ru.html, Allow ru, ua
	for (i = 0; i < count; i++) {
		Access* access;
		if (!strcasecmp(tokens[i], "allow")) {what = 0; i++;}
		else if (!strcasecmp(tokens[i], "deny")) {what = 403; i++;}

		if (i == count)
			return message(ERROR, "access value expected at token No%d", i);

		access = MALLOC(sizeof(Access));
		access->code = what;
		if (!parseAccess(access, tokens[i]))
			return message(ERROR, "cannot parse access value at token No%d: %s", i, tokens[i]);

		if (what && i < count - 2 && tokens[i + 1][0] == '>') {
			char* token = tokens[i += 2];
			int l = strlen(token);
			if (token[l - 1] == ',') token[--l] = 0;
			access->redirect = strdup(token);
			access->code = 302;
		}

		access->next = next;
		next = access;
	}

	if (context == CONFIG) {
		config->access = next;
	} else if (context == SERVER) {
		currentServer->access = next;
	} else {
		currentAlias->access = next;
	}

	return OK;
}


int
setFilter(char** tokens, int count)
{
	Filter* filter;

	//Filter content/type magic
	if (count < 2) return message(ERROR, "undefined filter handler");

	filter = MALLOC(sizeof(Filter));
	filter->type = strdup(tokens[0]);
	filter->typeLength = strlen(filter->type);

	if (strcasecmp(tokens[1], "STRIP_META") == 0)
		filter->handler = &filterStripMeta;
#ifdef _3DNEWS_ADDON
	else if (strcasecmp(tokens[1], "3DNEWS") == 0)
		filter->handler = &filter3DNews;
#endif
	else
		return message(ERROR, "unknown filter handler: %s", tokens[1]);

	if (context == CONFIG) {
		config->filter = filter;
	} else if (context == SERVER) {
		currentServer->filter = filter;
	} else {
		currentAlias->filter = filter;
	}

	return OK;
}


void
fillAliasDefaults(Alias* alias, Server* server)
{
	if (!alias->timeout) alias->timeout = server->timeout;
	if (!alias->expires) alias->expires = server->expires;
	if (!alias->charset) alias->charset = server->charset;
	if (!alias->srcCharset) alias->srcCharset = server->srcCharset;
	if (!alias->indexes) alias->indexes = server->indexes;
	if (!alias->reverse) alias->reverse = server->reverse;
	if (!alias->cache) alias->cache = server->cache;
	if (!alias->access) alias->access = server->access;
	if (!alias->filter) alias->filter = server->filter;
	if (alias->keepalive < 0) alias->keepalive = server->keepalive;
	if (alias->log < 0) alias->log = server->log;
	if (alias->listing < 0) alias->listing = server->listing;
	if (alias->throttle < 0) alias->throttle = server->throttle;
}

void
fillServerDefaults(Server* server)
{
	Alias* alias = server->aliases;

	if (!server->admin) server->admin = config->admin;
	if (!server->timeout) server->timeout = config->timeout;
	if (!server->expires) server->expires = config->expires;
	if (!server->charset) server->charset = config->charset;
	if (!server->srcCharset) server->srcCharset = config->srcCharset;
	if (!server->indexes) server->indexes = config->indexes;
	if (!server->reverse) server->reverse = config->reverse;
	if (!server->access) server->access = config->access;
	if (!server->filter) server->filter = config->filter;
	if (server->keepalive < 0) server->keepalive = config->keepalive;
	if (server->log < 0) server->log = config->log;
	if (server->listing < 0) server->listing = config->listing;
	if (server->throttle < 0) server->throttle = config->throttle;

	while(alias) {
		fillAliasDefaults(alias, server);
		alias = alias->next;
	}
}

void
fillDefaults()
{
	Server* server = config->servers;
	while(server) {
		fillServerDefaults(server);
		server = server->next;
	}
}


int
readConfig(int argc, char** argv)
{
	char buf[LONG_STR];
	char name[PATH_LEN];
	FILE* f;
	int line = 0;
	int i;
	int offset = 0;

	setDefaults();

	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "-c") == 0) {
			if (++i == argc) return message(ERROR, "config name is not defined in command line");
			FREE(config->configName);
			config->configName = strdup(argv[i]);
		}
	}

	snprintf(name, PATH_LEN, "%s/%s", CONF, config->configName);
	if (!(f = fopen(name, "r")))
		return message(SYS|ERROR, "cannot open %s", name);

	context = CONFIG;

	while(LONG_STR > offset && fgets(buf + offset, LONG_STR - offset, f)) {
		line++;
		trimSpaces(buf);
		if (*buf && buf[strlen(buf) - 1] == '\\') {
			offset = strlen(buf) - 1;
		} else {
			char* tokens[MAX_TOKENS];
			int count = getTokens(buf, tokens, MAX_TOKENS);
			offset = 0;
			if (!count) continue;
			if (!setOption(tokens, count, name, line))
				return message(ERROR, "config error at %s (%s:%d)",
							   tokens[0], name, line);
		}
	}
	if (LONG_STR <= offset)
		return message(ERROR, "too long line in config (%s:%d)", name, line);

	fclose(f);

	fillDefaults();

	return OK;
}

