/* $Id: core.c,v 1.10 2002/05/15 10:28:57 maxim Exp $
 */

#define _GNU_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>


#include "common/settings.h"
#include "common/signames.h"
#include "common/strings.h"
#include "common/country.h"
#include "common/misc.h"


#include "core.h"
#include "httpd.h"
#include "flood.h"
#include "proxy.h"
#include "config.h"
#include "handlers.h"
#include "files.h"
#include "buffer.h"
#include "access.h"

#if TRANSFER_MODE == TRANSFER_SENDFILE

#if __linux__
#include <sys/sendfile.h>
#endif

#endif /* TRANSFER_MODE == TRANSFER_SENDFILE */


#if MODEL == MODEL_KEVENT
#include <sys/event.h>
#endif

#ifdef COUNTER_ADDON
#include "handle_counter.h"
#endif


static char *states[] = {
	"UNDEF",
	"READ_REQUEST",
	"KEEP_ALIVE",
	"READ_HEADER",
	"READ_POST",
	"HANDLE_REQUEST",
	"HANDLING",
	"PREPARE_HEADER",
	"WRITE_HEADER",
	"WRITE_BODY",
	"WRITE_CONTENT",
	"WAIT_DATA",
	"DONE",
	"CLOSE",
	"NOT_OK"
};


#define REQUESTS_STEP 64

char* methods[] = {"?", "HEAD", "GET", "POST"};

FD* fds = NULL;
int maxFD = 0;
int fdLimit = 0;

time_t cleanupTime = 0;

char past[SHORT_STR];
char future[SHORT_STR];

Request* currentRequest = NULL;
Buffer* currentBuffer = NULL;

static Request* requests = NULL;
static Request* freeRequests = NULL;
static Request** requestsBlocks = NULL;
static int nRequests = 0, maxRequests = 0, nRequestsBlocks = 0;

static unsigned totalAccepts = 0, totalRequests = 0;
static unsigned lastUser = 0, lastSys = 0, lastAccepts = 0, lastRequests = 0;


static int requestsFD = -1;

#if MODEL == MODEL_RTSIG
int rtSignal = 0;
#elif MODEL == MODEL_KEVENT
int kQueue = 0;
int nChanges, aChanges;
struct kevent *kChanges;
#endif

char* getStatusLine(int code);
int makeErrorReply(Request* r, char* buffer, int size);


time_t
timerfc(char *s)
{
	static const int daytab[2][12] = {
		{0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
		{0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 }
	};
	enum { D_START, D_END, D_MON, D_DAY, D_YEAR, D_HOUR, D_MIN, D_SEC};
	unsigned sec, min, hour, day, mon, year;
	char month[3];
	register char c;
	register unsigned n;
	register char flag;
	register char state;
	register char isctime;
	sec = 60;
	min = 60;
	hour = 24;
	day = 32;
	year = 1969;
	isctime = 0;
	month[0] = 0;
	state = D_START;
	n = 0;
	flag = 1;
	do {
		c = *s++;
		switch (state) {
		case D_START:
			if (c == ' ') {
				state = D_MON;
				isctime = 1;
			} else if (c == ',') state = D_DAY;
			break;
		case D_MON:
			if ('a' <= c && c <= 'z') {
				if (n < 3) month[n++] = c;
			} else {
				if (n < 3) return -1;
				n = 0;
				state = isctime ? D_DAY : D_YEAR;
			}
			break;
		case D_DAY:
			if (c == ' ' && flag)
				;
			else if ('0' <= c && c <= '9') {
				flag = 0;
				n = 10 * n + (c - '0');
			} else {
				day = n;
				n = 0;
				state = isctime ? D_HOUR : D_MON;
			}
			break;
		case D_YEAR:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				year = n;
				n = 0;
				state = isctime ? D_END : D_HOUR;
			}
			break;
		case D_HOUR:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				hour = n;
				n = 0;
				state = D_MIN;
			}
			break;
		case D_MIN:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				min = n;
				n = 0;
				state = D_SEC;
			}
			break;
		case D_SEC:
			if ('0' <= c && c <= '9')
				n = 10 * n + (c - '0');
			else {
				sec = n;
				n = 0;
				state = isctime ? D_YEAR : D_END;
			}
			break;
		}
	} while (state != D_END && c);
	switch (month[0]) {
	case 'a':
		mon = (month[1] == 'p') ? 4 : 8;
		break;
	case 'd':
		mon = 12;
		break;
	case 'f':
		mon = 2;
		break;
	case 'j':
		mon = (month[1] == 'a') ? 1 : ((month[2] == 'l') ? 7 : 6);
		break;
	case 'm':
		mon = (month[2] == 'r') ? 3 : 5;
		break;
	case 'n':
		mon = 11;
		break;
	case 'o':
		mon = 10;
		break;
	case 's':
		mon = 9;
		break;
	default:
		return -1;
	}
	if (year <= 100)
		year += (year < 70) ? 2000 : 1900;
	--mon;
	--day;
	if (sec >= 60 || min >= 60 || hour >= 60 || day >= 31 || year < 1970)
		return -1;
	return sec + 60L * (min + 60L * (hour + 24L * (day + daytab[year % 4 == 0][mon] + 365L * (year - 1970L) + ((year - 1969L) >> 2))));
}



int
copyListener(Listener* listener, Listener* old)
{
	while (old) {
		if (listener->addr.s_addr == old->addr.s_addr && listener->port == old->port) {
			if (old->fd >= 0) {
				listener->fd = old->fd;
				listener->accepts = old->accepts;
				listener->acceptsToday = old->acceptsToday;
				listener->acceptsLast = old->acceptsLast;
				listener->droppedS = old->droppedS;
				listener->droppedC = old->droppedC;

				fds[old->fd].listener = listener;
				old->fd = -1;
				return OK;
			}
			return FAILED;
		}
		old = old->next;
	}
	return FAILED;
}


int
initListeners(Listener* old)
{
	Listener* listener;
	int inited = 0;
	uid_t uid = getuid();
	uid_t euid = geteuid();

	for (listener = config->listeners; listener; listener = listener->next) {
		struct sockaddr_in sa;
		int fd;
		int onoff = 1;
		int err;

		debug("initListener(%s:%d)", inet_ntoa(listener->addr), listener->port);
		if (old && copyListener(listener, old)) {
			message(MESSAGE, "using old listener %s:%d", inet_ntoa(listener->addr), listener->port);
			inited++;
			continue;
		}

		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		fd = socket(AF_INET, SOCK_STREAM, 0);
		err = errno;
		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);

		if (fd < 0) {
			char e[LONG_STR];
			errno = err;
			snprintf(e, LONG_STR, "cannot create socket: %s", strerror(errno));
			listener->error = strdup(e);
			message(SYS|ERROR, "cannot create listener socket");
			continue;
		}

		if (fd >= fdLimit) {
			char e[LONG_STR];
			snprintf(e, LONG_STR, "too big fd (%d > %d)", fd, fdLimit);
			listener->error = strdup(e);
			message(ERROR, "cannot init listener %s:%d: too big fd (%d > %d)", inet_ntoa(listener->addr), listener->port, fd, fdLimit);
			close(fd);
			continue;
		}

		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &onoff, sizeof(onoff)))
			message(SYS|ERROR, "cannot setsockopt(SO_REUSEADDR)");
#if __FreeBSD__
		if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &onoff, sizeof(onoff)))
			message(SYS|ERROR, "cannot setsockopt(SO_REUSEPORT)");
#endif

		//fcntl(fd, F_SETFD, FD_CLOEXEC);

#if MODEL == MODEL_RTSIG
		if (fcntl(fd, F_SETSIG, rtSignal))
			message(SYS|ERROR, "cannot fcntl(F_SETSIG)");
		if (fcntl(fd, F_SETOWN, myPID))
			message(SYS|ERROR, "cannot fcntl(F_SETOWN)");
		if (fcntl(fd, F_SETFL, O_NONBLOCK|O_ASYNC))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#elif MODEL == MODEL_KEVENT
		watchEvent(fd, EVFILT_READ, EV_ADD, 0, 0);
		if (fcntl(fd, F_SETFL, O_NONBLOCK))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#else
		if (fcntl(fd, F_SETFL, O_NONBLOCK))
			message(SYS|ERROR, "cannot fcntl(F_SETFL)");
#endif

		memset(&sa, 0, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_addr = listener->addr;
		sa.sin_port = htons(listener->port);

		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		onoff = bind(fd, (struct sockaddr*) &sa, sizeof sa);
		err = errno;
		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);
		if (onoff < 0) {
			char e[LONG_STR];
			errno = err;
			snprintf(e, LONG_STR, "cannot bind: %s", strerror(errno));
			listener->error = strdup(e);
			message(SYS|ERROR, "cannot bind on %s:%d", inet_ntoa(listener->addr), listener->port);
			continue;
		}

		if (listen(fd, config->connections) < 0) {
			char e[LONG_STR];
			snprintf(e, LONG_STR, "cannot listen: %s", strerror(errno));
			listener->error = strdup(e);
			message(SYS|ERROR, "cannot listen %s:%d", inet_ntoa(listener->addr), listener->port);
			continue;
		}
		fds[fd].listener = listener;
		listener->fd = fd;

		message(MESSAGE, "listening %s:%d", inet_ntoa(listener->addr), listener->port);
		inited++;
	}
	if (!inited) return message(ERROR, "nothing to listen");

	{
		struct rlimit rl;
		int err;
		if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
		getrlimit(RLIMIT_NOFILE, &rl);
		rl.rlim_cur = fdLimit;
		if (rl.rlim_max < fdLimit) {
			message(MESSAGE, "warning: rlim_max(%d) < fdLimit(%d)", (int)(rl.rlim_max), fdLimit);
			rl.rlim_max = fdLimit;
		}
		inited = setrlimit(RLIMIT_NOFILE, &rl);
		err = errno;
		if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);
		errno = err;
		if (inited < 0) message(SYS|ERROR, "cannot set RLIMIT_NOFILE=%d/%d", (int)(rl.rlim_cur), (int)(rl.rlim_max));
	}

	if (config->detachRoot) {
		uid_t gid = getgid();
		uid_t egid = getegid();

		if (uid != euid && setreuid(euid, euid)) message(SYS|ERROR, "cannot setreuid(%d, %d)", euid, euid);
		if (gid != egid && setregid(egid, egid)) message(SYS|ERROR, "cannot setregid(%d, %d)", euid, euid);
	}
	return OK;
}

char*
serverTokens()
{
	return
		config->serverTokens == 2 ? "Server: 0W" RN :
		config->serverTokens ? "Server: 0W/" VERSION RN :
        "";
}

int
makeBusyReply(char* buffer, int size, char* message)
{
	return catstr(buffer, size,
				  "HTTP/1.0 503 Service Unavailable" RN,
                  serverTokens(),
				  "Date: ", now_str, RN
				  "Content-type: text/plain" RN RN,
				  message ? : "The server is temporarily busy. Try again later." RN,
				  NULL);
}

void
acceptRequest(Listener* listener)
{
	debug("%d> acceptRequest()", listener->fd);
	do {
		struct sockaddr_in sa;
		int lsa, fd;
		Request* r;
		unsigned ip;


		lsa = sizeof sa;
		if ((fd = accept(listener->fd, (struct sockaddr *) &sa, &lsa)) < 0) {
			if (errno != EAGAIN) message(SYS|ERROR, "cannot accept");
			break;
		}
		debug("%d> accepted from %s", fd, inet_ntoa(sa.sin_addr));

		totalAccepts++;
		listener->accepts++;
		if (fd >= fdLimit) {
			char reply[SHORT_TEXT];
			int n = makeBusyReply(reply, SHORT_TEXT, NULL);
			write(fd, reply, n);
			message(ERROR, "connection to %s dropped: too many fds", inet_ntoa(sa.sin_addr));
			if (close(fd)) message(SYS|ERROR, "cannot close");
			listener->droppedS++;
			continue;
		}
		if (nRequests == config->connections) {
			char reply[SHORT_TEXT];
			int n = makeBusyReply(reply, SHORT_TEXT, NULL);
			write(fd, reply, n);
			message(ERROR, "connection to %s dropped: too many requests", inet_ntoa(sa.sin_addr));
			if (close(fd)) message(SYS|ERROR, "cannot close");
			listener->droppedS++;
			continue;
		}

		ip = ntohl(sa.sin_addr.s_addr);
		if (config->access) {
			Request req;
			Request* r = &req;
			memset(r, 0, sizeof(Request));
			r->ip = ip;
			if ((r->status = checkAccess(r, config->access, &r->location))) {
				char reply[MEDIUM_TEXT];
				char loc[LONG_STR];
				int n;
				if (r->location) catstr(loc, LONG_STR, "Location: ", r->location, RN, NULL);
				else *loc = 0;
				n = catstr(reply, MEDIUM_TEXT,
						   "HTTP/1.0 ", itoa(r->status), " ",
						   getStatusLine(r->status), RN,
						   serverTokens(),
						   "Date: ", now_str, RN
						   "Content-type: text/html" RN,
						   loc, RN,
						   NULL);
				n += makeErrorReply(r, reply + n, MEDIUM_TEXT - n);
				write(fd, reply, n);
				if (close(fd)) message(SYS|ERROR, "cannot close");
				listener->droppedC++;
				continue;
			}
		}

		if (!addIP(ip)) {
			char reply[SHORT_TEXT];
			int n = makeBusyReply(reply, SHORT_TEXT,
								  fmt("Too many simultaneous connections from one IP, current limit: %d." RN,
									  config->connectionsPerIP));
			write(fd, reply, n);
			if (close(fd)) message(SYS|ERROR, "cannot close");
			listener->droppedC++;
			debug("connection to %s dropped: too many connections per IP", inet_ntoa(sa.sin_addr));
			continue;
		}

		if (!(r = freeRequests)) {
			debug("acceptRequest: allocating new requests block");
			freeRequests = MALLOC(REQUESTS_STEP * sizeof(Request));
			if (!freeRequests) {
				char reply[SHORT_TEXT];
				int n = makeBusyReply(reply, SHORT_TEXT, NULL);
				write(fd, reply, n);
				message(ERROR, "connection to %s dropped: not enough memory", inet_ntoa(sa.sin_addr));
				if (close(fd)) message(SYS|ERROR, "cannot close");
				decIP(ip);
				listener->droppedS++;
				continue;
			}
			for (r = freeRequests; r < freeRequests + REQUESTS_STEP - 1; r++)
				r->next = r + 1;
			r = requestsBlocks[nRequestsBlocks++] = freeRequests;
		}

		freeRequests = freeRequests->next;
		if ((r->next = requests)) r->next->prev = r;
		requests = r;
		r->prev = NULL;


		r->type = REQUEST;
		r->state = READ_REQUEST;
		r->fd = fd;
		r->rangeFrom = r->rangeTo = r->inContentLength =
			r->file = r->length = -1;
		r->major = 1;
		r->ip = ip;
		r->last = now + config->requestTimeout;
		r->start = now;

		fds[fd].request = r;

		if (fd > maxFD) maxFD = fd;

		if (++nRequests > maxRequests) maxRequests = nRequests;

#if MODEL == MODEL_RTSIG
		fcntl(fd, F_SETSIG, rtSignal);
		fcntl(fd, F_SETOWN, myPID);
		fcntl(fd, F_SETFL, O_NONBLOCK|O_ASYNC);
#else
		fcntl(fd, F_SETFL, O_NONBLOCK);
#endif
		//fcntl(fd, F_SETFD, FD_CLOEXEC);

		http(r);
	} while (TRUE);
}


int
handleEvent(int fd)
{
	char* con_type;

	debug("%d> handleEvent()", fd);

	//	if (!(type & 0xFFF))
	//		return message(MESSAGE, "unsupported event: %X", type);

	if (fd < 0 || fd >= fdLimit)
		return message(MESSAGE, "bad fd: %d", fd);

	if ((con_type = fds[fd].type) == NULL)
		return FAILED;
	//return message(MESSAGE, "closed fd: %d", fd);

	if (*con_type == REQUEST) http(fds[fd].request);
	else if (*con_type == LISTENER) acceptRequest(fds[fd].listener);
	else if (*con_type == PROXY_REQUEST) proxy(fds[fd].proxy);
	else return message(MESSAGE, "bad connection type: %d:%X", fd, *fds[fd].type);
	currentRequest = NULL;


	return OK;
}

void
getUsage()
{
	static int prevAccepts;
	static unsigned prevRequests;
	static struct rusage prevUsage;
	struct rusage usage;
	static int fd = -2;
	Server* server = config->servers;
	Listener* listener = config->listeners;
	Backend* backend = config->backends;

	getrusage(RUSAGE_SELF, &usage);

	lastUser = (usage.ru_utime.tv_sec - prevUsage.ru_utime.tv_sec)*1000L + (usage.ru_utime.tv_usec - prevUsage.ru_utime.tv_usec)/1000L;
	lastSys = (usage.ru_stime.tv_sec - prevUsage.ru_stime.tv_sec)*1000L + (usage.ru_stime.tv_usec - prevUsage.ru_stime.tv_usec)/1000L;
	lastAccepts = totalAccepts - prevAccepts;
	lastRequests = totalRequests - prevRequests;

	if (fd >= 0) {
		char buf[NORM_STR];
		int n = catstr(buf, NORM_STR, itoa(totalAccepts), "\n", itoa(totalRequests), "\n", NULL);
		static int prev = 1111;
		write(fd, buf, n);
		if (prev > n) ftruncate(fd, n);
		prev = n;
		lseek(fd, 0, SEEK_SET);
	} else if (fd == -2) {
		if (strcmp(config->configName, "httpd.conf"))
			fd = -3;
		else if ((fd = open("data/httpd/counts", O_CREAT|O_WRONLY, 0644)) < 0)
			message(SYS|ERROR, "cannot open %s", "data/httpd/counts");
	}


	if (totalAccepts >= 2000000000) totalAccepts -= 2000000000;
	if (totalRequests >= 2000000000) totalRequests -= 2000000000;

	prevUsage = usage;
	prevAccepts = totalAccepts;
	prevRequests = totalRequests;

	while(server) {
		server->requestsLast = server->requests;
		server->sentLast = server->sent;
		server->requestsToday += server->requests;
		server->sentToday += server->sent;

		server->requests = server->sent = 0;

		server = server->next;
	}

	while(listener) {
		listener->acceptsLast = listener->accepts;
		listener->acceptsToday += listener->accepts;

		listener->accepts = 0;
		listener = listener->next;
	}

	while (backend) {
		backend->requestsLast = backend->requests;
		backend->requestsToday += backend->requests;

		backend->requests = 0;
		backend = backend->next;
	}

}

void
resetToday()
{
	Listener* listener = config->listeners;
	Server* server = config->servers;
	Backend* backend = config->backends;

	message(MESSAGE, "Day totals:");
	while(listener) {
		if (!listener->error)
			message(MESSAGE, "listener %s:%d: %s accepts, %s + %s dropped",
					listener->addr.s_addr ? inet_ntoa(listener->addr): "*",
					listener->port,
					itos(listener->accepts),
					itos(listener->droppedS),
					itos(listener->droppedC));
		listener->acceptsToday = listener->droppedS = listener->droppedC = 0;
		listener = listener->next;
	}

	while(server) {
		message(MESSAGE, "server %s: %s requests, %sb sent",
				server->name,
				itos(server->requestsToday),
				humanSize(server->sentToday));

		server->requestsToday = 0;
		server->sentToday = 0;
		server = server->next;
	}

	while (backend) {
		backend->requestsToday = backend->maxRequests = 0;
		backend = backend->next;
	}

	maxRequests = 0;

	resetTodayProxy();
	resetTodayFiles();
	resetTodayFlood();
}

void
cleanup()
{
	static int prev_day;
	static time_t usageTime = 0;
	static time_t rangesTime = 0;
	static time_t filesTime = 0;
	Request* r;

	debug("cleanup");

	cleanupProxy();

	r = requests;
	while(r) {
		Request* next = r->next;
		if (now > r->last) {
			debug("%d> timeout in %s", r->fd, states[r->state]);
			if (!r->throttled) r->state = CLOSE;
			http(r);
		}
		r = next;
	}


	if (filesTime/5 != now/5) {
		filesTime = now;
		cleanupFiles();
	}

	if (usageTime/5 != now/5) {
		usageTime = now;
		getUsage();
	}

	if (rangesTime/60 != now/60) {
		rangesTime = now;
		refreshRanges();
	}

	if (prev_day != now_day) {
		if (prev_day) resetToday();

		strcpy(past, time2str(now - 365*24*60*60, HTTP_TIME));
		strcpy(future, time2str(now + 365*24*60*60, HTTP_TIME));
		prev_day = now_day;
	}

	cleanupTime = now;
}



void
closeListeners(Listener* listener)
{
	while (listener) {
		if (listener->fd >= 0) {
			if (close(listener->fd) < 0)
				message(SYS|ERROR, "cannot close %s:%d", inet_ntoa(listener->addr), listener->port);
			fds[listener->fd].listener = NULL;
			listener->fd = -1;
		}
		listener = listener->next;
	}
}

void
freeAll()
{
	int i;
	Request* r = requests;

	message(MESSAGE, "freing all resources");

	closeListeners(config->listeners);

	while (r) {
		Request* next = r->next;
		r->state = CLOSE;
		http(r);
		r = next;
	}

	resetToday();
	fds = FREE(fds);
	currentBuffer = freeBuffer(currentBuffer);
	for (i = 0; i < nRequestsBlocks; i++)
		FREE(requestsBlocks[i]);
	requestsBlocks = FREE(requestsBlocks);
	freeRequests = NULL;
	freeProxyRequests();
	freeFiles();

#ifdef COUNTER_ADDON
	freeCounterRequests();
#endif
#if MODEL == MODEL_KEVENT
	kChanges = FREE(kChanges);
	nChanges = aChanges = 0;
#endif
	nRequests = maxRequests = nRequestsBlocks = maxFD = 0;
	config = freeConfig(config);
	message(CLOSE_LOGS, NULL);
}

int
copyServerInfo(Server* server, Server* old)
{
	while (old) {
		if (server->hash == old->hash && !strcmp(server->name, old->name)) {
			server->requests	  = old->requests;
			server->sent		  = old->sent;
			server->requestsLast  = old->requestsLast;
			server->sentLast	  = old->sentLast;
			server->requestsToday = old->requestsToday;
			server->sentToday	  = old->sentToday;
			return OK;
		}
		old = old->next;
	}
	return FAILED;
}

int
copyBackendInfo(Backend* backend, Backend* old)
{
	while (old) {
		if (backend->addr.s_addr == old->addr.s_addr && backend->port == old->port) {
			backend->requests	   = old->requests;
			backend->requestsToday = old->requestsToday;
			return OK;
		}
		old = old->next;
	}
	return FAILED;
}


int
reloadConfig(int argc, char** argv)
{
	Config* old = config;
	int oldLimit = fdLimit;
	Server* server;
	Backend* backend;

	if (configured == now)
		return message(MESSAGE, "skipping double reload"), OK;

	message(MESSAGE, "reloading config");


	if (!readConfig(argc, argv)) {
		message(MESSAGE, "config error, using old config");
		config = old;
		return OK;
	}

	fdLimit = 2 * config->connections + config->proxyConnections + config->nOfListeners + 16;
	if (fdLimit > oldLimit &&
		!(fds = REALLOC(fds, sizeof(FD) * fdLimit)))
		return FAILED;

	if (config->connections > old->connections &&
		!(requestsBlocks = REALLOC(requestsBlocks, ((config->connections / REQUESTS_STEP) + 1) * sizeof(Request*))))
		return FAILED;

	if (config->proxyConnections > old->proxyConnections  &&
		!initProxyRequests())
		return FAILED;

	if (fdLimit > oldLimit && !initFiles())
		return FAILED;

	if (!initListeners(old->listeners))
		return FAILED;

	closeListeners(old->listeners);

	for (server = config->servers; server; server = server->next)
		copyServerInfo(server, old->servers);

	for (backend = config->backends; backend; backend = backend->next)
		copyBackendInfo(backend, old->backends);

	if (!old->active) freeConfig(old);


	if (nRequests > config->connections) {
		int i = nRequests - config->connections;
		Request* r = requests;
		while (i--) {
			Request* next = r->next;
			r->state = CLOSE;
			http(r);
			r = next;
		}
	}

	{
		Request* r = requests;
		while (r) {
			Request* next = r->next;
			if (getActive(r->ip) > config->connectionsPerIP) {
				r->state = CLOSE;
				http(r);
			}
			r = next;
		}
	}

	configured = now;

	return OK;
}


int
init(int argc, char** argv)
{
	getTime();
	configured = now;
	strcpy(past, time2str(now - 365*24*60*60, HTTP_TIME));
	strcpy(future, time2str(now + 365*24*60*60, HTTP_TIME));

	return
		(readConfig(argc, argv) &&
		 (fdLimit = 2 * config->connections + config->proxyConnections + config->nOfListeners + 16) &&
		 (fds = MALLOC(sizeof(FD) * fdLimit)) &&
		 (currentBuffer = newBuffer(INITIAL_BUFFER_SIZE)) &&
		 (requestsBlocks = MALLOC(((config->connections / REQUESTS_STEP) + 1) * sizeof(Request*))) &&
		 initProxyRequests() &&
		 initFiles() &&
#ifdef COUNTER_ADDON
		 initCounterRequests() &&
#endif
		 initListeners(NULL));
}



#if MODEL == MODEL_POLL
int
fillListenersPoll(struct pollfd* pollfds)
{
	struct pollfd* pollfd = pollfds;
	Listener* listener = config->listeners;
	while (listener) {
		if (listener->fd >= 0) {
			pollfd->events = POLLIN;
			pollfd->fd = listener->fd;
			pollfd++;
		}
		listener = listener->next;
	}
	return pollfd - pollfds;
}

int
fillPoll(struct pollfd* pollfds)
{
	Request* r;
	struct pollfd* pollfd = pollfds;

	for (r = requests; r; r = r->next) {
		short ev = r->watchRead ? POLLIN : r->watchWrite ? POLLOUT : 0;
		if (ev) {
			pollfd->events = ev;
			pollfd->fd = r->fd;
			pollfd++;
		}
	}
	return pollfd - pollfds;
}


char
work(int argc, char** argv)
{

	while (!doQuit) {
		struct pollfd* pollfds;
		int l;

		if (!init(argc, argv)) return FAILED;

		pollfds = MALLOC(sizeof(struct pollfd) * (config->nOfListeners + config->connections));
		l = fillListenersPoll(pollfds);

		while (!(doQuit | doRestart | doReload)) {
			int n = l;
			int i;
			int result;
			n += fillPoll(pollfds + n);
			n += fillProxyPoll(pollfds + n);
			usleep(10000);
			debug("polling %d fds...", n);
			result = poll(pollfds, n, 5000);
			debug("ready %d fds", result);
			getTime();
			if (result > 0) {
				struct pollfd* pollfd = pollfds;
				for (i = 0; i < n && result; i++, pollfd++) {
					if (pollfd->revents) {
						handleEvent(pollfd->fd);
						result--;
					}
				}
			} else if (result < 0) {
				if (errno != EINTR) message(SYS|ERROR, "cannot poll");
				if (errno == ENOMEM) sleep(10);
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) {
					doQuit = YES;
				} else {
					FREE(pollfds);
					pollfds = MALLOC(sizeof(struct pollfd) * (config->nOfListeners + config->connections));
					l = fillListenersPoll(pollfds);
				}
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
		}
		FREE(pollfds);
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#elif MODEL == MODEL_RTSIG
char
work(int argc, char** argv)
{
	//	int count = 0;
	sigset_t sigset;
	rtSignal = SIGRTMIN;
	sigemptyset(&sigset);
	sigaddset(&sigset, rtSignal);
	sigprocmask(SIG_BLOCK, &sigset, NULL);

	while (!doQuit) {
		if (!init(argc, argv)) return FAILED;
		while (!(doQuit | doRestart)) {
			int r;
			siginfo_t info;
			sigset_t set = sigset;
			struct timespec timeout = {5, 0};
			memset(&info, 0, sizeof(siginfo_t));

			debug("waiting for signal...");
			r = sigtimedwait(&set, &info, &timeout);
			//if ((count++ & 0xFF) == 0) getTime();
			getTime();
			if (r > 0) {
				if (info.si_signo == rtSignal)
					handleEvent(info.si_fd);
				else
					message(ERROR,"hmm, sigwaitinfo returned unwanted signal: %s", signame(info.si_signo));
			} else if (errno != EINTR && errno != EAGAIN) {
				message(SYS|ERROR, "cannot sigwaitinfo");
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) doQuit = YES;
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
		}
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#elif MODEL == MODEL_KEVENT

void
watchEvent(int fd, int filter, int flags, u_int fflags, int data)
{
	if (nChanges == aChanges) {
		if (aChanges) aChanges <<= 1;
		else aChanges = 32;
		kChanges = REALLOC(kChanges, aChanges * sizeof(struct kevent));
	}

	debug("Watching event[%d]: fd=%d, %s %s (%d,%d)", nChanges, fd, filter == EVFILT_READ ? "read" : "write", flags&EV_DELETE? "(delete)" : flags&EV_ONESHOT? "(oneshot)" : flags&EV_ENABLE? "(enable)" : flags&EV_DISABLE? "(disable)" : "", fflags, data);
	EV_SET(kChanges + nChanges, fd, filter, flags, fflags, data, NULL);
	nChanges++;
}

char
work(int argc, char** argv)
{
	if ((kQueue = kqueue()) < 0)
		return message(SYS|ERROR, "cannot create kqueue");

	while (!doQuit) {
		if (!init(argc, argv)) return FAILED;
		while (!(doQuit | doRestart | doReload)) {
			struct timespec timeout = {5, 0};
			int r;

			debug("waiting for event...");
			r = kevent(kQueue, kChanges, nChanges,
					   kChanges, aChanges,
					   &timeout);
			nChanges = 0;
			getTime();
			if (r > 0) {
				struct kevent* ke = kChanges;
				int i;
				for (i = 0; i < r; i++, ke++) {
					debug("Event: ident=%d, filter=%d, flags=%d, fflags=%d, data=%d",
						  ke->ident, ke->filter, ke->flags, ke->fflags, ke->data);
					handleEvent(ke->ident);
				}
			} else if (r < 0 && errno != EINTR && errno != EAGAIN) {
				message(SYS|ERROR, "cannot wait event");
			}
			if (now != cleanupTime) cleanup();
			if (doReload) {
				if (!reloadConfig(argc, argv)) doQuit = YES;
				message(CLOSE_LOGS, NULL);
				doReload = NO;
			}
		}
		freeAll();
		doRestart = NO;
	}
	return OK;
}

#else
#error "select model not ready yet"
#endif /* MODEL */


int
parseRequestLine(char* line, Request* r)
{
	char *method, *uri, *proto, *major, *minor;
	method = line;

	if (!(uri = strchr(method, ' '))) return message(MESSAGE, "400-no uri: %s", method), 400;
	*uri++ = 0;
	while(*uri==' ') uri++;

	if (!(proto = strchr(uri, ' '))) return message(MESSAGE, "400-no proto: %s", uri), 400;
	if (proto - uri >= LONG_STR) uri[LONG_STR - 1] = 0;

	*proto++ = 0;
	while(*proto==' ') proto++;

	if (!(major = strchr(proto, '/'))) return message(MESSAGE, "400-no major: %s", proto), 400;
	*major++ = 0;

	if (!(minor = strchr(major, '.'))) return message(MESSAGE, "400-no minor: %s", major), 400;
	*minor++ = 0;


	if (strcmp(strtolower(method), "get") == 0)
		r->method = GET;
	else if (strcmp(method, "head") == 0)
		r->method = HEAD;
	else if (strcmp(method, "post") == 0)
		r->method = POST;
	else return message(MESSAGE, "501-bad method: %s", method), 501;

	if (strcmp(strtolower(proto), "http"))
		return message(MESSAGE, "400-bad proto: %s", proto), 400;

	r->major = atoi(major); if (r->major < 1) r->major = 1;
	r->minor = atoi(minor); if (r->minor < 0) r->minor = 0;

	if (r->major > 1) return message(MESSAGE, "505-bad major: %s", major), 505;

	r->keepalive = r->major && r->minor;

	r->uri = uri;

	return 0;
}

char*
unifyHostName(char* host)
{
	char* port;

	strtolower(host);
	if (config->skipWWW && !memcmp(host, "www.", strlen("www.")))
		host += strlen("www.");

	if ((port = strchr(host, ':'))) {
		if (config->skipPort) *port = 0;
		else if (!memcmp(port, ":80", strlen(":80") + 1)) *port = 0;
	}

	port = host + strlen(host) - 1;
	while (*port == '.' && port >= host) *port-- = 0;

	return host;
}



int
parseHeaderLine(char* line, Request* r)
{
	char* value = strchr(line, ':');
	if (!value) return message(MESSAGE, "strange header line: %s", line);
	*value++ = 0;
	while (isspace(*value)) value++;
	if (!*value) return 0;

	strtolower(line);

	if (!(r->host || strcmp(line, "host"))) {
		r->host = unifyHostName(value);
	} else if (!(r->userAgent || strcmp(line, "user-agent"))) {
		r->userAgent = value;
	} else if (!(r->referrer || strcmp(line, "referer"))) {
		r->referrer = value;
	} else if (!(r->cookie || strcmp(line, "cookie"))) {
		r->cookie = value;
	} else if (!(r->authorization || strcmp(line, "authorization"))) {
		r->authorization = value;
	} else if (!(r->ims || strcmp(line, "if-modified-since"))) {
		r->ims = timerfc(strtolower(value));
		if (r->ims == (time_t) -1) r->ims = 0;
	} else if (!(r->iums || strcmp(line, "if-unmodified-since"))) {
		r->iums = timerfc(strtolower(value));
		if (r->iums == (time_t) -1) r->iums = 0;
	} else if (!(r->ifRange || strcmp(line, "if-range"))) {
		r->ifRange = timerfc(strtolower(value));
		if (r->ifRange == (time_t) -1) r->ifRange = 0;
	} else if (!strcmp(line, "range")) {
		int n;
		char *end;
		if (strncmp(value, "bytes=", strlen("bytes="))) return 0;
		value += strlen("bytes=");
		errno = 0;
		if (*value == '-') { // last N bytes: bytes=-N
			n = strtol(++value, &end, 10);
			if (n && end > value && !*end && !errno) r->rangeTo = n;
			return 0;
		}


		n = strtol(value, &end, 10);
		if (end == value || *end != '-' || errno) return 0;
		r->rangeFrom = n;

		value = end + 1;
		if (!*value) return 0; // skip first N bytes: bytes=N-

		errno = 0;
		n = strtol(value, &end, 10);
		if (end == value || *end != 0 || errno) return 0;
		r->rangeTo = n;		   // bytes=M-N

	} else if (!strcmp(line, "connection")) {
		strtolower(value);
		if (r->minor) {
			if (!strcmp(value, "close"))
				r->keepalive = NO;
		} else {
			if (!strcasecmp(value, "keep-alive"))
				r->keepalive = YES;
		}
	} else if (!(r->realIP || strcmp(line, "x-real-ip"))) {
		r->realIP = string2ip(value);
	} else if (!(r->forwardedFor || strcmp(line, "x-forwarded-for"))) {
		r->forwardedFor = string2ip(value);
		if (r->forwardedFor < 0x01000000) r->forwardedFor = -1;
		//message(MESSAGE, "strange ff1: %s\n", value);
	} else if (!(r->forwardedFor || strcmp(line, "forwarded"))) {
		if ((value = rindex(value, ' '))) {
			r->forwardedFor = string2ip(value + 1);
			if (r->forwardedFor < 0x01000000) r->forwardedFor = -1;
			//message(MESSAGE, "strange ff2: %s\n", value + 1);
		}
	} else if (!(r->inContentType || strcmp(line, "content-type"))) {
		if (r->method == POST) r->inContentType = value;
	} else if (!(r->inContentLength >= 0 || strcmp(line, "content-length"))) {
		if (r->method == POST) r->inContentLength = atoi(value);
	} else {
		int i = 0;

		for (i = 0; i < MAX_OTHER_HEADERS; i++) {
			if (!r->inHeaders[i]) {
				line[strlen(line)] = ':';
				r->inHeaders[i] = line;
				break;
			}
		}
	}
	return 0;
}

#define COPY_FIELD(name) if (r->name >= start && r->name < end) r->name += diff

Buffer*
copyRequestBuffer(Request* r, Buffer* src, int increase)
{
	int diff, i;
	char* start = src->buffer;
	char* end = src->buffer + src->count;
	if (!(r->buffer = newBuffer(increase ? src->size << 1 : src->size)))
		return NULL;
	memcpy(r->buffer->buffer, src->buffer,
		   r->buffer->count = src->count);
	r->buffer->start = src->start;

	diff = r->buffer->buffer - src->buffer;
	COPY_FIELD(uri);
	COPY_FIELD(path);
	COPY_FIELD(params);
	COPY_FIELD(query);
	COPY_FIELD(userAgent);
	COPY_FIELD(referrer);
	COPY_FIELD(authorization);
	COPY_FIELD(user);
	COPY_FIELD(cookie);
	COPY_FIELD(host);
	COPY_FIELD(inContentType);
	for (i = 0; r->inHeaders[i]; i++)
		COPY_FIELD(inHeaders[i]);

	COPY_FIELD(contentType);
	COPY_FIELD(location);
	COPY_FIELD(auth);
	COPY_FIELD(setCookie);
	COPY_FIELD(body);
	for (i = 0; r->outHeaders[i]; i++)
		COPY_FIELD(outHeaders[i]);
	return r->buffer;
}

char*
addStringToBuffer(Request* r, char* string, int n)
{
	char* s;

	if (!r->buffer) return NULL;
	while (r->buffer->count + n > r->buffer->size) {
		Buffer* old = r->buffer;
		debug("addStringToBuffer(%s): doubling %sbuffer",
			  string, old == currentBuffer ? "current " : "");
		if (!copyRequestBuffer(r, r->buffer, YES)) return NULL;
		if (old != currentBuffer) freeBuffer(old);
		if (r->buffer->count + n > r->buffer->size)
			message(MESSAGE, "wow! doubling is insufficient for %d bytes: %s", n, string);
		else
			break;
	}

	s = r->buffer->buffer + r->buffer->start;
	if ((r->buffer->count - r->buffer->start) > 0) // pipelined
		memmove(s + n, s, r->buffer->count - r->buffer->start);

	memcpy(s, string, n);
	r->buffer->count += n;
	r->buffer->start += n;

	return s;
}



char*
getLine(Request* r)
{
	if (!r || r->fd < 0) {
		message(ERROR, "getLine for bad request: %s", r ? "not connected" : "NULL");
		return NULL;
	}

	while (1) {
		int cnt;
		if (r->buffer->count - r->buffer->start > 0) {
			char* line = r->buffer->buffer + r->buffer->start;
			char* eol;
			if ((eol = memchr(line, '\n', r->buffer->count - r->buffer->start))) {
				r->buffer->start = eol + 1 - r->buffer->buffer;
				*eol-- = 0;
				while (eol >= line && isspace(*eol)) *eol-- = 0;
				return line;
			}
		}

		if (r->buffer->count == r->buffer->size)  {
			if (r->buffer->size < 16000) {
				Buffer* old = r->buffer;
				debug("getLine(): doubling %sbuffer", old == currentBuffer ? "current " : "");
				if (!copyRequestBuffer(r, r->buffer, YES)) return NULL;
				if (old != currentBuffer) freeBuffer(old);
			} else {
				char s[NORM_STR];
				int n;
				if (requestsFD < 0) requestsFD = open("logs/requests", O_CREAT|O_APPEND|O_WRONLY, 0644);
				n = snprintf(s, NORM_STR, "\n%02d:%02d:%02d:%s (buffer overflow)",
							 now_tm.tm_hour, now_tm.tm_min, now_tm.tm_sec,
							 ip2string(r->ip));
				write(requestsFD, s, n);
				write(requestsFD, r->buffer->buffer, r->buffer->count);
				message(MESSAGE, "buffer overflow while reading from %s", ip2string(r->ip));
				return NULL;
			}
		}

		errno = 0;
		cnt = read(r->fd, r->buffer->buffer + r->buffer->count, r->buffer->size - r->buffer->count);
		debug("%d> reading %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
#if DEBUGLOG
		if (config->log && cnt >= 0) {
			char s[NORM_STR];
			int n;
			if (requestsFD < 0) requestsFD = open("logs/requests", O_CREAT|O_APPEND|O_WRONLY, 0644);
			n = snprintf(s, NORM_STR, "\n%02d:%02d:%02d:%s%c ",
						 now_tm.tm_hour, now_tm.tm_min, now_tm.tm_sec,
						 ip2string(r->ip), r->buffer->count ? '+' : '>'
						);
			write(requestsFD, s, n);
			write(requestsFD, r->buffer->buffer + r->buffer->count, cnt);
		}
#endif
		if (cnt <= 0) {
			if (errno == EWOULDBLOCK) return NULL;

			if (cnt < 0 && errno != ECONNRESET) message(SYS|ERROR, "cannot read from %s", ip2string(r->ip));
			return NULL;
		}
		r->buffer->count += cnt;
	}
}

char*
getStatusLine(int code)
{
	switch (code) {
	case 100: return "Continue";
	case 101: return "Switching Protocols";

	case 200: return "OK";
	case 201: return "Created";
	case 202: return "Accepted";
	case 203: return "Non-Authoritative Information";
	case 204: return "No Content";
	case 205: return "Reset Content";
	case 206: return "Partial Content";

	case 300: return "Multiple Choices";
	case 301: return "Moved Permanently";
	case 302: return "Moved Temporarily";
	case 303: return "See Other";
	case 304: return "Not Modified";
	case 305: return "Use Proxy";

	case 400: return "Bad Request";
	case 401: return "Unauthorized";
	case 402: return "Payment Required";
	case 403: return "Forbidden";
	case 404: return "Not Found";
	case 405: return "Method Not Allowed";
	case 406: return "Not Acceptable";
	case 407: return "Proxy Authentication Required";
	case 408: return "Request Timeout";
	case 409: return "Conflict";
	case 410: return "Gone";
	case 411: return "Length Required";
	case 412: return "Precondition Failed";
	case 413: return "Request Entity Too Large";
	case 414: return "Request-URI Too Long";
	case 415: return "Unsupported Media Type";

	case 500: return "Internal Server Error";
	case 501: return "Not Implemented";
	case 502: return "Bad Gateway";
	case 503: return "Service Unavailable";
	case 504: return "Gateway Timeout";
	case 505: return "HTTP Version Not Supported";
	default:  return "";
	}
}

int
fillHeaders(Request* r, char* buffer, int size)
{
	int l = 0, i;
	char* status = getStatusLine(r->status);
	char* b = buffer;

	debug("%d> fillHeaders()", r->fd);
	size -= 2;

	l = catstr(b, size,
			   "HTTP/", itoa(r->major), ".", itoa(r->minor), " ",
			   itoa(r->status), " ", status, RN,
			   serverTokens(),
			   "Date: ", now_str, RN, NULL);

	if (r->keepalive) {
		if (r->minor == 0)
			l += catstr(b + l, size - l, "Connection: Keep-Alive" RN, NULL);
	} else if (r->minor) {
		l += catstr(b + l, size - l, "Connection: Close" RN, NULL);
	}

	if (r->status == 405)
		l += catstr(b + l, size - l, "Allow: GET, HEAD, POST" RN, NULL);

	if (r->contentType)
		l += catstr(b + l, size - l, "Content-Type: ", r->contentType, RN, NULL);

	if (r->location)
		l += catstr(b + l, size - l, "Location: ", r->location, RN, NULL);

	if (r->rangeTotal)
		l += catstr(b + l, size - l, "Content-Range: bytes ",
					itoa(r->rangeFrom), "-",
					itoa(r->rangeTo), "/",
					itoa(r->rangeTotal), RN,
					NULL);

	if (r->length >= 0)
		l += catstr(b + l, size - l, "Content-Length: ", itoa(r->length), RN, NULL);

	if (r->acceptRanges)
		l += catstr(b + l, size - l, "Accept-Ranges: bytes" RN, NULL);

	if (r->modified > 0)
		l += catstr(b + l, size - l, "Last-Modified: ", time2str(r->modified, HTTP_TIME), RN, NULL);

	if (r->auth)
		l += catstr(b + l, size - l, "WWW-Authenticate: ", r->auth, RN, NULL);

	if (r->expires == EXPIRES_PAST)
		l += catstr(b + l, size - l,
					"Expires: ", past, RN
					"Pragma: no-cache" RN
					"Cache-control: no-cache" RN,
					NULL);
	else if (r->expires == EXPIRES_FUTURE)
		l += catstr(b + l, size - l, "Expires: ", future, RN, NULL);
	else if (r->expires)
		l += catstr(b + l, size - l, "Expires: ", time2str(r->expires, HTTP_TIME), RN, NULL);

	if (r->setCookie)
		l += catstr(b + l, size - l,
					"P3P: policyref=\"/w3c/p3p.xml\", CP=\"UNI\"" RN
					"Set-Cookie: ", r->setCookie, RN,
					NULL);

	for (i = 0; r->outHeaders[i]; i++) {
		int n = strlen(r->outHeaders[i]);
		if (l + n + 2 < size) {
			memcpy(b + l, r->outHeaders[i], n); l += n;
			memcpy(b + l, RN, 2); l += 2;
		}
	}

	memcpy(b + l, RN, 2);
	return l + 2;
}

int
makeErrorReply(Request* r, char* buffer, int size)
{
	int l = 0;
	char* status = getStatusLine(r->status);
	char* b = buffer;

	debug("%d> makeErrorReply()", r->fd);
	l += catstr(b + l, size - l,
				"<html><head><title>[", itoa(r->status) ,"] ", status,
				"</title></head>\n<body><h3>", status, "</h3>\n", NULL);
	switch (r->status) {
	case 301:
	case 302:
		l += catstr(b + l, size - l, "This document has moved <a href=\"",
					escapeLink(r->location), "\">here</a>.", NULL);
		break;
	case 401:
		l += catstr(b + l, size - l, "You need proper authorization to use this resource.", NULL);
		break;
	case 408:
		l += catstr(b + l, size - l, "Time is out for waiting your request.", NULL);
		break;
	case 400:
	case 405:
	case 501:
	case 505:
		l += catstr(b + l, size - l, "Your request was not understood or not allowed by this server.", NULL);
		break;
	case 403:
		l += catstr(b + l, size - l, "Access to this resource has been denied to you.", NULL);
		break;
	case 404:
		if (r->server)
			l += catstr(b + l, size - l, "Requested file <b>", quote(r->uri),
						"</b> could not be found on server <b>", r->host ? : r->server->name,
						"</b>.", NULL);
		else
			l += catstr(b + l, size - l, "Requested virtual server <b>",
						quote(r->host), "</b> could not be found.", NULL);
		break;
	case 412:
		l += catstr(b + l, size - l, "The precondition given in the request evaluated to false.", NULL);
		break;
	case 413:
		l += catstr(b + l, size - l, "The post request is too large.", NULL);
		break;
	case 503:
		l += catstr(b + l, size - l, "The server is temporarily busy. Try again later.", NULL);
		break;
	case 504:
		l += catstr(b + l, size - l, "The upstream server does not response. Try again later.", NULL);
		break;
	default:
		l += catstr(b + l, size - l, "An internal server error has occurred.", NULL);
		break;
	}
	if (r->server && r->server->admin)
		l += catstr(b + l, size - l, "\n<hr>If you beleive this is incorrect contact the site administrator at "
					"<a href=\"mailto:", r->server->admin, "\">",
					r->server->admin, "</a>.",
					NULL);

	l += catstr(b + l, size - l, "\n</body></html>\n", NULL);
	return l;
}

void
logRequest(Request* r)
{
	char* methods[] = {"-", "HEAD", "GET", "POST"};
	char* strings[] = {r->cookie, r->uri, r->referrer, r->userAgent};
	char fwd[17];
	int i;

	if (r->query) r->query[-1] = '?';
	if (r->params) r->params[-1] = ';';

	for (i = 0; i < sizeof(strings)/sizeof(strings[0]); i++) {
		char* c;
		if ((c = strings[i])) while((c = strchr(c, '\t'))) *c++ = ' ';
	}

	if (r->forwardedFor) {
		*fwd = '-';
		strcpy(fwd + 1, ip2string(r->forwardedFor));
	} else {
		*fwd = 0;
	}
	message(ACCESS, "\t%s%s\t%s\t%d\t%s\t%s%s\t%d\t%s\t%s",
			ip2string(r->realIP?:r->ip), fwd, r->cookie?:"-",
			r->status, methods[r->method], r->host?:"-", r->uri?:"-",
			r->sent, r->referrer?:"-", r->userAgent?:"-");
}


void
setRange(Request* r)
{
	if (r->content) return;
	if (!r->acceptRanges || r->length < 0 || r->status != 200) {
		r->acceptRanges = NO;
		r->rangeFrom = 0;
		r->rangeTo = 0;
		return;
	}

	if (r->rangeFrom >= 0 || r->rangeTo >= 0) {
		r->status = 206;
		if (r->rangeFrom < 0) { // bytes=-500
			if (r->rangeTo < r->length) {
				r->rangeTotal = r->length;
				r->length = r->rangeTo;
				r->rangeTo = r->rangeTotal - 1;
				r->rangeFrom = r->rangeTotal - r->length;
				return;
			}
		} else if (r->rangeTo < 0) { // bytes=500-
			if (r->rangeFrom < r->length) {
				r->rangeTotal = r->length;
				r->length = r->rangeTotal - r->rangeFrom;
				r->rangeTo = r->rangeTotal - 1;
				return;
			}
		} else if (r->rangeFrom <= r->rangeTo && r->rangeFrom < r->length) {
			// bytes=500-1000
			r->rangeTotal = r->length;
			if (r->rangeTo >= r->rangeTotal)
				r->rangeTo = r->rangeTotal - 1;
			r->length = r->rangeTo - r->rangeFrom + 1;
			return;
		}
	}
	r->status = 200;
	r->rangeFrom = 0;
	r->rangeTo = r->length - 1;
}

void
http(Request* r)
{
	char* line;
	char buffer[BLOCK_SIZE];
	int watchRead = 0;
	int watchWrite = 0;
#ifdef LOWAT_SUPPORT
	int watermark = 0;
#endif


	currentRequest = r;

	if (!r->buffer) {
		r->buffer = currentBuffer;
		currentBuffer->start = currentBuffer->count = 0;
	}

	while(1) {
		switch (r->state) {
		case UNDEF:
			debug("%d> http: %s", r->fd, states[r->state]);
			message(ERROR, "strange: http in UNDEF state (%d)", r->fd);
			goto end;

		case READ_REQUEST:
			watchRead = 1;

		case KEEP_ALIVE:
			debug("%d> http: %s", r->fd, states[r->state]);

			if (!(line = getLine(r))) {
				if (errno == EWOULDBLOCK) goto end;
				r->state = CLOSE;
				break;
			}
			if (!*line) break; // skip empty lines before request

			r->wasRequest = YES;
			r->start = now;

			if ((r->status = parseRequestLine(line, r))) {
				r->state = NOT_OK;
				break;
			}
			r->state = READ_HEADER;
			//break;

		case READ_HEADER:
			debug("%d> http: %s", r->fd, states[r->state]);
			while (1) {
				if (!(line = getLine(r))) {
					if (errno == EWOULDBLOCK) goto end;
					r->state = CLOSE;
					break;
				}
				if (!*line) {
					int count = r->buffer->count - r->buffer->start;
					r->state = HANDLE_REQUEST;

					if (r->method == POST) {
						if (r->inContentLength < 0) {
							message(MESSAGE, "warning: post without content-length [%s %s%s]", ip2string(r->ip), r->host?:"-", r->uri);
							r->status = 400;
							r->state = NOT_OK;
						} else if (r->inContentLength > config->maxPostSize) {
							message(MESSAGE, "warning: post size too big (%d > %d) [%s %s%s]", r->inContentLength, config->maxPostSize, ip2string(r->ip), r->host?:"-", r->uri);
							r->status = 413;
							r->state = NOT_OK;
						} else if (r->inContentLength) {
							debug("creating postBuffer(%d in %d)", r->inContentLength, count);
							r->postBuffer = newBuffer(r->inContentLength);
							if (count > r->inContentLength) count = r->inContentLength;
							if (count) {
								int pipelined = r->buffer->count - r->buffer->start - count;
								memcpy(r->postBuffer->buffer, r->buffer->buffer + r->buffer->start, r->postBuffer->count = count);
								if (pipelined)
									memmove(r->buffer->buffer + r->buffer->start,
											r->buffer->buffer + r->buffer->start + count,
											pipelined);
								debug("shrinking buffer by from %d to %d (pipelined: %d)", r->buffer->count, r->buffer->count - count, pipelined);
								r->buffer->count -= count;
							}

							if (count < r->inContentLength)
								r->state = READ_POST;
						}
					}

					break;
				}

				if ((r->status = parseHeaderLine(line, r))) {
					r->state = NOT_OK;
					break;
				}
			}
			break;

		case READ_POST: {
			int cnt;
			Buffer* buffer = r->postBuffer;
			debug("%d> http: %s", r->fd, states[r->state]);

			errno = 0;
			cnt = read(r->fd, buffer->buffer + buffer->count, buffer->size - buffer->count);
			debug("%d> reading %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
			if (cnt <= 0) {
				if (errno == EWOULDBLOCK) goto end;
				if (cnt < 0 && errno != ECONNRESET) message(SYS|ERROR, "cannot read from %s", ip2string(r->ip));
				r->state = CLOSE;
				break;
			}
			buffer->count += cnt;

			if (buffer->count == buffer->size) r->state = HANDLE_REQUEST;

			break;
		}

		case HANDLE_REQUEST: {
			debug("%d> http: %s", r->fd, states[r->state]);

			watchRead = -1;

			r->pipelined = r->buffer->count - r->buffer->start;

			totalRequests++;

			r->config = config;
			config->active++;

			if ((r->status = handleRequest(r))) {
				r->state = NOT_OK;
				break;
			}
			if (r->state == HANDLING) goto end;

			r->state = PREPARE_HEADER;
			break;
		}

		case HANDLING:
			debug("%d> http: %s", r->fd, states[r->state]);
			goto end;


		case PREPARE_HEADER: {
			debug("%d> http: %s", r->fd, states[r->state]);

			if (r->status == 0) r->status = 200;
			if (r->length < 0 && !(r->method == HEAD || r->status == 304))
				r->keepalive = NO;

			if (!r->expires && r->alias && r->alias->expires)
				r->expires =
					r->alias->expires == EXPIRES_PAST ? EXPIRES_PAST :
					r->alias->expires == EXPIRES_FUTURE ? EXPIRES_FUTURE :
					now + r->alias->expires;

			setRange(r);

			r->state = WRITE_HEADER;
			watchWrite = 1;

			if (config->sendBuffer)
				if (setsockopt(r->fd, SOL_SOCKET, SO_SNDBUF, &config->sendBuffer, sizeof(config->sendBuffer)))
					message(SYS|ERROR, "cannot setsockopt(SO_SNDBUF=%d)", config->sendBuffer);
			//break;
		}

		case WRITE_HEADER: {
			char header[MEDIUM_TEXT];
			int count = 0;
			int l;
			int length;
			int body = 0;
			struct iovec buffers[2];

			debug("%d> http: %s", r->fd, states[r->state]);

			l = fillHeaders(r, header, MEDIUM_TEXT);

			length = (r->method == HEAD || r->status == 304) ? 0 : r->length;

			l -= r->offset;
			buffers[0].iov_base = header + r->offset;
			buffers[0].iov_len = l;
			buffers[1].iov_len = 0;

			if (r->content) {
				if (r->contentType && r->alias && r->alias->filter &&
					r->alias->filter->type &&
					!memcmp(r->contentType, r->alias->filter->type, r->alias->filter->typeLength)) {
					r->filter.state = 1;
					r->filter.pos = 0;
					r->filter.quote = 0;
					r->alias->filter->handler(r, NO);
				} else {
					r->filtered = r->received;
				}
				buffers[1].iov_base = r->content;
				buffers[1].iov_len = body = r->filtered;
			} else if (length > 0) {
				if (r->body) {
					buffers[1].iov_base = r->body + r->rangeFrom;
					buffers[1].iov_len = body = length;
				} else if (r->file >= 0) {
#if TRANSFER_MODE == TRANSFER_PLAIN
					if (r->rangeFrom) lseek(r->file, r->rangeFrom, SEEK_SET);
					if ((body = read(r->file, buffer, BLOCK_SIZE)) < 0) {
						message(SYS|ERROR, "cannot read %s%s", r->server ? r->server->name : "?", r->uri);
						r->state = CLOSE;
						break;
					}
					buffers[1].iov_base = buffer;
					buffers[1].iov_len = body;

#elif TRANSFER_MODE == TRANSFER_SENDFILE
#if __linux__
					int onoff = 1;
					if (setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
						message(SYS|ERROR, "cannot setsockopt(TCP_CORK=1)");
#elif __FreeBSD__
					off_t sbytes = -1;
					struct sf_hdtr header;
					header.headers = buffers;
					header.hdr_cnt = 1;
					header.trailers = NULL;
					header.trl_cnt = 0;

					sendfile(r->file, r->fd, r->rangeFrom, length, &header, &sbytes, 0);
					if (sbytes < buffers[0].iov_len) {
						debug("%d> written %d bytes (%s)", r->fd, sbytes, errno?strerror(errno):"OK");
						if (errno == EWOULDBLOCK) {
							debug("blocked header write (%d bytes)", buffers[0].iov_len);
							goto end;
						}
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot sendfile to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					count = sbytes;
					body = length;

#endif

#elif TRANSFER_MODE == TRANSFER_MMAP
					r->mmap = mmap(NULL, r->fileSize, PROT_READ, MAP_PRIVATE, r->file, 0);
					freeFD(r->file);
					r->file = -1;
					if (r->mmap == MAP_FAILED) {
						message(SYS|ERROR, "cannot mmap %s%s", r->server ? r->server->name : "?", r->uri);
						r->state = CLOSE;
						break;
					}
					buffers[1].iov_base = r->mmap + r->rangeFrom;
					buffers[1].iov_len = body = length;
#else
#error unknown TRANSFER_MODE
#endif
				}
			} else {
				r->body = NULL;
			}

			l += body;

			errno = 0;
			if (buffers[1].iov_len) {
				if ((count = writev(r->fd, buffers, 2)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
			} else if (!count) {
				if ((count = write(r->fd, buffers[0].iov_base, buffers[0].iov_len)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
			}
			debug("%d> written %d bytes (%s)", r->fd, count, errno?strerror(errno):"OK");

			if (r->server) r->server->sent += count;

			if (count < l) {
				if (count < buffers[0].iov_len) {
					debug("incomplete header write (%d/%d bytes)", count, buffers[0].iov_len);
					r->offset += count;
					goto end;
				}
				body -= l - count;
			}

			r->sent += body;
			r->length = length;

			if (body == r->length) {
				r->state = DONE;
				break;
			}
			r->offset = 0;

#ifdef LOWAT_SUPPORT
			watermark = count;
#endif

			if (r->content) {
				r->state = WRITE_CONTENT;
				if (r->length < 0 || r->contentSize < r->length) {
					if (r->received > r->sent) memmove(r->content, r->content + r->sent, r->received - r->sent);
					r->offset = r->sent;
				}
			} else {
				if (r->body && !r->alloc &&
					!(r->body >= r->buffer->buffer &&
					  r->body < r->buffer->buffer + r->buffer->count))
					r->body = r->alloc = memcpy(MALLOC(r->length + r->rangeFrom), r->body, r->length + r->rangeFrom);
				r->state = WRITE_BODY;
			}

			if (r->content) break;

#if __linux__ && TRANSFER_MODE == TRANSFER_SENDFILE
			if (r->file >= 0) break;
#endif

			goto end;
		}

		case WRITE_BODY:
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->body) {
				int written;
				errno = 0;
				if ((written = write(r->fd, r->body + r->sent + r->rangeFrom, r->length - r->sent)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
				r->sent += written;
				if (r->server) r->server->sent += written;
				if (r->sent == r->length) {
					r->state = DONE;
					break;
				}
				goto end;
			} else if (r->file >= 0) {
#if TRANSFER_MODE == TRANSFER_PLAIN
				lseek(r->file, r->sent + r->rangeFrom, SEEK_SET);
				while (1) {
					int written;
					int readed = r->length - r->sent;
					errno = 0;
					if (readed > BLOCK_SIZE) readed = BLOCK_SIZE;
					if ((readed = read(r->file, buffer, readed)) < 0) {
						message(SYS|ERROR, "cannot read localfile");
						r->state = CLOSE;
						break;
					}
					if (readed == 0) {
						message(SYS|ERROR, "unexpected end of file");
						r->state = CLOSE;
						break;
					}
					if ((written = write(r->fd, buffer, readed)) <= 0) {
						debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
						if (errno == EWOULDBLOCK) goto end;
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");

					r->sent += written;
					if (r->server) r->server->sent += written;
					if (written < readed) goto end;
					if (r->sent == r->length) {
						r->state = DONE;
						break;
					}
				}
#elif TRANSFER_MODE == TRANSFER_SENDFILE
				errno = 0;
				{
					off_t offset = r->sent + r->rangeFrom;
					int sent;
#if __FreeBSD__
					off_t sbytes = 0;
#endif

					if (r->sent && r->alias && r->alias->throttle) {
						if (!r->throttled) {
							debug("%d> throttle: go to sleep", r->fd);
							watchWrite = -1;
							r->throttled = YES;
							goto end;
						}
						debug("%d> throttle: awaked", r->fd);
						watchWrite = 1;
						r->throttled = NO;
					}

#if __linux__
					sent = sendfile(r->fd, r->file, &offset, r->length - r->sent);

					if (r->sent == 0 && !r->pipelined) {
						int onoff = 0;
						if (setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
							message(SYS|ERROR, "cannot setsockopt(TCP_CORK=0)");
					}
#elif __FreeBSD__
					sendfile(r->file, r->fd, offset, r->length - r->sent, NULL, &sbytes, 0);
					if (sbytes > 0) sent = sbytes;
					else sent = -1;
#endif

					debug("%d> sendfile %d bytes (%s)", r->fd, sent, errno?strerror(errno):"OK");
					if (sent > 0) {
						r->sent += sent;
						if (r->server) r->server->sent += sent;
						if (r->sent == r->length) {
							r->state = DONE;
							break;
						}
					}
					if (sent <= 0 || errno) {
						if (errno == EWOULDBLOCK) goto end;
						if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot sendfile to %s", ip2string(r->ip));
						r->state = CLOSE;
						break;
					}
					goto end;
				}
#elif TRANSFER_MODE == TRANSFER_MMAP
				message(MESSAGE, "strange, r->file in mmap mode");
#else
#error unknown TRANSFER_MODE
#endif
#if TRANSFER_MODE == TRANSFER_MMAP
			} else if (r->mmap) {
				int written;
				errno = 0;
				if ((written = write(r->fd, r->mmap + r->rangeFrom + r->sent, r->length - r->sent)) <= 0) {
					debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) goto end;
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> written %d bytes (%s)", r->fd, written, errno?strerror(errno):"OK");
				r->sent += written;
				if (r->server) r->server->sent += written;
				if (r->sent == r->length) {
					r->state = DONE;
					break;
				}
				goto end;
#endif // TRANSFER_MMAP
			} else {
				r->state = DONE;
			}
			break;

		case WRITE_CONTENT:
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->filter.state) r->alias->filter->handler(r, NO);
			else r->filtered = r->received;
			while (1) {
				int count = r->filtered - r->sent;
				if (count == 0) {
					if (r->proxy && r->proxy->state == WAIT_MEMORY) proxy(r->proxy);
					watchWrite = -1;
					r->state = WAIT_DATA;
					break;
				}
				debug("WRITE_CONTENT: size: %d, off: %d, r: %d, s: %d", r->contentSize, r->offset, r->filtered, r->sent);
				if ((count = write(r->fd, r->content + r->sent - r->offset, count)) <= 0) {
					if (errno == EWOULDBLOCK) {
						debug("%d> blocked write", r->fd);
						if (r->proxy && r->proxy->state == WAIT_MEMORY) proxy(r->proxy);
						goto end;
					}
					if (errno != EPIPE && errno != ECONNRESET && errno != ENOTCONN) message(SYS|ERROR, "cannot write to %s", ip2string(r->ip));
					r->state = CLOSE;
					break;
				}
				debug("%d> writing %d bytes", r->fd, count);
				r->sent += count;
				if (r->server) r->server->sent += count;
				if (r->filtered == r->sent && (!r->proxy || r->proxy->state == FINISHED)) {
					if (r->filtered == r->received) {
						r->state = DONE;
						break;
					} else {
						r->alias->filter->handler(r, YES);
						continue;
					}
				}
				if (r->length < 0 || r->offset + r->contentSize < r->length) {
					if (r->received > r->sent) memmove(r->content, r->content + r->sent - r->offset, r->received - r->sent);
					r->offset = r->sent;
				}
			}
			break;


		case WAIT_DATA:
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->filter.state) r->alias->filter->handler(r, NO);
			else r->filtered = r->received;
			if (r->filtered == r->sent) {
				if (!r->proxy || r->proxy->state == FINISHED) {
					if (r->filtered < r->received) {
						r->alias->filter->handler(r, YES);
						r->state = WRITE_CONTENT;
						watchWrite = 1;
					} else	{
						r->state = DONE;
					}
				} else {
					goto end;
				}
			} else {
				r->state = WRITE_CONTENT;
				watchWrite = 1;
			}
			break;

		case DONE: {
#if TRANSFER_MODE == TRANSFER_SENDFILE && __linux__
			int wasFile = r->file >= 0 ? YES : NO;
#endif
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->mmap) {
				munmap(r->mmap, r->fileSize);
				r->mmap = NULL;
			}
			if (r->file >= 0 && freeFD(r->file) < 0)
				message(SYS|ERROR, "cannot close localfile");
			r->file = -1;
			if (r->postBuffer) r->postBuffer = freeBuffer(r->postBuffer);
			if (r->alloc) r->alloc = FREE(r->alloc);
			if (r->proxy) r->proxy = closeProxy(r->proxy);
			if (r->contentSize) {
				r->content = FREE(r->content);
				r->contentSize = 0;
			}

			if (r->keepalive) {
				unsigned ip = r->ip;
				int fd = r->fd;
				Buffer* buffer = r->buffer;
				int pipelined = buffer->count - buffer->start;
				char wRead, wWrite;
				char* s;
				Request *prev, *next;

				int l = r->wasRequest ?
					r->alias ? r->alias->log :
					r->server ? r->server->log :
					config->log : NO;
				if (l == OK || (l == ERRORS && r->status >= 400)) logRequest(r);

				for (s = buffer->buffer + buffer->start; pipelined; s++, pipelined--)
					if (*s != '\r' && *s != '\n') break;

				if (pipelined) {
					memmove(buffer->buffer, s, pipelined);
					buffer->count = pipelined;
					buffer->start = 0;
				} else if (buffer != currentBuffer) {
					buffer = freeBuffer(buffer);
				} else {
					buffer = NULL;
				}

				if (r->config && (--(r->config->active)) == 0 && r->config != config)
					freeConfig(r->config);

				wRead = r->watchRead;
				wWrite = r->watchWrite;
				prev = r->prev;
				next = r->next;

				memset(r, 0, sizeof(Request));

				r->type = REQUEST;
				r->state = KEEP_ALIVE;
				r->fd = fd;
				r->major = 1;
				r->rangeFrom = r->rangeTo = r->file =
					r->length = r->inContentLength = -1;
				r->ip = ip;
				r->last = now + config->keepAliveTimeout;
				r->start = now;
				r->buffer = buffer;
				r->watchRead = wRead;
				r->watchWrite = wWrite;

				r->prev = prev;
				r->next = next;

				watchRead = 1;
				watchWrite = -1;
#ifdef LOWAT_SUPPORT
				watermark = 0;
#endif

				if (!pipelined) {
#if TRANSFER_MODE == TRANSFER_SENDFILE && __linux__
					int onoff = 0;
					if (wasFile && setsockopt(r->fd, SOL_TCP, TCP_CORK, &onoff, sizeof(onoff)))
						message(SYS|ERROR, "cannot setsockopt(TCP_CORK=0)");
#endif
					goto end;
				}

				break;
			}

			r->state = CLOSE;
			//break;
		}

		case CLOSE: {
			int l = r->wasRequest ?
				r->alias ? r->alias->log :
				r->server ? r->server->log :
				config->log :
				NO;

			debug("%d> http: %s", r->fd, states[r->state]);

			if (l == YES || (l == ERRORS && r->status >= 400)) logRequest(r);

			decIP(r->ip);
			if (r->mmap) munmap(r->mmap, r->fileSize);
			if (r->file >= 0 && freeFD(r->file) < 0)
				message(SYS|ERROR, "cannot close localfile");
			if (r->fd >= 0) {
				fds[r->fd].request = NULL;
				// some weirdness: if any bytes in socket buffer
				// connection reset appears,
				// quick-fix: read them all
				if (r->method == POST) while(read(r->fd, buffer, BLOCK_SIZE) > 0);
				if (close(r->fd) < 0)
					message(SYS|ERROR, "cannot close %d", r->fd);
			}
			if (r->buffer && r->buffer != currentBuffer) freeBuffer(r->buffer);
			if (r->postBuffer) freeBuffer(r->postBuffer);
			if (r->alloc) FREE(r->alloc);
			if (r->proxy) closeProxy(r->proxy);
			if (r->contentSize) FREE(r->content);

			if (r->config && (--(r->config->active)) == 0 && r->config != config)
				freeConfig(r->config);

			if (r->next) r->next->prev = r->prev;
			if (r->prev) r->prev->next = r->next;
			if (requests == r) requests = r->next;

			memset(r, 0, sizeof(Request));

			r->next = freeRequests;
			freeRequests = r;
			nRequests--;
			r = NULL;
			goto end;
		}

		case NOT_OK:
			debug("%d> http: %s", r->fd, states[r->state]);
			if (r->status == 400) {
				char s[NORM_STR];
				int n;
				if (requestsFD < 0) requestsFD = open("logs/requests", O_CREAT|O_APPEND|O_WRONLY, 0644);
				n = snprintf(s, NORM_STR, "\n%02d:%02d:%02d:%s (400 error) ",
							 now_tm.tm_hour, now_tm.tm_min, now_tm.tm_sec,
							 ip2string(r->ip));
				write(requestsFD, s, n);
				write(requestsFD, r->buffer->buffer, r->buffer->count);
			}
			if (r->status != 304 && !r->body) {
				if (r->method != HEAD)
					r->length = makeErrorReply(r, r->body = buffer, BLOCK_SIZE);
				r->contentType = "text/html";
			}
			if (r->status >= 400) r->keepalive = NO;
			r->state = PREPARE_HEADER;
			break;
		}
	}

end:

	if (r) {
#if MODEL == MODEL_KEVENT
		if (watchRead > 0) {
			if (!r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_ADD, 0, 0);
				r->watchRead = YES;
			}
		} else if (watchRead < 0) {
			if (r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_DELETE, 0, 0);
				r->watchRead = NO;
			}
		}

		if (watchWrite > 0) {
			if (!r->watchWrite) {
#ifdef LOWAT_SUPPORT
				watermark = watermark ?
					config->lowWatermark ? :
					watermark >= 3000 + 2048 ? watermark - 3000 :
					0 : 0;
				if (watermark) {
					debug("%d> setting low watermark to %d", r->fd, watermark);
					watchEvent(r->fd, EVFILT_WRITE, EV_ADD, NOTE_LOWAT, watermark);
				} else {
					watchEvent(r->fd, EVFILT_WRITE, EV_ADD, 0, 0);
				}
#else
				watchEvent(r->fd, EVFILT_WRITE, EV_ADD, 0, 0);
#endif
				r->watchWrite = YES;
			}
		} else if (watchWrite < 0) {
			if (r->watchWrite) {
				watchEvent(r->fd, EVFILT_WRITE, EV_DELETE, 0, 0);
				r->watchWrite = NO;
			}
		}
#elif MODEL == MODEL_POLL
		if (watchRead) r->watchRead = watchRead > 0 ? YES : NO;
		if (watchWrite) r->watchWrite = watchWrite > 0 ? YES : NO;
#endif //MODEL

		if (r->buffer == currentBuffer) {
			if (currentBuffer->count) copyRequestBuffer(r, currentBuffer, NO);
			else r->buffer = NULL;
		}

#if defined(LOWAT_SUPPORT) && MODEL != MODEL_KEVENT
		watermark = watermark ?
			config->lowWatermark ? :
			watermark >= 3000 + 2048 ? watermark - 3000 :
			0 : 0;
		if (watermark >= 3000 + 2048) {
			watermark -= 3000;
			debug("%d> setting low watermark to %d", r->fd, watermark);
			if (setsockopt(r->fd, SOL_SOCKET, SO_SNDLOWAT, &watermark, sizeof(watermark)))
				message(SYS|ERROR, "cannot setsockopt(SO_SNDLOWAT=%d)", watermark);
		}
#endif

		if (r->state < READ_HEADER);
		else if (r->state < HANDLE_REQUEST) r->last = now + config->timeout;
		else if (r->throttled) r->last = now + 1 + (nRequests & 1);
		else r->last = now + (r->alias ? r->alias->timeout : r->server ? r->server->timeout : config->timeout);
	}

	currentRequest = NULL;
}


#define split_sec(sec, h, m, s) {\
	h = sec/60/60;\
	m = (sec%(60*60))/60;\
	s = (sec%60);\
	}

int
handleStatusRequest(Request* r)
{
	char* body;
	int real = 0, system = 0, user = 0;
	int h, m, s, l;
	time_t finish;
	struct rusage ru;
	Listener* listener;
	Server* server;
	int size = r->query ? HUGE_TEXT : MEDIUM_TEXT;
	char show;
	SimpleList* extension;

	unsigned sumCount = 0, sumRequestsLast = 0, sumSentLast = 0;
	unsigned sumRequestsToday = 0, sumDroppedS = 0, sumDroppedC = 0;
	unsigned sumActive = 0;
	long long sumSentToday = 0;


	if (!(body = MALLOC(size))) return 500;
	l = SNPRINTF(body, size,
				 "<html><head><title>%s</title></head>\n<body><h3>%s</h3><pre>\n"
				 "Server version: %s, built: %s\n"
				 "Current time: %s\n"
				 "Started at    %s\n"
				 "Configured at %s\n",
				 "Server status", "Server status",
				 VERSION, built,
				 time2str(now, LOCAL_TIME),
				 time2str(start, LOCAL_TIME),
				 time2str(configured, LOCAL_TIME)
				);

	finish = time(NULL);
	real = finish - start;
	l += SNPRINTF(body + l, size - l, "Elapsed time:\n");
	split_sec(real, h, m, s);
	l += SNPRINTF(body + l, size - l, "  real:  %4d:%02d:%02d\n", h, m, s);

	if (getrusage(RUSAGE_SELF, &ru)) {
		message(SYS|ERROR, "cannot getrusage");
	} else {
		user = ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec < 500000 ? 0 : 1);
		system = ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec < 500000 ? 0 : 1);
		split_sec(user, h, m, s);
		l += SNPRINTF(body + l, size - l, "  user:   %3d:%02d:%02d (%3.1f%%)\n", h, m, s, user*100.0/real);

		split_sec(system, h, m, s);
		l += SNPRINTF(body + l, size - l, "  system: %3d:%02d:%02d (%3.1f%%)\n", h, m, s, system*100.0/real);
	}

	l += SNPRINTF(body + l, size - l, "\nLast 5 seconds: %s acc/sec, %s req/sec, CPU usage: %3.1f%% user, %3.1f%% system\n", itos((lastAccepts*2+5)/10), itos((lastRequests*2+5)/10), lastUser/50.0, lastSys/50.0);

	m = strlen("99 listeners");
	for(listener = config->listeners; listener; listener = listener->next) {
		s = strlen(listener->addr.s_addr ? inet_ntoa(listener->addr): "*") + 1 + strlen(itoa(listener->port));
		if (s > m) m = s;
	}
	l += SNPRINTF(body + l, size - l,
				  "\n%-*s %5s %10s %10s %10s\n",
				  m + 2, "Listener", "acc/s", "acc/today", "srv-drops", "cli-drops");

	for(listener = config->listeners; listener; listener = listener->next) {
		char name[SHORT_STR];
		SNPRINTF(name, SHORT_STR, "%s:%d",
				 listener->addr.s_addr ? inet_ntoa(listener->addr) : "*",
				 listener->port);

		if (listener->error)
			l += SNPRINTF(body + l, size - l, "  %s%-*s %-38s%s\n",
						  listener->next || !sumCount ? "" : "<u>",
						  m, name,
						  listener->error,
						  listener->next || !sumCount ? "" : "</u>"
						 );
		else
			l += SNPRINTF(body + l, size - l, "  %s%-*s %5s %10s %10s %10s%s\n",
						  listener->next || !sumCount ? "" : "<u>",
						  m, name,
						  itos((listener->acceptsLast*2+5)/10),
						  itos(listener->acceptsToday),
						  itos(listener->droppedS),
						  itos(listener->droppedC),
						  listener->next || !sumCount ? "" : "</u>"
						 );
		sumCount++;
		sumRequestsLast += listener->acceptsLast;
		sumRequestsToday += listener->acceptsToday;
		sumDroppedS += listener->droppedS;
		sumDroppedC += listener->droppedC;
	}
	if (sumCount > 1)
		l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %10s %10s %10s\n",
					  sumCount,
					  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "listeners",
					  itos((sumRequestsLast*2+5)/10),
					  itos(sumRequestsToday),
					  itos(sumDroppedS),
					  itos(sumDroppedC)
					 );

	sumCount = 0; sumRequestsLast = 0; sumRequestsToday = 0;

	m = strlen("99 servers");
	for(server = config->servers; server; server = server->next) {
		s = strlen(server->name);
		if (s > m) m = s;
	}
	l += SNPRINTF(body + l, size - l,
				  "\n%-*s %5s %7s %11s %11s\n",
				  m + 2, "Server", "req/s", "bytes/s", "req/today", "bytes/today");

	for(server = config->servers; server; server = server->next) {
		l += SNPRINTF(body + l, size - l, "  %s%-*s %5s %7s %11s %11s%s\n",
					  server->next || !sumCount ? "" : "<u>",
					  m, server->name,
					  itos((server->requestsLast*2+5)/10),
					  humanSize((server->sentLast*2+5)/10),
					  itos(server->requestsToday),
					  humanSize(server->sentToday),
					  server->next || !sumCount ? "" : "</u>"
					 );
		sumCount++;
		sumRequestsLast += server->requestsLast;
		sumSentLast += server->sentLast;
		sumRequestsToday += server->requestsToday;
		sumSentToday += server->sentToday;
	}

	if (sumCount > 1)
		l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %7s %11s %11s\n",
					  sumCount,
					  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "servers",
					  itos((sumRequestsLast*2+5)/10),
					  humanSize((sumSentLast*2+5)/10),
					  itos(sumRequestsToday),
					  humanSize(sumSentToday)
					 );


	if (config->backends) {
		Backend* backend;
		int m = strlen("99 backends");
		sumCount = 0; sumRequestsLast = 0; sumRequestsToday = 0;

		for(backend = config->backends; backend; backend = backend->next) {
			int n = strlen(backend->name) + 1 + strlen(itoa(backend->port));
			if (n > m) m = n;
		}

		l += SNPRINTF(body + l, size - l, "\n"
					  "%-*s state active max req/s req/today\n", m + 2, "Backend");
		for(backend = config->backends; backend; backend = backend->next) {
			char name[SHORT_STR];
			SNPRINTF(name, SHORT_STR, "%s:%d", backend->name, backend->port);
			l += SNPRINTF(body + l, size - l, "  %s%-*s %5s %6d %3d %5s %9s%s\n",
						  backend->next || !sumCount ? "" : "<u>",
						  m, name, backend->state ? "UP" : "DOWN",
						  backend->nRequests, backend->maxRequests,
						  itos((backend->requestsLast*2+5)/10),
						  itos(backend->requestsToday),
						  backend->next || !sumCount ? "" : "</u>"
						 );
			sumCount++;
			sumActive += backend->nRequests;
			sumRequestsLast += backend->requestsLast;
			sumRequestsToday += backend->requestsToday;
		}

		if (sumCount > 1)
			l += SNPRINTF(body + l, size - l, "  %d %-*s %5s %6s %3s %5s %9s\n",
						  sumCount,
						  m - (sumCount < 10 ? 2 : sumCount < 100 ? 3 : 4), "backends",
						  "",
						  itos(sumActive),
						  "",
						  itos((sumRequestsLast*2+5)/10),
						  itos(sumRequestsToday)
						 );
	}


	show = r->query && strstr(r->query, "requests");
	l += SNPRINTF(body + l, size - l, "\n"
				  "<a href=\"%s%s\">Requests</a> (active/max): %s / %s\n",
				  r->uri, show ? "" : "?requests",
				  itos(nRequests), itos(maxRequests));

	if (show) {
		Request* r;
		if (requests)
			l += SNPRINTF(body + l, size - l,
						  "%-14s %-15s %4s [%5s/%-5s] %s\n",
						  "State", "IP", "Time", "Sent", "Total", "URL");
		for (r = requests; r && size > l; r = r->next) {
			if (r->buffer && r->buffer->start)
				l += SNPRINTF(body + l, size - l,
							  "%-14s %-15s %4s [%5s/%-5s] %s%s%s%s\n",
							  states[r->state], ip2string(r->ip),
							  humanPeriod(now - r->start),
							  humanSize(r->sent),
							  r->length < 0 ? "?" : humanSize(r->length),
							  r->host?:"-", r->uri?:"",
							  r->query?"?":"", r->query?:"");
			else
				l += SNPRINTF(body + l, size - l, "%-14s %-15s %4s\n",
							  states[r->state], ip2string(r->ip),
							  humanPeriod(now - r->start));
		}
	}

	if (size > l) l += getProxyStats(body + l, size - l, r->query && strstr(r->query, "proxy"), r->uri);

	if (size > l) l += getIPStats(body + l, size - l, r->query && strstr(r->query, "ips"), r->uri);

	if (size > l) l += getFilesStats(body + l, size - l, r->query && strstr(r->query, "files"), r->uri);

	if (size > l) l += SNPRINTF(body + l, size - l, "</pre></body></html>\n");


	r->body = r->alloc = body;
	r->length = l < size ? l : size;
	r->expires = r->modified = now;

	for (extension = config->extensions; extension; extension = extension->next) {
		if (extension->name[0] == '*' ||
			strcmp("html", extension->name) == 0) {
			r->contentType = extension->link;
			break;
		}
	}
	return 0;
}


