#include <stdio.h>

#include "common/settings.h"
#include "common/strings.h"
#include "common/misc.h"

#include "config.h"

#define LOW_SIZE 1024


typedef struct IPInfo_ IPInfo;

struct IPInfo_ {
	unsigned ip;
	unsigned active;
	unsigned accepted;
	unsigned dropped;

	IPInfo* next;
	IPInfo* prev;
};

static IPInfo** ips = NULL;
static IPInfo* freeIPs = NULL;
static int count = 0;
static int maxCount = 0;
static int floods = 0;
static int dropped = 0;


char
addIP(unsigned ip)
{
	unsigned low = ip & (LOW_SIZE - 1);
	IPInfo* i;
	if (!ips && !(ips = MALLOC(LOW_SIZE * sizeof(IPInfo*)))) return OK;

	i = ips[low];
	while(i) {
		if (i->ip == ip) {
			if (++(i->active) > config->connectionsPerIP) {
				i->dropped++;
				i->active--;
				dropped++;
				return FAILED;
			}

			i->accepted++;
			return OK;
		}
		i = i->next;
	}


	if (!(i = freeIPs)) {
		IPInfo* block = MALLOC(1024 * sizeof(IPInfo));
		if (!block) return OK;
		for (i = block; i < block + 1024; i++) {
			i->next = freeIPs;
			freeIPs = i;
		}
		i = freeIPs;
	}
	freeIPs = freeIPs->next;


	if ((i->next = ips[low])) i->next->prev = i;
	ips[low] = i;
	i->prev = NULL;

	i->ip = ip;
	i->dropped = 0;
	i->active = i->accepted = 1;

	if (++count > maxCount) maxCount = count;

	return OK;
}

unsigned
decIP(unsigned ip)
{
	unsigned low = ip & (LOW_SIZE - 1);
	IPInfo* i = ips[low];
	while(i) {
		if (i->ip == ip) {
			if (--(i->active) <= 0) {
				if (i->dropped) {
					floods++;
					message(MESSAGE, "flood from %s detected, dropped: %d, accepted: %d", ip2string(i->ip), i->dropped, i->accepted);
				}

				if (i->next) i->next->prev = i->prev;
				if (i->prev) i->prev->next = i->next;
				if (ips[low] == i) ips[low] = i->next;

				i->next = freeIPs;
				freeIPs = i;
				--count;
			}
			return OK;
		}
		i = i->next;
	}

	return message(MESSAGE, "strange: cannot find %s in ip-list", ip2string(ip));
}


int
getActive(unsigned ip)
{
	IPInfo* i;
	unsigned low = ip & (LOW_SIZE - 1);
	if (!ips) return 0;

	i = ips[low];
	while(i) {
		if (i->ip == ip) return i->active;
		i = i->next;
	}
	return 0;
}


int
getIPStats(char* buffer, int size, char show, char* uri)
{
	int l = SNPRINTF(buffer, size, "\n"
					 "<a href=\"%s%s\">Unique IPs</a> (current/max): %s / %s\n"
					 "Connection floods: %s times, %s accepts was dropped\n",
					 uri, show ? "" : "?ips",
					 itos(count), itos(maxCount), itos(floods), itos(dropped));
	if (show) {
		unsigned low;
		l += SNPRINTF(buffer + l, size - l, "%-16s %s %s %s\n", "IP-address", "Active", "Accepted", "Dropped");
		for (low = 0; low < LOW_SIZE; low++) {
			IPInfo* i = ips[low];
			while(i) {
				if (size > l)
					l += SNPRINTF(buffer + l, size - l, "%-16s %6d %8d %7d\n", ip2string(i->ip), i->active, i->accepted, i->dropped);
				i = i->next;
			}
		}
	}
	return l;
}

void
resetTodayFlood()
{
	maxCount = count;
	floods = dropped = 0;
}
