/* $Id: core.c,v 1.10 2002/05/15 10:28:57 maxim Exp $
 */

#define _GNU_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "common/settings.h"
#include "common/signames.h"
#include "common/strings.h"
#include "common/country.h"
#include "common/misc.h"


#include "core.h"
#include "httpd.h"
#include "flood.h"
#include "proxy.h"
#include "config.h"
#include "access.h"
#include "files.h"
#include "listing.h"

#ifdef COUNTER_ADDON
#include "handle_counter.h"
#endif

#ifdef _3DNEWS_ADDON
#include "common/cmd.h"
#endif

char*
findCtype(char* ext)
{
	int n = strlen(ext) + 1;
	SimpleList* extension = config->extensions;
	while(extension) {
		if (extension->name[0] == '*' ||
            (ext[0] == extension->name[0] &&
			 memcmp(ext, extension->name, n) == 0)) {
			return extension->link;
			break;
		}
		extension = extension->next;
	}
    return NULL;
}

int
findServer(Request* r)
{
	unsigned h;

	if (!(r->server = config->servers)) return FAILED;
	if (!r->host) return OK;

	h = hash(r->host);
	while(r->server) {
		int i;
		if (r->server->name[0] == '*' ||
			(r->server->hash == h &&
			 !strcmp(r->server->name, r->host))) return OK;
		for (i = 0; i < r->server->nOfAlters; i++)
			if (r->server->alterHash[i] == h &&
				!strcmp(r->server->alter[i], r->host)) return OK;
		r->server = r->server->next;
	}
	return FAILED;
}

int
findAlias(Request* r)
{
	char uri[LONG_STR];
    char* ext = NULL;
	int l = strlen(r->uri);
	int cl = 0;
	Alias* alias = r->server ? r->server->aliases : NULL;
	*uri = 0;
	while(alias) {
		int n = alias->length;
		if (l >= n && memcmp(alias->alias, r->uri, n) == 0 &&
			(r->uri[n] == 0 || r->uri[n] == '/')) {
			int i;
			if (alias->exact) {
				if (r->uri[n] == 0 || r->uri[n + 1] == 0) break;
			} else if (alias->nOfCtype) {
				if (!r->contentType) {
					if (!*uri) strtolower(strncpy_(uri, r->uri, LONG_STR));
					if (!ext) {
						for (i = l - 1; i > 0; i--) {
							if (uri[i] == '/') break;
							if (uri[i] == '.') {ext = uri + i + 1; break;}
						}
						if (!ext) {
							if (i == l - 1) ext = "html";
							else ext = "";
						}
					}
					if ((r->contentType = findCtype(ext)))
						cl = strlen(r->contentType);
  				} else {
					cl = strlen(r->contentType);
				}
				if (cl) for (i = 0; i < alias->nOfCtype; i++)
					if (cl >= alias->ctypeLength[i] &&
                        alias->ctype[i][0] == r->contentType[0] &&
						memcmp(alias->ctype[i], r->contentType,
							   alias->ctypeLength[i]) == 0) goto end;
			} else if (alias->nOfTails) {
				if (!*uri) strtolower(strncpy_(uri, r->uri, LONG_STR));
				for (i = 0; i < alias->nOfTails; i++)
					if (l >= alias->tailLength[i] &&
						memcmp(alias->tail[i],
							   uri + l - alias->tailLength[i],
							   alias->tailLength[i]) == 0) goto end;
			} else {
				break;
			}
		}
		alias = alias->next;
	}
end:
	if ((r->alias = alias)) {
		r->path = r->uri + alias->length;
		return OK;
	} else {
		return FAILED;
	}
}


int
handleLocalRequest(Request* r)
{
	char fname[PATH_LEN];
	struct stat* s;
	char ext[SHORT_STR];
	int fd;
	SimpleList* extension;
	int n;
	char* ctype = r->alias->exactLocation ? NULL : r->contentType;
    r->contentType = NULL;

	n = catstr(fname, PATH_LEN, r->alias->location, r->alias->exactLocation ? "" : r->path, NULL);

	if (!n) {
		message(ERROR, "too long filename: %s%s", r->host?:"-", r->uri);
		return 414;
	}

	*ext = 0;
	if (!ctype || r->alias->reverse) {
        int i;
		for (i = n - 1; i > 0; i--) {
			if (fname[i] == '/') break;
			if (fname[i] == '.') {
				strtolower(strncpy_(ext, fname + i + 1, SHORT_STR));
				break;
			}
		}

		if (*ext && r->alias->reverse) {
			SimpleList* rev = r->alias->reverse;
			while(rev) {
				if (ext[0] == rev->name[0] && !strcmp(ext, rev->name)) {
					reverse(fname + n - strlen(ext));
					break;
				}
				rev = rev->next;
			}
		}
	}

	if ((fd = getFD(fname, n)) < 0) {
		message(SYS|ERROR, "cannot open %s", fname);
		return
			errno == ENOENT ? 404 :
			errno == ENOTDIR ? 404 :
			errno == EACCES ? 403 :
			errno == ENOMEM ? 503 :
			errno == EMFILE ? 503 :
			errno == ENFILE ? 503 :
			500;
	}

	if (!(s = getFDStat(fd))) {
		freeFD(fd);
		message(ERROR, "strange: cannot find stat for fd %d:%s", fd, fname);
		return 500;
	}

	if (S_ISDIR(s->st_mode)) {
		SimpleList* index;
        char* dot;
		freeFD(fd);
		if (fname[n - 1] != '/') {
			char location[LONG_STR];
			int n = catstr(location, LONG_STR, "http://",
						   r->host?:r->server->name, escape(r->uri, NO), "/",
						   NULL);
			if (!n) return 414;
			if (r->params)
				n += catstr(location + n, LONG_STR - n, ";", r->params, NULL);
			if (r->query)
				n += catstr(location + n, LONG_STR - n, "?", r->query, NULL);
			if (!(r->location = addStringToBuffer(r, location, n + 1)))
				return 414;
			return 301;
		}

		index = r->alias->indexes;
		while(index) {
			int l = strlen(index->name);
			if (n + l >= PATH_LEN) return 414;
			memcpy(fname + n, index->name, l + 1);
			if ((fd = getFD(fname, n + l)) >= 0) break;
			index = index->next;
		}
		if (!index) {
			if (r->alias->listing) {
				fname[n] = 0;
				if (r->ims && r->ims >= s->st_mtime) return 304;
				if (r->iums && r->iums < s->st_mtime) return 412;
				r->modified = s->st_mtime;
				return showListing(r, fname);
			} else {
				message(SYS|ERROR, "cannot open %s", fname);
				return errno == EACCES ? 403 :
					errno == ENOENT ? 404 :
					errno == ENOTDIR ? 404 :
					errno == ENOMEM ? 503 :
					errno == EMFILE ? 503 :
					errno == ENFILE ? 503 :
					500;
			}
		}
        ctype = NULL;
		if ((dot = strrchr(index->name, '.'))) strtolower(strncpy_(ext, dot + 1, SHORT_STR));
		else *ext = 0;
		if (!(s = getFDStat(fd))) {
			freeFD(fd);
			message(ERROR, "strange: cannot find stat for fd %d:%s", fd, fname);
			return 500;
		}
	}

	if (r->ims && r->ims >= s->st_mtime) {
		freeFD(fd);
		return 304;
	}
	if (r->iums && r->iums < s->st_mtime) {
		freeFD(fd);
		return 412;
	}
	if (r->ifRange && r->ifRange < s->st_mtime)
		r->rangeFrom = r->rangeTo = -1;

	if (ctype) {
		r->contentType = ctype;
	} else {
		extension = config->extensions;
		if (*ext) {
			n = strlen(ext) + 1;
			while(extension) {
				if (extension->name[0] == '*' ||
					(ext[0] == extension->name[0] &&
					 memcmp(ext, extension->name, n) == 0)) {
					r->contentType = extension->link;
					break;
				}
				extension = extension->next;
			}
		} else {
			while(extension) {
				if (extension->name[0] == '*') {
					r->contentType = extension->link;
					break;
				}
				extension = extension->next;
			}
		}
	}

	r->modified = s->st_mtime;
	r->length = r->fileSize = s->st_size;
	r->acceptRanges = YES;

	if (r->method == HEAD) {
		freeFD(fd);
		return 0;
	}

	r->file = fd;

	return 0;
}


int
handleCGIRequest(Request* r)
{
	return 500;
}


int
handleRedirectRequest(Request* r)
{
	char location[LONG_STR];
	int n = catstr(location, LONG_STR, r->alias->location,
				   r->alias->exactLocation ? "" : escape(r->path, NO), NULL);
	if (!n) return 414;

	if (r->params)
		n += catstr(location + n, LONG_STR - n, ";", r->params, NULL);
	if (r->query)
		n += catstr(location + n, LONG_STR - n, "?", r->query, NULL);

	if (!(r->location = addStringToBuffer(r, location, n + 1)))
		return 414;
	return 302;
}


int
handleEmptyRequest(Request* r)
{
	SimpleList* extension;
	char* dot;
	char ext[SHORT_STR];
	int n;
	char* file = strrchr(r->path, '/');
	if (file) file++;
	else file = r->path;

	if ((dot = strrchr(r->path, '.'))) strtolower(strncpy_(ext, dot + 1, SHORT_STR));
	else if (!r->path[0] || r->path[strlen(r->path) - 1] == '/') memcpy(ext, "html", 5);
	else *ext = 0;

	n = strlen(ext) + 1;
	extension = config->extensions;
	while(extension) {
		if (extension->name[0] == '*' ||
			memcmp(ext, extension->name, n) == 0) {
			r->contentType = extension->link;
			break;
		}
		extension = extension->next;
	}
	r->length = 0;
	return 0;
}

int
handlePixelRequest(Request* r)
{
	static char PixelGIF[] =
		"GIF87a\x01\x00\x01\x00\xF0\x00\x00\x00\x00\x00\x00\x00\x00\x2C"
		"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3B";
	static char BlankGIF[] =
		"GIF89a\x01\x00\x01\x00\x80\xFF\x00\xC0\xC0\xC0\x00"
		"\x00\x00\x21\xF9\x04\x01\x00\x00\x00\x00\x2C\x00\x00"
		"\x00\x00\x01\x00\x01\x00\x00\x02\x02\x44\x01\x00\x3B";

	if (r->query) {
		unsigned color = strtoul(r->query, NULL, 16);
		char* c = (char*)&color;
		char* pixel = addStringToBuffer(r, PixelGIF, sizeof(PixelGIF) - 1);
		pixel[13] = c[2];
		pixel[14] = c[1];
		pixel[15] = c[0];
		r->body = pixel;
		r->length = sizeof(PixelGIF) - 1;
	} else if (r->alias && r->alias->aux[0]) {
		char* pixel = addStringToBuffer(r, PixelGIF, sizeof(PixelGIF) - 1);
		memcpy(pixel + 13, r->alias->aux[0], 3);
		r->body = pixel;
		r->length = sizeof(PixelGIF) - 1;
	} else {
		r->body = BlankGIF;
		r->length = sizeof(BlankGIF) - 1;
	}
	r->contentType = "image/gif";
	r->modified = 1000000000; // 09-Sep-2001
	return 0;
}

#ifdef _3DNEWS_ADDON

int
handle3DNewsRedirectRequest(Request* r)
{
	char location[LONG_STR];
	static int country = -2;
	int n;

	if (country == -2) country = domain2id(r->alias->aux[0]);

	n = catstr(location, LONG_STR,
			   r->alias->aux[(ip2country(r->ip) == country) ? 1 : 2],
			   escape(r->path, NO),
			   NULL);
	if (!n) return 414;

	message(ACCESS, "3D_FILES> %s %s %s",
			ip2string(r->ip),
			id2domain(ip2country(r->ip)),
			location);

	if (!(r->location = addStringToBuffer(r, location, n + 1)))
		return 414;

	return 302;
}

int
handle3DNewsMirrorRequest(Request* r)
{
	char fname[PATH_LEN];
	struct stat s;

	int n = catstr(fname, PATH_LEN, r->alias->location, r->path, NULL);
	if (!n) {
		message(ERROR, "too long filename: %s%s", r->host?:"-", r->uri);
		return 414;
	}

	if (stat(fname, &s) < 0) {
		char location[LONG_STR];
		catstr(location, LONG_STR, r->alias->aux[0], r->path, NULL);
		putCmd("httpd", "getit", "download", fmt("%s\t%s", fname, location));
		if (!(r->location = addStringToBuffer(r, location, n + 1)))
			return 414;
		return 302;
	}
	return handleLocalRequest(r);
}


#endif

int
handleRequest(Request* r)
{
	int result;
	static int recursion = 0;
	char* c;
	char* redirect = NULL;
	debug("%d> handleRequest(%s%s)", r->fd, r->host?:"-", r->uri?:"-");

	if (r->uri[0] != '/') {
		char* slash;
		int l;
		if (memcmp(r->uri, "http://", strlen("http://"))) return 400;
		slash = strchr(r->uri + strlen("http://"), '/');
		if (!slash) return 400;
		l = slash - r->uri - strlen("http://");
		memmove(r->uri, r->uri + strlen("http://"), l);
		r->uri[l] = 0;
		r->host = unifyHostName(r->uri);
		r->uri = slash;
	}
	if ((r->query = strchr(r->uri, '?')))
		*(r->query++) = 0;
	if ((r->params = strchr(r->uri, ';')))
		*(r->params++) = 0;

	unescape(r->uri, NO, NO);

	// compress double-slashes
	c = r->uri;
	while((c = strstr(c, "//"))) strcpy(c, c + 1);

	// compress dotted-paths
	c = r->uri;
	while((c = strstr(c, "/./"))) strcpy(c, c + 2);

	// remove parent-dirs
	while ((c = (strstr(r->uri, "/../")))) {
		if (c == r->uri) {
			r->uri += strlen("/..");
		} else {
			char* slash = c - 1;
			while (*slash != '/') slash--;
			strcpy(slash + 1, c + strlen("/../"));
		}
	}

	if (!(findServer(r) && findAlias(r))) return 404;
	if (r->keepalive) r->keepalive = r->alias->keepalive;
	r->server->requests++;

	if (++recursion > 4) {
		message(ERROR, "recursive(%d) handleRequest for %s%s", recursion, r->host?:"-", r->uri?:"-");
		--recursion;
		return 500;
	}


	if (r->alias->access && (result = checkAccess(r, r->alias->access, &redirect))) {
		--recursion;
		if (result == 302) {
			char location[LONG_STR];
			if (!redirect) {
				message(ERROR, "undefined access redirect for %s%s", r->host?:"-", r->uri?:"-");
				return 500;
			} else if (*redirect == '/') {
				int n = catstr(location, LONG_STR, "http://",
							   r->host?:r->server->name, redirect, NULL);
				if (!n || !(r->location = addStringToBuffer(r, location, n + 1)))
					return 414;
			} else if (!memcmp(redirect, "http://", strlen("http://"))) {
				if (redirect[strlen(redirect) - 1] == '/') {
					int n = catstr(location, LONG_STR,
								   redirect, r->path[0] ? r->path + 1 : "", NULL);
					if (!n || !(r->location = addStringToBuffer(r, location, n + 1)))
						return 414;
				} else {
					r->location = redirect;
				}
			} else {
				int n = catstr(location, LONG_STR, "http://",
							   r->host?:r->server->name, r->alias->alias,
							   "/", redirect, NULL);
				if (!n || !(r->location = addStringToBuffer(r, location, n + 1)))
					return 414;
			}
		}
		return result;
	}

	switch (r->alias->type) {
	case LOCAL: result = handleLocalRequest(r); break;
	case CGI: result = handleCGIRequest(r); break;
	case REDIRECT: result = handleRedirectRequest(r); break;
	case PROXY: result = handleProxyRequest(r); break;
	case STATUS: result = handleStatusRequest(r); break;
	case EMPTY: result = handleEmptyRequest(r); break;
	case PIXEL: result = handlePixelRequest(r); break;
#ifdef COUNTER_ADDON
	case COUNTER: result = handleCounterRequest(r); break;
	case COUNTER_LOGO: result = handleCounterLogoRequest(r); break;
	case COUNTER_CLICK: result = handleCounterClickRequest(r); break;
	case COUNTER_VALUES: result = handleCounterValuesRequest(r); break;
#endif
#ifdef _3DNEWS_ADDON
	case _3DNEWS_REDIRECT: result = handle3DNewsRedirectRequest(r); break;
	case _3DNEWS_MIRROR: result = handle3DNewsMirrorRequest(r); break;
#endif
	default:
		result = 500;
		message(ERROR, "bad alias type %d for %s%s", r->alias->type, r->server->name, r->alias->alias);
	}
	--recursion;
	return result;
}


