/* $Id: httpd.h,v 1.5 2002/05/15 10:28:57 maxim Exp $
 */

#ifndef HTTPD_H
#define HTTPD_H

#include <limits.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "common/settings.h"
#include "common/pairs.h"

#include "buffer.h"

#define VERSION "0.7l"

#define TRANSFER_PLAIN	  0
#define TRANSFER_SENDFILE 1
#define TRANSFER_MMAP	  2

#define MODEL_SELECT 0
#define MODEL_POLL	 1
#define MODEL_RTSIG	 2
#define MODEL_KEVENT 3

//
// method of transferring local files
// plain: uses simple read/write,  most compability, low performance
// mmap: maps files and use write, wide compability, good performance
// sendfile:					   linux, freebsd,	 best performance
//

//#define TRANSFER_MODE TRANSFER_PLAIN
#define TRANSFER_MODE TRANSFER_SENDFILE
//#define TRANSFER_MODE TRANSFER_MMAP



//
//	event notification mode
//	select:	   most compability, low performance  (not ready yet)
//	poll:	   wide compability, good performance
//	rt-signal: Linux 2.4+,		 best performance
//	kevent:	   FreeBSD 4.1+,	 best performance
//


#ifndef MODEL
#if __linux__
#define MODEL MODEL_RTSIG

#elif __FreeBSD__
#define MODEL MODEL_KEVENT

#define LOWAT_SUPPORT

#else
//#define MODEL MODEL_SELECT
#define MODEL MODEL_POLL
//#define MODEL MODEL_RTSIG
//#define MODEL MODEL_KEVENT

#endif
#endif

#define EXPIRES_FUTURE INT_MAX
#define EXPIRES_PAST   INT_MIN



//
// maximum number of unkown headers which transmitted to/from backend
//
#define MAX_OTHER_HEADERS 15
#define INITIAL_BUFFER_SIZE 1000



////////////////////////////////////////////////////////////////////


enum {
	UNKNOWN,
	HEAD,
	GET,
	POST
};

#define ERRORS 2


typedef struct SimpleList_struct SimpleList;
typedef struct Config_struct Config;
typedef struct Listener_struct Listener;
typedef struct Server_struct Server;
typedef struct Alias_struct Alias;
typedef struct Cache_struct Cache;
typedef struct Access_struct Access;
typedef struct Filter_struct Filter;
typedef struct Proxy_struct Proxy;
typedef struct CGIServer_struct CGIServer;
typedef struct Request_struct Request;
typedef struct Backend_struct Backend;
typedef struct ProxyRequest_struct ProxyRequest;
typedef struct FilterState_struct FilterState;
typedef struct CGIRequest_struct CGIRequest;
typedef union FD_union FD;

struct Config_struct {
	char* configName;
	char* admin;

	int connections;
	int proxyConnections;
	int cgiConnections;
	int connectionsPerIP;
	int keepAliveTimeout;
	int requestTimeout;
	int timeout;
	int expires;
	int keepalive;
	int charset;
	int srcCharset;
	int log;
	int listing;
	int throttle;
	int maxBuffer;
	int maxPostSize;
	int skipWWW;
	int skipPort;
	int filesCheck;
	int filesKeep;
	int sendBuffer;
	int lowWatermark;
	int detachRoot;
	int serverTokens;

	Listener* listeners;
	SimpleList* extensions;
	SimpleList* mimes;
	SimpleList* indexes;
	SimpleList* reverse;
	Server* servers;
	Access* access;
	Filter* filter;
	Backend* backends;

	int nOfListeners;
	int active;
};

struct Server_struct {
	char* name;
	char* admin;

	int timeout;
	int expires;
	int keepalive;
	int charset;
	int srcCharset;
	int log;
	int listing;
	int throttle;

	unsigned hash;
	int nOfAlters;
	char** alter;
	unsigned* alterHash;


	SimpleList* indexes;
	SimpleList* reverse;
	Alias* aliases;
	Cache* cache;
	Access* access;
	Filter* filter;

	int requests;
	int sent;

	int requestsLast;
	int sentLast;

	int requestsToday;
	long long sentToday;

	Server* next;
};


struct Alias_struct {
	int timeout;
	int expires;
	int keepalive;
	int log;
	int listing;
	int throttle;
	int charset;
	int srcCharset;

	int type;
	int length;
	char* alias;
	char* location;
	char* aux[8];

	char exact;
	char exactLocation;

	int nOfTails;
	char** tail;
	int* tailLength;

	int nOfCtype;
	char** ctype;
	int* ctypeLength;

	SimpleList* indexes;
	SimpleList* reverse;
	Cache* cache;
	Proxy* proxy;
	CGIServer* cgi;
	Access* access;
	Filter* filter;

	Alias* next;
};

struct Listener_struct {
	unsigned char type;
	struct in_addr addr;
	int port;
	int fd;
	unsigned accepts;
	unsigned acceptsLast;
	unsigned acceptsToday;
	unsigned droppedS;
	unsigned droppedC;
	char* error;
	Listener* next;
};


struct Cache_struct {
	int size;
	int age;
};

struct Backend_struct {
    char* name;
	struct in_addr addr;
	unsigned short port;
	unsigned short state;
	time_t last;
	unsigned nRequests;
	unsigned maxRequests;
	unsigned requests;
	unsigned requestsLast;
	unsigned requestsToday;
	unsigned lastRequests;

	Backend* next;
};

struct Proxy_struct {
	Backend* backend;
	char* location;

	Proxy* next;
};

struct CGI_struct {
};


struct Filter_struct {
	char* type;
	int typeLength;
	void (*handler)(Request* request, char flush);
};

struct FilterState_struct {
	char state;
	char quote;
	short pos;
};

struct SimpleList_struct {
	char* name;
	void* link;
	SimpleList* next;
};

struct Access_struct {
	int code;

	int type;

	char* realm;
	char* passwd;
	char* file;
	Pairs* users;
	time_t mtime;

	int country;

	int network;
	int mask;

	int refType;
	char* referrer;

	char* agent;

	char* redirect;

	Access* next;
};


struct Request_struct {
	unsigned char type;
	unsigned char state;
	char watchRead;
	char watchWrite;

	Server* server;
	Alias* alias;
	unsigned ip;
	unsigned realIP;
	int fd;
	int file;
	time_t last;
	int throttled;

	/*	request parameters */
	time_t start;
	short status;
	unsigned char method;
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	char wasRequest;
	int pipelined;

	char* uri;
	char* path;
	char* params;
	char* query;
	Buffer* buffer;
	Buffer* postBuffer;

	/*	request headers */
	char* userAgent;
	char* referrer;
	char* authorization;
	char* user;
	char* auth;
	char* cookie;
	char* host;
	char* inContentType;
	int inContentLength;
	time_t ims;
	time_t iums;
	time_t ifRange;
	unsigned forwardedFor;
	char* inHeaders[MAX_OTHER_HEADERS + 1];

	int rangeTotal;
	int rangeFrom;
	int rangeTo;

	/* reply headers */
	char* contentType;
	char* location;
	char* setCookie;
	int length;
	int fileSize;
	time_t modified;
	time_t expires;
	char* outHeaders[MAX_OTHER_HEADERS + 1];

	char* body;
	char* alloc;
	char* content;
	int offset;
	int received;
	int filtered;
	int sent;
	int contentSize;

	void* mmap;

	ProxyRequest* proxy;
	CGIRequest* cgi;
	FilterState filter;
	Config* config;

	Request* prev;
	Request* next;
};


union FD_union {
	char* type;
	Listener* listener;
	Request* request;
	ProxyRequest* proxy;
	CGIRequest* cgi;
};

struct ProxyRequest_struct {
	unsigned char type;
	unsigned char state;

	char watchRead;
	char watchWrite;

	int fd;
	time_t last;


	/*	request parameters */
	time_t start;
	Request* request;
	Proxy* proxy;
	Buffer* buffer;
	int posted;

	/* reply headers */
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	int status;
	char* contentType;
	char* location;
	char* setCookie;
	char* auth;
	int length;
	int rangeFrom;
	int rangeTo;
	int rangeTotal;
	time_t modified;
	time_t expires;
	char* outHeaders[MAX_OTHER_HEADERS + 1];
	int offset;

	ProxyRequest* prev;
	ProxyRequest* next;
};

struct CGIRequest_struct {
	unsigned char type;
	unsigned char state;

	int fd;
	time_t last;

	/*	request parameters */
	Request* request;
	Buffer* buffer;
	int posted;

	/* reply headers */
	char major;
	char minor;
	char keepalive;
	char acceptRanges;
	int status;
	char* contentType;
	char* location;
	char* setCookie;
	char* auth;
	int length;
	int rangeFrom;
	int rangeTo;
	int rangeTotal;
	time_t modified;
	time_t expires;
	char* outHeaders[MAX_OTHER_HEADERS + 1];
};



extern int doRestart, doReload, doShutdown, doQuit;
extern int myPID;
extern time_t start, configured;
extern char built[];
extern char past[SHORT_STR];
extern char future[SHORT_STR];

#endif
